#ifdef WIN32
#pragma warning(disable : 4786 )
#endif

#include "CSAction.h"
#include "SDLMain.h"
#include "CSXMLHelper.h"

CSActionLoader CSActionLoader::INSTANCE;
const char *CSAction::CLASS = "CSAction";

/**
 Private Constructor of an CSAction. Actions can only be instantiated through the
 CSActionLoader
*/
CSAction::CSAction(const std::string &filename)
{
	static char *functionName="CSAction";
	LOG_ENTER 
	CSActionData data;
	loadActionData(filename, data);
	initialize(data);
	LOG_EXIT
}

CSAction::CSAction(const CSAction &action)
{
	static char *functionName="CSAction";
	LOG_ENTER 
	int i;
	initialize();
	mIsBreakable = action.mIsBreakable;
	mRepeatable = action.mRepeatable;
	mSizePhases = action.mSizePhases;
	mEnvoker = action.mEnvoker;
	mType = action.mType;

	mPhases = new unsigned int[mSizePhases];
	for (i=0; i<mSizePhases; i++)
	{
		mPhases[i] = action.mPhases[i];
	}

	mAnimations = new CSAnimations();
	CSAnimations::const_iterator iter = action.mAnimations->begin();
	while (iter != action.mAnimations->end())
	{
		// animations are singletons!
		// that means, we can (and should) re-use the animation we find at
		// the action that we copy from
		mAnimations->push_back(*iter);	
		iter++;
	}
	mId = strdup(action.mId);
	mNextAction = strdup(action.mNextAction);
	LOG_EXIT
}

void CSAction::initialize(void)
{
	static char *functionName="initialize";
	LOG_ENTER 
	mIsBreakable = true;
	mRepeatable = CS_REPEATABLE;
	mSizePhases = 0;
	mPhases = 0;
	mType = 0;
	mEnvoker = 0;
	mId = 0;
	mNextAction = 0;
	mAnimations = 0;
	LOG_EXIT
}

void CSAction::initialize(const CSActionData &data)
{
	static char *functionName="initialize";
	LOG_ENTER 
	int i;
	initialize();

	mType = data.type;
	mIsBreakable = data.isBreakable;
	mRepeatable = data.repeatable;
	mSizePhases = data.sizePhases;
	mEnvoker = data.envoker;

	mPhases = new unsigned int[mSizePhases];
	for (i=0; i<mSizePhases; i++)
	{
		mPhases[i] = data.phases[i];
	}
		
	mAnimations = new CSAnimations();
	for (StringVector::iterator iter = data.animationNames->begin(); iter != data.animationNames->end(); iter++)
	{
		char *help = (char *) *iter; 
		if (help == NULL)
		{
			LOG_EXIT
			SDLMain::shutdown((std::string)"Char Pointer == 0 Action Loading: " + SDL_GetError(), 1);
		}
		// create all animations (with ini - filename)
		CSAnimation *animation = CSAnimationLoader::INSTANCE.load((char *)(help));
		mAnimations->push_back(animation);	
	} 

	mId = strdup(data.id);
	mNextAction = strdup(data.nextAction);
	LOG_EXIT
}

CSAction::~CSAction()
{
	static char *functionName="~CSAction";
	LOG_ENTER 
	if (mPhases != 0)
	{
		delete []mPhases;
		mPhases = 0;
	}

	if (mAnimations != 0)
	{
		// Animation are singletons -> do not delete them 
		// Animation are deleted by CSAnimationLoader!
		delete mAnimations;
		mAnimations = 0;
	}
	
	if (mId != 0)
	{
		free(mId);
		mId = 0;
	}

	if (mNextAction != 0)
	{
		free(mNextAction);
		mNextAction = 0;
	}
	LOG_EXIT
}

// build an initial state for this action, sets variables to defaults
ActionState CSAction::buildState()
{
	static char *functionName="buildState";
	ActionState state;
	state.mFinished = false;
	state.mForward = true;
	if (mSizePhases > 0)
	{
		state.mCurrentPhase = 0;
		state.mCurrentAnimation = mPhases[state.mCurrentPhase];
		state.mAnimation = mAnimations->at(state.mCurrentAnimation);
		state.mAnimationState = state.mAnimation->buildState();
	}
	return state;
}

// resets a state to default values
// doesn't  update mDisplayParams! 
// (this position and mDisplayPicture do not change!)
void CSAction::resetState(ActionState& state)
{
	static char *functionName="resetState";
	state.mFinished = false;
	state.mForward = true;
	if (mSizePhases > 0)
	{
		state.mCurrentPhase = 0;
		state.mCurrentAnimation = mPhases[state.mCurrentPhase];
		state.mAnimation = mAnimations->at(state.mCurrentAnimation);
		state.mAnimation->resetState(state.mAnimationState);
	}
}
		
// resets a state to start values of this action -> immediately
// - happens when an event for another action occurs or so
void CSAction::startAction(ActionState &state)
{
	static char *functionName="startAction";
	state.mFinished = false;
	state.mForward = true;
	if (mSizePhases > 0)
	{
		state.mCurrentPhase = 0;
		state.mCurrentAnimation = mPhases[state.mCurrentPhase];
		state.mAnimation = mAnimations->at(state.mCurrentAnimation);
		state.mAnimation->startAnimation(state.mAnimationState);
	}
}

// resets a state to start values of this action
// this will not be displayed during next display() call
// - happens when one action ends during update, and a next
// action is initiated - this means the next action is
// active from the start of next update, doesn't interfere with
// current mDisplayParams!
void CSAction::startNextAction(ActionState &state)
{
	static char *functionName="startNextAction";
	state.mFinished = false;
	state.mForward = true;
	if (mSizePhases > 0)
	{
		state.mCurrentPhase = 0;
		state.mCurrentAnimation = mPhases[state.mCurrentPhase];
		state.mAnimation = mAnimations->at(state.mCurrentAnimation);
		state.mAnimation->startNextAnimation(state.mAnimationState);
	}
}

void CSAction::setSecPerFrame(float secPerFrame)
{
	static char *functionName="setSecPerFrame";
	LOG_ENTER 
	CSAnimations::const_iterator iter = mAnimations->begin();
	while (iter != mAnimations->end())
	{
		CSAnimation *animation = *iter;
		animation->setSecPerFrame(secPerFrame);
		iter++;
	}
	LOG_EXIT
}

void CSAction::setAlpha(int alpha)
{
	static char *functionName="setAlpha";
	LOG_ENTER 
	CSAnimations::const_iterator iter = mAnimations->begin();
	while (iter != mAnimations->end())
	{
		CSAnimation *animation = *iter;
		animation->setAlpha(alpha);
		iter++;
	}
	LOG_EXIT
}

// if finished and not repeat -> false, true otherwise
// does not display anything, but prepares (update) for next displaying
bool CSAction::next(ActionState &state)
{
	static char *functionName="next";
	bool finished = false;
	if (state.mFinished)
	{
		return !state.mFinished;
	}

	finished = !(state.mAnimation->next(state.mAnimationState));
	if (finished)
	{
		if (mRepeatable == CS_RANDOM)
		{
			// TODO
		}
		else
		{
			if (state.mForward)
			{
				state.mCurrentPhase++;
				if (state.mCurrentPhase == mSizePhases)
				{
					if (mRepeatable == CS_INVERSE_REPEATABLE)
					{
						state.mForward = !state.mForward;
						if (mSizePhases > 1)
						{
							state.mCurrentPhase-=2;
						}
						else
						{
							state.mCurrentPhase--;
						}
					}
					else
					if (mRepeatable == CS_INVERSE_ONCE)
					{
						state.mForward = !state.mForward;
						if (mSizePhases > 1)
						{
							state.mCurrentPhase-=2;
						}
						else
						{
							state.mCurrentPhase--;
						}
					}
					else
					{
						if (mRepeatable == CS_NOT_REPEATABLE)
						{
							state.mFinished = true;
							return !state.mFinished;
						}
						state.mCurrentPhase = 0;
					}
				}
			}
			else // mForward
			{
				state.mCurrentPhase--;
				if (state.mCurrentPhase == -1)
				{
					if (mRepeatable == CS_INVERSE_REPEATABLE)
					{
						state.mForward = !state.mForward;
						if (mSizePhases > 1)
						{
							state.mCurrentPhase+=2;
						}
						else
						{
							state.mCurrentPhase++;
						}
					}
					else
					{
						if (mRepeatable == CS_INVERSE_ONCE)
						{
							state.mFinished = true;
							return !state.mFinished;
						}
						state.mCurrentPhase = 0;
					}

				}
			}

		} // else (mRepeatable == CS_RANDOM)
		state.mCurrentAnimation = mPhases[state.mCurrentPhase];
		state.mAnimation = mAnimations->at(state.mCurrentAnimation);
		state.mAnimation->startAnimation(state.mAnimationState);
	} // (mDelayCounter == mDelays[mPhases[mCurrentPhase]])

	return !state.mFinished;
}

// not much to do here, to display an animation (picture)
void CSAction::display(ActionState &state)
{
	static char *functionName="display";
	state.mAnimation->display(state.mAnimationState);
}

bool CSAction::isBreakable(ActionState &state)
{
	static char *functionName="isBreakable";
	if (!mIsBreakable)
	{
		return mIsBreakable;
	}
	return state.mAnimation->isBreakable();
}

bool CSAction::isAction(char *id)
{
	static char *functionName="isAction";
	return !(strcmp(id, mId));
}

// of all pictures belonging to this action 
// return the largest width
unsigned int CSAction::getMaxX()
{
	static char *functionName="getMaxX";
	unsigned int maxX = 0;
	unsigned int x;
	for (CSAnimations::iterator iter = mAnimations->begin(); iter != mAnimations->end(); iter++)
	{
		CSAnimation *ani = *iter;
		x = ani->getMaxX(); if (maxX < x) maxX = x;
	}
	return maxX;
}

// of all pictures belonging to this action 
// return the largest height
unsigned int CSAction::getMaxY()
{
	static char *functionName="getMaxY";
	unsigned int maxY = 0;
	unsigned int y;

	for (CSAnimations::iterator iter = mAnimations->begin(); iter != mAnimations->end(); iter++)
	{
		CSAnimation *ani = *iter;
		y = ani->getMaxY(); if (maxY < y) maxY = y;
	}
	return maxY;
}

void CSAction::loadActionData(const std::string &filename, CSActionData &data)
{
	static char *functionName="loadActionData";
	LOG_ENTER 
	char tmp[10];
	
	CSXMLHelper xmlSupport(filename, "ACTION");
	try
	{
		if (xmlSupport.getError())
		{
			LOG_EXIT
			throw "error";
		}

		data.id = strdup(xmlSupport.getString("ID").c_str());
		data.isBreakable = xmlSupport.getInt("IS_BREAKABLE") == 1;
		data.type = xmlSupport.getInt("TYPE");
		data.repeatable = xmlSupport.getInt("REPEATABLE");

		data.envoker = xmlSupport.getInt("ENVOKER");

		if (xmlSupport.getInt("count(/ACTION/DEFAULT_NEXT_ACTION)") == 1)
		{
			data.nextAction = strdup(xmlSupport.getString("DEFAULT_NEXT_ACTION").c_str());
		}
		data.sizePhases = xmlSupport.getInt("count(/ACTION/PHASE)");
		data.phases = new unsigned int[data.sizePhases];
		int i;
		for (i=0; i<data.sizePhases; i++)
		{
			// i+1, predicates in xpath start with 1
			data.phases[i] = xmlSupport.getInt((std::string)"/ACTION/PHASE["+itoa(i+1,tmp,10)+"]");
		}

		int animationCount = xmlSupport.getInt("count(/ACTION/ANIMATION)");
		data.animationNames = new StringVector();
		for (i=0; i<animationCount; i++)
		{
			// i+1, predicates in xpath start with 1
			std::string a = xmlSupport.getString((std::string)"/ACTION/ANIMATION["+itoa(i+1,tmp,10)+"]");
			data.animationNames->push_back(strdup(a.c_str()));
		}
	}
	catch(...)
	{
		LOG_EXIT
		SDLMain::shutdown((std::string)"XML error \"" + filename + "\": " + xmlSupport.getErrorMessage().c_str(), 1);
	}
	LOG_EXIT
}
