#ifndef CSACTION_H
#define CSACTION_H

#include <string>
#include <vector>
#include "CSTypes.h"
#include "CSAnimation.h"
#include "CSLoadable.h"
#include "CSLog.h"

class CSActionLoader;
class CSAction;

const int NORTH_ACTION = 0;
const int SOUTH_ACTION = 1;
const int WEST_ACTION = 2;
const int EAST_ACTION = 3;
const int STAY_ACTION = 4;

typedef std::vector<CSAction *> CSActions;

//! data classes are used for loading a loadable class
//! first the data is loaded via xml into this "DATA"-class
//! from this data - the actual action is created
//! the data can be reused to create another instance of an
//! action class - though probably not needed...
class CSActionData
{
	public:
		char *id;								//!< id of action, unique
		int type;
		char *nextAction;						//!< (default) nextAction (id)
		bool isBreakable;						//!< is this action breakable in the midth of phases
		unsigned int  repeatable;				//!< status of repeatable, see animation

		unsigned int envoker;					//!< from what is this action envoked
		unsigned int sizePhases;				//!< how many phases?	
		unsigned int *phases;					//!< arrayPointer # sizePhases
		StringVector *animationNames;			//!< animations filenames (xml)

		CSActionData()
		{
			id = 0;
			nextAction = 0;
			envoker = 0;
			isBreakable = true;
			repeatable = 0;
			animationNames = 0;
			sizePhases = 0;
			phases = 0;
		}

		~CSActionData()
		{
			if (id != 0)
			{
				free (id);
				id = 0;
			}
			if (nextAction != 0)
			{
				free (nextAction);
				nextAction = 0;
			}
			if (phases != 0)
			{
				delete []phases;
				phases = 0;
			}
			if (animationNames != 0)
			{
				for (StringVector::iterator iter = animationNames->begin(); iter != animationNames->end(); iter++)
				{
					delete *iter;
				}
				delete animationNames;
				animationNames = 0;
			}	
		}
};

//! the states hold all variable parts needed for printing the 
//! action on screen
//! putting these in a state class, enables us to
//! reuse the memory expensive action class!
class ActionState
{
	public:
		AnimationState mAnimationState;		//!< part of each action are animations
											//!< so is the state part of this state
		
		bool mForward;						//!< in animation sequence going ++ or --
		bool mFinished;						//!< if a NOT_REPEATABLE animation,
											//!< this is false upon "one round"
		CSAnimation *mAnimation;			//!< current animation
		signed int mCurrentPhase;			//!< current "active" phaseNo
		unsigned int mCurrentAnimation;		//!< current active animation, corresponse to #
		ActionState() 
		{
			mForward = true;				//!< in animation sequence going ++ or --
			mFinished = false;				//!< if a NOT_REPEATABLE animation,
			mAnimation = 0;
			mCurrentPhase = -1;
			mCurrentAnimation = -1;			//!< current active animation, corresponse to 
		}
		virtual ~ActionState() {}
};

/**
	This file defines CSAction:

	  CSAction can only be created thru a loader (CSActionLoader).
	  CSAction are defined by (now ONLY thru) an xml-File, like:
	  DTD:
\verbatim
		<!ELEMENT ACTION (ID,IS_BREAKABLE,REPEATABLE,(PHASE)+,(ANIMATION)+,ENVOKER,(DEFAULT_NEXT_ACTION)?)>
		<!ELEMENT ID (#PCDATA)>
		<!ELEMENT IS_BREAKABLE (#PCDATA)>
		<!ELEMENT REPEATABLE (#PCDATA)>
		<!ELEMENT PHASE (#PCDATA)>
		<!ELEMENT ANIMATION (#PCDATA)>
		<!ELEMENT ENVOKER (#PCDATA)>
		<!ELEMENT DEFAULT_NEXT_ACTION (#PCDATA)>
\endverbatim
	  Example:
  
\verbatim
		<ACTION>
			<ID>PAC_DOWN_ACTION</ID>
			<IS_BREAKABLE>&true;</IS_BREAKABLE>
			<REPEATABLE>&INVERSE_ONCE;</REPEATABLE>
			<PHASE>0</PHASE>
			<ANIMATION>animation\\pac_down.xml</ANIMATION>
			<ENVOKER>&DOWN;</ENVOKER>
			<DEFAULT_NEXT_ACTION>PAC_DOWN_STAY_ACTION</DEFAULT_NEXT_ACTION>
		</ACTION>
\endverbatim
<PRE>
	  ID	       - supposed a unique identifier
	  IS_BREAKABLE - if an action can be interrupted by some action
	  REAPEATABLE  - one of:
					 ACTION_NOT_REPEATABLE 		0
					 ACTION_REPEATABLE 			1
					 ACTION_INVERSE_ONCE 		2
					 ACTION_INVERSE_REPEATABLE 	3
					 ACTION_RANDOM 				4

	  PHASES       - many phases, each phase is represented by a number
					 the number corresponse to the ANIMATION that is
					 loaded with the ini - file (numbered from top to
					 bottom.
					 NOTE:
						REAPEATABLE			ACTION_INVERSE_ONCE
						PHASES				0
						PHASES				1
						PHASES				2
						PHASES				3
						PHASES				4
				
				     gives the same result as:
						REAPEATABLE			ACTION_NOT_REPEATABLE
						PHASES				0
						PHASES				1
						PHASES				2
						PHASES				3
						PHASES				4
						PHASES				3
						PHASES				2
						PHASES				1
						PHASES				0

	  ANIMATION    - many animations are possible!
					 describes the animation to be loaded,
	                 parameter gives the ini-file for a CSAnimation

	  ENVOKER      - what 'envokes' the action, usually for a sprite
					 this will be something like UP - DOWN 
					 (see definition in message.h)
	  
	  DEFAULT_NEXT_ACTION - the unique id of another action, that will be envoked
					 if this action 'ends' 
					 (JUMP - might be an action, but at some stage JUMP WILL end,
					  than another action will per default be envoked,
					  like STAND)
					  This parameter is optional, it overrides (if set), the
					  DEFAULT_ACTION of a CSSprite.
					  This default - actions is not loaded with this action.
					  To have any effect it must also be loaded with the
					  corresponding CSSprite.

	  Methods:
		unsigned int getMaxX(); 
		unsigned int getMaxY(); 
		void setSecPerFrame(float secPerFrame);

		ActionState buildState();
		void resetState(ActionState &state);
		void startAction(ActionState &state);
		void startNextAction(ActionState &state);
		bool isBreakable(ActionState &state);
		virtual bool next(ActionState &state);			// if finished and not repeat -> false, true otherwise
		virtual void display(ActionState &state);	

		// inlines
		unsigned int getEnvoker() 
		bool isAction(char *id);
		char *getId()									// do not free return pointer!
		char *getDefaultAction()						// do not free return pointer!
</PRE>		
	the action class has no instance relevant data (state independant)
	thus one instance can be reused 
	one action can be (re) used severall times
	the data used for printing must come from the outside (ActionState)
	*/
class CSAction
{
	friend CSActionLoader;
	private:
		bool mIsBreakable;					//!< can action be interrupted?
											//!< by events 
		char *mId;							//!< id of action (unique)
		int mType;
		char *mNextAction;					//!< id of next (default) action
		unsigned int mRepeatable;			//!< status of repeatable, see animation
		unsigned int mSizePhases;			//!< how many phases
		unsigned int *mPhases;				//!< arrayPointer # mSizePhases
		unsigned int mEnvoker;				//!< envoker corresponds to subtype of a ActionMessage
											//!< see message.h
		CSAnimations *mAnimations;			//!< all animations this action is made of

		//! constructor using "*.xml" file
		CSAction(const std::string &filename);
		static void loadActionData(const std::string &filename, CSActionData &data);

		void initialize(const CSActionData &data);
		void initialize();
	
	public:
		CSAction(const CSAction &action);
		virtual ~CSAction();
		static const char *CLASS;
		virtual std::string getType() {return (std::string) CLASS;}

		unsigned int getMaxX(); 
		unsigned int getMaxY(); 
		void setSecPerFrame(float secPerFrame);
		bool isAction(char *id);

		ActionState buildState();
		void resetState(ActionState &state);
		void startAction(ActionState &state);
		void startNextAction(ActionState &state);
		bool isBreakable(ActionState &state);
		virtual bool next(ActionState &state);			//!< if finished and not repeat -> false, true otherwise
		virtual void display(ActionState &state);	
		void setAlpha(int alpha);

		unsigned int getEnvoker() {return mEnvoker;}
		char *getId() {return mId;}						//!< do not free return pointer!
		int getActionType() {return mType;}
		char *getDefaultAction() {return mNextAction;}	//!< do not free return pointer!
};		

//! factory - class
//! that is used to create CSAction
//! uses xml - data-files, produces "singletons"
class CSActionLoader : public Loadable<CSAction>
{
	protected:
		virtual CSAction *create(const std::string &filename)
		{
			return new CSAction(filename);
		}

	public:
		static CSActionLoader INSTANCE;
};

#endif // CSACTION_H
