#ifndef CSAnimation_h
#define CSAnimation_h

// USES SDL_Surface

#include <string>
#include <vector>
#include "CSTypes.h"
#include "SDLMain.h"
#include "CSPicture.h"
#include "CSLoadable.h"
#include "CSLog.h"

class CSAnimation;
class CSAnimationLoader;
class CSAnimationPicture;

typedef std::vector<CSAnimationPicture *> CSAnimationPictures;
typedef std::vector<CSAnimation *> CSAnimations;

static const int CS_NOT_REPEATABLE = 0;
static const int CS_REPEATABLE = 1;
static const int CS_INVERSE_ONCE = 2;
static const int CS_INVERSE_REPEATABLE = 3;
static const int CS_RANDOM = 4;				//!< (not implemented)
static const int ANIMATION_SPEED = 65536;
static const int ADJUST_SPEED = 65535;


//! class to define a single animated picture-frame, data wise -> see below
class CSAnimationPicture
{
	public:
		char *filename;							//!< defining picture filename!
		int animationRate;						//!< delay for one update of the animation (per picture)
		signed int moveX;						//!< delta X to move (after delay) (per picture)
		signed int moveY;						//!< delta Y to move (after delay) (per picture)
		CSAnimationPicture()
		{
			filename = 0;
		}
		~CSAnimationPicture()
		{
			if (filename != 0)
			{
				free(filename);
				filename = 0;
			}
		}
};

//! data classes are used for loading a loadable class
//! first the data is loaded via xml into this "DATA"-class
//! from this data - the actual animation is created
//! the data can be reused to create another instance of a
//! animation class - though probably not needed...
class CSAnimationData
{
	public:
		char *id;								//!< defining picture filename!
		bool isBreakable;						//!< cann animation be stopped while 
												//!< in between
		unsigned int repeatable;				//!< see above definition
		unsigned int sizePhases;				//!< how many phases?	
		unsigned int *phases;					//!< arrayPointer # sizePhases
		CSAnimationPictures *pictures;			//!< above data for animation pictures
		
		CSAnimationData()
		{
			phases = 0;
			id = 0;
			pictures = 0;
		}

		~CSAnimationData()
		{
			if (id != 0)
			{
				free (id);
				id = 0;
			}
			if (pictures != 0)
			{
				CSAnimationPictures::iterator iter = pictures->begin();

				while (iter != pictures->end())
				{
					CSAnimationPicture *pic = *iter;
					delete pic;
					iter++;
				}
				delete pictures;
				pictures = 0;
			}	
			if (phases != 0)
			{
				delete []phases;
				phases = 0;
			}
		}

};

/** a phase is a  sequence of pictures. Same pictures can occur at different stages
   during an animation.
   Example:
		pictures a1, a2, a3, a4 (pictures actually belonging to animation
		phases a1, a2, a1, a2, a3, a1, a2, a3, a4
		The phases defines "sequences" of pictures during one animation.

		the states hold all variable parts needed for printing the animation on screen
		putting these in a state class, enables us to
		 reuse the memory expensive animation class!
*/
class AnimationState
{
	public:
		float mXPos;						//!< current x position on screen
		float mYPos;						//!< current y position on screen
		int mSpeedX;						//!< current x speed, pixels per frame
		int mSpeedY;						//!< current y speed, pixels per frame
		unsigned int mAdjustSpeed;
		float mAnimationRateCounter;		//!< "speed" of anaimtion
		signed int mCurrentPhase;			//!< current "active" phaseNo
		unsigned int mCurrentPicture;		//!< current active pictureNo, corresponse to 
											//!< picture selected by active phase
		bool mForward;						//!< in animation sequence going ++ or --
		bool mFinished;						//!< if a NOT_REPEATABLE animation,
		CSPicture *mPicture;				//!< currenct active CSPicture, corresponse to 
											//!< picture selected by active phase
											//!< this need not be the picture to be displayed
											//!< this round, since update and display 
											//!< are two different methods
											//!< the mDisplayPicture will be set during
											//!< update to display the picture that belongs to 
											//!< the "current" frame
											//!< at the end of the update method the 
											//!< the current picture of the state
											//!< can differ, because the update can
											//!< sets the picture to the NEXT picture
											//!< (animation counter reaches 0, next frame
											//!< will be selected, but the old frame
											//!< should still be displayed)
											//!< !NOT TO BE FREED!
		CSPicture *mDisplayPicture;			//!< picture that will be displayed when calling display()
											//!< !NOT TO BE FREED!
		CSDisplayParams mDisplayParams;		//!< position on screen and in world of the current mDisplaypicture

		AnimationState() 
		{
			mXPos = 0;
			mYPos = 0;
			mSpeedX = ANIMATION_SPEED;
			mSpeedY = ANIMATION_SPEED;
			mAdjustSpeed = 0;
			mAnimationRateCounter = 0;
			mCurrentPhase = 0;
			mCurrentPicture = 0;
			mForward = true;
			mFinished = false;
			mPicture = 0;
			mDisplayPicture = 0;

			mDisplayParams.mXDisplayStart = 0;
			mDisplayParams.mYDisplayStart = 0;
			mDisplayParams.mXWorldStart = 0;
			mDisplayParams.mYWorldStart = 0;
			mDisplayParams.mXPos = 0;
			mDisplayParams.mYPos = 0;
		}
		virtual ~AnimationState() {}
};

/*!
	CSAnimation
	  CSAnimation can only be created thru a loader (CSAnimationLoader).
	  CSAnimation are defined by (now ONLY thru) an xml-File, like:
		DTD:
\verbatim

		<!ENTITY % types SYSTEM "types.dtd"> %types;

		<!ELEMENT ANIMATION (ID,IS_BREAKABLE,REPEATABLE,(PHASE)+,(IMAGE_DATA)+)>
		<!ELEMENT ID (#PCDATA)>
		<!ELEMENT IS_BREAKABLE (#PCDATA)>
		<!ELEMENT REPEATABLE (#PCDATA)>
		<!ELEMENT PHASE (#PCDATA)>
		<!ELEMENT IMAGE_DATA (PICTURE,ANIMATION_RATE,MOVE_X,MOVE_Y)>
		<!ELEMENT PICTURE (#PCDATA)>
		<!ELEMENT ANIMATION_RATE (#PCDATA)>
		<!ELEMENT MOVE_X (#PCDATA)>
		<!ELEMENT MOVE_Y (#PCDATA)>
\endverbatim

		  Example:
\verbatim
		<ANIMATION>
			<ID>PAC_DOWN</ID>
			<IS_BREAKABLE>&true;</IS_BREAKABLE>
			<REPEATABLE>&INVERSE_ONCE;</REPEATABLE>
			<PHASE>0</PHASE>
			<PHASE>1</PHASE>
			<PHASE>2</PHASE>
			<PHASE>3</PHASE>
			<PHASE>4</PHASE>
			<IMAGE_DATA>
				<PICTURE>picture\\pac_down1.xml</PICTURE>
				<ANIMATION_RATE>&A_RATE;</ANIMATION_RATE>
				<MOVE_X>0</MOVE_X>
				<MOVE_Y>&SPEED;</MOVE_Y>
			</IMAGE_DATA>
			<IMAGE_DATA>
				<PICTURE>picture\\pac_down2.xml</PICTURE>
				<ANIMATION_RATE>&A_RATE;</ANIMATION_RATE>
				<MOVE_X>0</MOVE_X>
				<MOVE_Y>&SPEED;</MOVE_Y>
			</IMAGE_DATA>
			...
		</ANIMATION>
\endverbatim
<PRE>	  
	  ID	       - supposed a unique identifier
	  IS_BREAKABLE - if an animation can be interrupted by some action
	  REAPEATABLE  - one of:
						CS_NOT_REPEATABLE = 0;
						CS_REPEATABLE = 1;
						CS_INVERSE_ONCE = 2;
						CS_INVERSE_REPEATABLE = 3;
						CS_RANDOM = 4;				// (not implemented)

	  PHASES       - many phases, each phase is represented by a number
					 the number corresponse to the IMAGE_DATA that is
					 loaded with the ini - file (numbered from top to
					 bottom.
					 NOTE:
						REAPEATABLE			CS_INVERSE_ONCE
						PHASES				0
						PHASES				1
						PHASES				2
						PHASES				3
						PHASES				4
				
				     gives the same result as:
						REAPEATABLE			CS_NOT_REPEATABLE
						PHASES				0
						PHASES				1
						PHASES				2
						PHASES				3
						PHASES				4
						PHASES				3
						PHASES				2
						PHASES				1
						PHASES				0

	  if an animation has only one image, it is not really
	  much of an aniamtion ->
	  thus an aniamtion with only one picture never uses aniamtion rate, 
	  and never uses the repeatable flag
	  IMAGE_DATA   - many image_data's are possible
					 describes the image to be loaded, 
	                 PICTURE parameter gives the ini-file for a CSPicture,
					 ANIMATION_RATE  parameter gives the delay between two animation phases
		    						 negative animation rates are not in animtions per second
									 but aniamtion each # of seconds!
					 MOVE_X parameter gives the value (speed) with which the
					    the x-coordinates wil be updated (negative and positive) pixels per second
					 MOVE_Y parameter gives the value (speed) with which the
					    the y-coordinates wil be updated (negative and positive) pixels per second

	  CSAnimation are singletons in the sense, that two "loaded" CSAnimations of 
	  the same "ini" file ARE the same instances. 

	</PRE>
	the animation class has no instance relevant data (state independant)
	thus one instance can be reused 
	- one aniamtion can be (re) used severall times
	the data used for printing must come from the outside (AnimationState)
	*/
class CSAnimation
{
	friend CSAnimationLoader;
	private:
		bool mIsBreakable;					//!< can animation be interrupted?
											//!< by events 
		unsigned int mRepeatable;			//!< above const definitions
		unsigned int mSizePictures;			//!< how many pictures
		unsigned int mSizePhases;			//!< how many phases
		float *mAnimationRates;				//!< arrayPointer # mSizePictures
		unsigned int *mPhases;				//!< arrayPointer # mSizePhases
		signed int *mMoveX;					//!< arrayPointer # mSizePictures
		signed int *mMoveY;					//!< arrayPointer # mSizePictures
		
		CSPictures *mPictures;				//!< all pictures in a vector (stl)

		bool mMoveAllowed;					//!< may at some stage be forbidden
											//!< from outside (hits a wall...)
											//!< this is false upon "one round"
		bool mIsSolid;						//!< has transparent pixels
		float mSecsPerFrame;				//!< set from outside, to provide speed independent movement
		
		//! constructor using "*.xml" file
		CSAnimation(const std::string &filename);
		static void loadAnimationData(const std::string &filename, CSAnimationData &data);

		virtual void initialize(const CSAnimationData &data);
		virtual void initialize();
	
	public:
		static const char *CLASS;
		virtual std::string getType() {return (std::string) CLASS;}

		CSAnimation(const CSAnimation &animation);
		virtual ~CSAnimation();

		unsigned int getMaxX();
		unsigned int getMaxY();
		void setSolid(bool b);
		void scale(int width, int tilesX, int height, int tilesY);

		AnimationState buildState();
		void resetState(AnimationState& state);	
		void startAnimation(AnimationState& state);
		void startNextAnimation(AnimationState& state); //!< comment
		bool next(AnimationState& state);	//!< if finished and not repeat -> false, true otherwise
		void display(AnimationState& state);	
		void setAlpha(int alpha);
		SDL_Surface *getScaledAnimation(AnimationState &mAnimationState, double factor)
		{
			//! \todo: do all animation states beforehand
			return mAnimationState.mPicture->getScaledPicture(factor);
		}
		

		bool isSolid() {return mIsSolid;}
		bool isBreakable() {return mIsBreakable;}
		void setSecPerFrame(float secPerFrame);
		void setSpeedX(int speed, AnimationState& state) {state.mSpeedX = speed;}
		void setSpeedY(int speed, AnimationState& state) {state.mSpeedY = speed;}
		void adjustSpeed(int speed, AnimationState& state) 
		{state.mSpeedX = state.mSpeedY = ADJUST_SPEED;state.mAdjustSpeed = speed;}
		void resetSpeed(AnimationState& state) {state.mSpeedX = state.mSpeedY = ANIMATION_SPEED;}
};

//! factory - class
//! that is used to create CSAniamtion
//! uses xml - data-files, produces "singletons"
class CSAnimationLoader : public Loadable<CSAnimation>
{
	protected:
		virtual CSAnimation *create(const std::string &filename)
		{
			return new CSAnimation(filename);
		}

	public:
		static CSAnimationLoader INSTANCE;
};
 
#endif // CSAnimation_h

