#ifndef CSGAME_H
#define CSGAME_H

// USES SDL_Rect 
// USES SDL_Events

#ifdef WIN32
#pragma warning(disable : 4786 )
#endif

#include "CSMessageDispatchable.h"
#include "CSMessageListener.h"
#include "CSMessage.h"
#include "CSSprite.h"
#include "CSWorld.h"
#include "CSFont.h"
#include "CSDesktop.h"
#include "CSMenuBar.h"
#include "CSMenu.h"
#include "CSMenuItem.h"
#include "CSXMLHelper.h"
#include "CSLog.h"
#include <map>

const int INIT_COUNT = 100;

const int CHECK_PIXEL = 0;
const int CHECK_HIT = 1;
const int CHECK_JUST_HIT = 2;

class CSGame;
class CSSample;

class TimedEvent;
typedef std::map<int, TimedEvent *> CSEventMap;
class TimedEvent
{
	public:
		int id;	
		bool useRepeat;
		int repeat;
		unsigned int finishTime;
		unsigned int intervall;
		void *data;
		void (*answer)(TimedEvent *);	//!< function pointer as CallBack
		TimedEvent()
		{
			id = 0;	
			useRepeat = false;
			repeat = 0;
			finishTime = 0;
			intervall = 0;
			data = 0;
			answer = 0;
		}
};

//! a data class just for level loading
class SpriteData
{
	public:
		char *name;
		unsigned int x;
		unsigned int y;
		unsigned int z;
		unsigned int speed;
		int behaviour;
};
typedef std::vector<SpriteData *> SpriteDatas;

//! a data class just for level loading
class SampleData
{
	public:
		char *name;
		char *sid;
};
typedef std::vector<SampleData *> SampleDatas;

class CSLevelData
{
	public:
		char *id;			
		SpriteData mainSprite;
		StringVector worlds;
		SpriteDatas sprites;
		SampleDatas samples;
		StringVector descriptions;

		CSLevelData()
		{
			id = 0;
		}

		~CSLevelData()
		{
			if (id != 0)
			{
				free (id);
				id = 0;
			}

			if (mainSprite.name != 0)
			{
				free(mainSprite.name);
				mainSprite.name = 0;
			}

			for (StringVector::iterator diter = descriptions.begin(); diter != descriptions.end(); diter++)
			{
				delete *diter;
			}
			descriptions.clear();

			for (StringVector::iterator witer = worlds.begin(); witer != worlds.end(); witer++)
			{
				delete *witer;
			}
			worlds.clear();

			for (SpriteDatas::iterator siter = sprites.begin(); siter != sprites.end(); siter++)
			{
				SpriteData *sprite = *siter;
				if (sprite->name != 0)
				{
					free(sprite->name);
				}
				delete sprite;
			}
			sprites.clear();

			for (SampleDatas::iterator ssiter = samples.begin(); ssiter != samples.end(); ssiter++)
			{
				SampleData *sample = *ssiter;
				if (sample->name != 0)
				{
					free(sample->name);
				}
				if (sample->sid != 0)
				{
					free(sample->sid);
				}
				delete sample;
			}
			samples.clear();
		}
};

class OutOfBoundsData
{
	public:
		int threshold;
		CSSprite *sprite;
		CSMessageListener *listener;
};

class TileCollisionCheckData
{
	public:
		TileCollisionCheckData()
		{
			sprite = 0;
			layer = 0;
			listener = 0;
			type = 0;
			threshold = 0;
		}
		CSSprite *sprite;
		int layer;
		CSMessageListener *listener;
		int type;
		int threshold;
};

class SpriteCollisionCheckData
{
	public:
		SpriteCollisionCheckData()
		{
			sprite = 0;
			other = 0;
			listener = 0;
			type = 0;
			threshold = 0;
		}
		CSSprite *sprite;
		CSSprite *other;
		CSMessageListener *listener;
		int type;
		int threshold;
};

typedef std::vector<OutOfBoundsData *> OutOfBoundsDatas;
typedef std::vector<TileCollisionCheckData *> TileCollisionCheckDatas;
typedef std::vector<SpriteCollisionCheckData *> SpriteCollisionCheckDatas;

class CSGame: public CSMessageDispatchable, public CSMessageListener
{
	private:
		char *mId;
		bool dummyGame;
		bool mIsGui;
		bool mLastIsGui;
		bool mEnableCheck;
		bool mDisplayScaled;
		unsigned int mThreshold;
		bool mIsWorldOutOfBounds;

		int mCurrentFrameRate;
		long mDiffTimeGUI;
		long mDiffTimeGame;
		long mDiffTimeAll;

		StringVector mLevels;
		CSEventMap *mEventMap;

		std::map<std::string, CSSample *> mSampleMap;
		TileCollisionCheckDatas mTileCollisionCheckDatas;
		SpriteCollisionCheckDatas mSpriteCollisionCheckDatas;
		OutOfBoundsDatas mOutOfBoundsDatas;
		
		void initialize();
		void toggleFullscreen();
		void toggleGui();
		void screenChanged();
		void adjustDisplay();
		void checkTimedEvent();
		void onMouseMotion(SDL_MouseMotionEvent event);
		void onMouseButtonDown(SDL_MouseButtonEvent event);
		void onMouseButtonUp(SDL_MouseButtonEvent event);
	
	protected:
		CSLevelData *mLevelData;
		bool mQuit;
		bool mPause;
		bool mLevelFinished;

		int mInitCount;
		float mCurrentTime;
		int mAverageFrameRate;
		float mGameTime;
		float mDiffTime;
		float mSecPerFrame;
		void handleEvents();
		int getLevelDataNum(CSSprite *sprite);

		CSDesktop *mDesktop;
		CSMenuBar *mMenuBar;
		CSMenu *mFileMenu;
		CSMenu *mHelpMenu;
		CSMenuItem *mMenuItemExit;
		CSMenuItem *mMenuItemAbout;
		CSSprites mSprites;
		CSSprite *mMainSprite;
		CSSprite *mScaleCenterSprite;
		CSWorld **mWorld;
		CSWorld *mCurrentWorld;
		unsigned int mCurrentLevel;
		unsigned int mStartLevel;
		unsigned int mWorldCount;
		CSFont *mFont;
		void initSpeed();
		void setLevelFinished(bool b) {mLevelFinished = true;}
		virtual void addMainSprite(CSSprite* sprite) 
		{
			mMainSprite = sprite;
			addSprite(sprite);
		}
		virtual void addSprite(CSSprite* sprite);
		virtual void removeSprite(CSSprite* sprite);
		virtual void startRun() {};
		virtual void endRun() {};
		virtual void displayGameData() {};
		virtual void checkCollisions();
		virtual void checkOutOfBounds();
		virtual void setDisplayScaled(bool b) {mDisplayScaled = b;}
		virtual bool isDisplayScaled() {return mDisplayScaled;}
		virtual void setScaleDisplayLayer(bool b, int layer);
		void loadLevelData(const std::string &filename, CSLevelData *data);
		void initLevel(std::string levelName);
		void freeLevel();

		virtual void loadGameData(const std::string &filename);
		
		void setCurrentWorld(unsigned int w);
		virtual SDL_Rect getDisplayArea();
		virtual SDL_Rect getScaledDisplayArea();
		virtual double getScaleFactor() {return 0.5;}
		void addSample(CSSample *sample, const char *sid);
		void removeSample(const char *sid);
		void clearSamples();
		void nextLevel();
		void previousLevel();

		void virtual reactOnTimedEvent(TimedEvent *event){}
		void disableSpriteMovement();
		void resetSprites();
		void clearTimedEvents();
		void enableCheck(bool b) {mEnableCheck = b;}

		virtual void loadLevelDataSpecific(CSXMLHelper &xmlSupport){}
		virtual void initLevelSpecific(CSLevelData *data){}
		virtual void freeLevelSpecific(){}
		virtual void resetSpritesSpecific(){}
		bool isSamplePlaying(void);

	public:
		CSGame(std::string gameFileName);
		virtual ~CSGame();

		static const char *CLASS;
		virtual std::string getType() {return (std::string) CLASS;}

		virtual void intro(void);
		virtual void run(void);
		void reactOnMessageGame(CSMessage *message);
		virtual void reactOnMessage(CSMessage *message);
		CSSprites *getSprites() {return &mSprites;}
		CSSprite *getMainSprite() {return mMainSprite;}
		CSWorld *getWorld() {return mCurrentWorld;}
		void checkWorld();

		void enableOutOfBoundsCheck(CSSprite *sprite, CSMessageListener *listener, int threshold);
		void enableTileCollisionCheckJustHit(CSSprite *sprite, int layer, CSMessageListener *listener);
		void enableTileCollisionCheck(CSSprite *sprite, int layer, CSMessageListener *listener);
		void enableTileCollisionCheckPixel(CSSprite *sprite, int layer, CSMessageListener *listener, int pixelThreshold = 1);
		void enableSpriteCollisionCheck(CSSprite *sprite, CSSprite *other, CSMessageListener *listener);
		void enableSpriteCollisionCheckPixel(CSSprite *sprite, CSSprite *other, CSMessageListener *listener, int pixelThreshold = 1);

		void disableOutOfBoundsCheck(CSSprite *sprite, CSMessageListener *listener);
		void disableTileCollisionCheck(CSSprite *sprite, int layer, CSMessageListener *listener);
		void disableSpriteCollisionCheck(CSSprite *sprite, CSSprite *other, CSMessageListener *listener);
		void setThreshold(unsigned int threshold) {mThreshold = threshold;}
		void setStartLevel(unsigned int l) {mStartLevel = l;}
		void playSample(const char *sid);
		void doPause(void);
		void addTimedEvent(unsigned int timeSpace, int id, void *data, int repeat = 1, bool repeatUse = false);
		void addTimedEvent(unsigned int timeSpace, int id, void *data, void (*answer)(TimedEvent *), int repeat = 1, bool repeatUse = false);
		void removeTimedEvent(int id);
		void removeTimedEvent(int id, void *data);
		void enableDesktop(bool b);

		void print(int x, int y, char *text)
		{
			if (mFont == 0)
			{
				printf(text);
			}
			else
			{
				mFont->putString(x, y, text);
			}
		}
		void print(int x, int y, int number)
		{
			char t[40];
			sprintf(t, "%i", number);
			print(x, y, t);
		}

		GameMessage MESSAGE_TOGGLE_FULLSCREEN;
		GameMessage MESSAGE_END;
		GameMessage MESSAGE_PAUSE;
		GameMessage MESSAGE_SPRITE_TILE_COLLISION;
		GameMessage MESSAGE_SPRITE_SPRITE_COLLISION;
		GameMessage MESSAGE_SPRITE_OUT_OF_BOUNDS;
		GameMessage MESSAGE_TOGGLE_GUI;

		ActionMessage MESSAGE_DEATH;
		ActionMessage MESSAGE_UP;
		ActionMessage MESSAGE_DOWN;
		ActionMessage MESSAGE_LEFT;
		ActionMessage MESSAGE_RIGHT;
		ActionMessage MESSAGE_FIRE;
		ActionMessage MESSAGE_DEFAULT;
		
		InputMessage MESSAGE_KEY_UP;
		InputMessage MESSAGE_KEY_DOWN;
		InputMessage MESSAGE_KEY_LEFT;
		InputMessage MESSAGE_KEY_RIGHT;
		InputMessage MESSAGE_KEY_FIRE;

		SpriteMessage MESSAGE_SPRITE_STATE_CHANGE;
};		

#endif // CSGAME_H
