#include "CSHTMLHelper.h"
#include "CSHelper.h"

#ifdef WIN32
#pragma warning(disable : 4786 )
#endif

const char *CSHTMLHelper::CLASS = "CSHTMLHelper";
const char *CSHTMLNode::CLASS = "CSHTMLNode";

/**	Holds the mapping all known entities (not all official html-entities are used!):
	<PRE>
	\verbatim 
		"&nbsp;" <-> " "
		"&quot;" <-> "\""
		"&#60;" <-> "<"
		"&#62;" <-> ">"
		"&lt;" <-> "<"
		"&gt;" <-> ">"
		"&amp;" <-> "&"
		"&copy;" <-> ""
		"&reg;" <-> ""
		"&cent;" <-> ""
		"&deg;" <-> ""
		"&sup2;" <-> ""
		"&raquo;" <-> ""
		"&laquo;" <-> ""
		"&frac14;" <-> ""
		"&frac12;" <-> ""
		"&frac34;" <-> ""
		"&plusmn;" <-> ""
		"&uuml;" <-> ""
		"&auml;" <-> ""
		"&ouml;" <-> ""
		"&Oslash;" <-> ""
		"&iexcl;" <-> ""
		"&ntilde;" <-> ""
		"&szlig;" <-> ""
	\endverbatim.
	</PRE>
*/
HTMLEntityMap CSHTMLNode::mHTMLEntityMap;

/** Return the next sibling-node (brother or sister :-)), if none is available an empty \a CSHTMLNode is returned.
	\return the next sibling-node (if available)
	\return an empty \a CSHTMLNode otherwise
  */
CSHTMLNode CSHTMLNode::nextSibling()
{
	static char *functionName="nextSibling";
	LOG_ENTER 
	if (!mNode) 
	{
		LOG_EXIT
		return CSHTMLNode(0);
	}
	LOG_EXIT
	return CSHTMLNode(mNode->NextSibling());
}

/** Return the previous sibling-node (brother or sister :-)), if none is available an empty \a CSHTMLNode is returned.
	\return the previous sibling-node (if available)
	\return an empty \a CSHTMLNode otherwise

	\todo This is not implemented! (returns allways an empty node!)
  */
CSHTMLNode CSHTMLNode::previousSibling()
{
	static char *functionName="previousSibling";
	LOG_ENTER 
	if (!mNode) 
	{
		LOG_EXIT
		return CSHTMLNode(0);
	}
	LOG_EXIT
	return CSHTMLNode(0);
}

/** Return the parent-node (father or mother :-)), if none is available an empty \a CSHTMLNode is returned.
	\return the parent-node
	\return an empty \a CSHTMLNode otherwise
  */
CSHTMLNode CSHTMLNode::parent()
{
	static char *functionName="parent";
	LOG_ENTER 
	if (!mNode) 
	{
		LOG_EXIT
		return CSHTMLNode(0);
	}
	LOG_EXIT
	return CSHTMLNode(mNode->Parent());
}

/** Return the first child-node (son or daughter :-)), if none is available an empty \a CSHTMLNode is returned.
	\a firstChild() or \a lastChild() must be called in order to
	successfully call \a nextChild() or \a previousChild().

	\return the first child-node
	\return an empty \a CSHTMLNode otherwise
  */
CSHTMLNode CSHTMLNode::firstChild()
{
	static char *functionName="firstChild";
	LOG_ENTER 
	if (!mNode) 
	{
		LOG_EXIT
		return CSHTMLNode(0);
	}
	mCurrentChild = mNode->FirstChild();
	LOG_EXIT
	return CSHTMLNode(mCurrentChild);
}

/** Return the last child-node (son or daughter :-)), if none is available an empty \a CSHTMLNode is returned.
	\a firstChild() or \a lastChild() must be called in order to
	successfully call \a nextChild() or \a previousChild().

	\return the last child-node
	\return an empty \a CSHTMLNode otherwise
  */
CSHTMLNode CSHTMLNode::lastChild()
{
	static char *functionName="lastChild";
	LOG_ENTER 
	if (!mNode) 
	{
		LOG_EXIT
		return CSHTMLNode(0);
	}
	mCurrentChild = mNode->LastChild();
	LOG_EXIT
	return CSHTMLNode(mCurrentChild);
}

/** Return the previous child-node (son or daughter :-)), if none is available an empty \a CSHTMLNode is returned.
	\a firstChild() or \a lastChild() must be called in order to
	successfully call \a nextChild() or \a previousChild().

	\return a child-node
	\return an empty \a CSHTMLNode otherwise
  */
CSHTMLNode CSHTMLNode::previousChild()
{
	static char *functionName="previousChild";
	LOG_ENTER 
	if (mCurrentChild)
	{
		mCurrentChild = mCurrentChild->PreviousSibling();
		if (!mCurrentChild)
		{
			std::string message="No previous child found!\n";
			message += "Tag="+getTag();
			message += ", ";
			message += "Value="+getValue();
			LOG_DEBUG_MESSAGE(message) 
		}
	}
	LOG_EXIT
	return CSHTMLNode(mCurrentChild);
}

/** Return the next child-node (son or daughter :-)), if none is available an empty \a CSHTMLNode is returned.
	\a firstChild() or \a lastChild() must be called in order to
	successfully call \a nextChild() or \a previousChild().

	\return a child-node
	\return an empty \a CSHTMLNode otherwise
  */
CSHTMLNode CSHTMLNode::nextChild()
{
	static char *functionName="nextChild";
	LOG_ENTER 
	if (mCurrentChild)
	{
		mCurrentChild = mCurrentChild->NextSibling();
		if (!mCurrentChild)
		{
			std::string message="No next child found!\n";
			message += "Tag="+getTag();
			message += ", ";
			message += "Value="+getValue();
			LOG_DEBUG_MESSAGE(message) 
		}
	}
	LOG_EXIT
	return CSHTMLNode(mCurrentChild);
}

/** A Text node is a node that does not hold TAG, or attribute information. The plain
    good old html-text!

	\return true if this node countains text 
	\return false if not
  */
bool CSHTMLNode::isText()
{
	return mNode->ToText() != 0;
}

/** If the node is a text-node, the string representing the text is returned.
    Known entities e.g.  "&amp;" <-> "&" are mapped to their "meaning".

	\return text if textnode
	\return an empty string otherwise
  */
std::string CSHTMLNode::getValue()
{
	static char *functionName="getValue";
	LOG_ENTER 
	if (!mNode) 
	{
		LOG_EXIT
		return std::string();
	}
	if (mNode->ToText() != 0)
	{
		LOG_EXIT
		std::string text = mNode->ToText()->Value();
		std::string entityFreeText = replaceHTMLEntities(text);
		return entityFreeText;
	}
	LOG_EXIT
	return std::string();
}

/** Check whether the current node is a single TAG-node (e.g. \<HR\>, \<BR\>).

	\return true if this node is a single TAG-node
	\return false if not (also for text only nodes)
  */
bool CSHTMLNode::isSingleTag()
{
	static char *functionName="isSingleTag";
	LOG_ENTER 
	if (!mNode) 
	{
		LOG_EXIT
		return false;
	}
	if (mNode->ToElement() != 0)
	{
		LOG_EXIT
		return mNode->ToElement()->isSingleTagElement();
	}
	LOG_EXIT
	return false;
}

/** Gets the TAG the current node is representing (e.g. HTML, BODY...)

	\return string that represents the TAG
	\return empty string if not a TAG-node
  */
std::string CSHTMLNode::getTag()
{
	static char *functionName="getTag";
	LOG_ENTER 
	if (!mNode) 
	{
		LOG_EXIT
		return std::string();
	}
	if (mNode->ToElement() != 0)
	{
		LOG_EXIT
		return mNode->ToElement()->Value();
	}
	LOG_EXIT
	return std::string();
}

/** Return the first attribut of the node (e.g. "color", "HREF", ...).
	\a firstAttribut() must be called in order to successfully call \a nextAttribut().

	\return string that represents the attribute
	\return empty string if not a TAG-node or no attribute was found
  */
std::string CSHTMLNode::firstAttribut()
{
	static char *functionName="firstAttribut";
	LOG_ENTER 
	if (!mCurrentElement)
	{
		LOG_EXIT
		return std::string();
	}
	mCurrentAttribute = mCurrentElement->FirstAttribute();
	if (!mCurrentAttribute)
	{
		LOG_EXIT
		return std::string();
	}
	LOG_EXIT
	return mCurrentAttribute->Name();
}

/** Return the next attribut of the node (e.g. "color", "HREF", ...).
	\a firstAttribut() must be called in order to successfully call \a nextAttribut().

	\return string that represents the attribute
	\return empty string if not a TAG-node or no attribute was found
  */
std::string CSHTMLNode::nextAttribut()
{
	static char *functionName="nextAttribut";
	LOG_ENTER 
	if (!mCurrentElement)
	{
		LOG_EXIT
		return std::string();
	}
	if (!mCurrentAttribute)
	{
		LOG_EXIT
		return std::string();
	}
	mCurrentAttribute = mCurrentAttribute->Next();
	if (!mCurrentAttribute)
	{
		LOG_EXIT
		return std::string();
	}
	LOG_EXIT
	return mCurrentAttribute->Name();
}


/** Return the value of an attribut of the node (e.g. for "color" -> "#ffffff").
	Attribute value given is indepenedent of upper or lower case!

	\return string that represents the value of the attribute
	\return empty string if not a TAG-node or no corresponding attribute was found
  */
std::string CSHTMLNode::attributValue(const std::string &attribute)
{
	static char *functionName="attributValue";
	LOG_ENTER 
	const char *p;
	if (!mCurrentElement)
	{
		LOG_EXIT
		return std::string();
	}

	// case independency, make the given attribute name UPPER
	std::string	upperAttribute;
	
	for( p = attribute.c_str(); p < attribute.c_str() + strlen( attribute.c_str() ); p++ )
	{
		upperAttribute += toupper(*p);
	}

	mCurrentAttribute = mCurrentElement->FirstAttribute();
	while (mCurrentAttribute)
	{
		std::string name =	mCurrentAttribute->Name();

		// case independency, make this attribute name UPPER
		std::string upperName;
		for( p = name.c_str(); p < name.c_str() + strlen( name.c_str() ); p++ )
		{
			upperName += toupper(*p);
		}

		// and compare them
		if (strcmp(upperAttribute.c_str(), upperName.c_str()) == 0)
		{
			// equal -> great, return it
			LOG_EXIT
			return mCurrentAttribute->Value();
		}
		mCurrentAttribute = mCurrentAttribute->Next();
	}
	LOG_EXIT
	return std::string();
}
/** this is realy only a quick hack!
	\todo Entities are not all done (at least numbers (e.g "&#55;") should done alright)
*/
std::string CSHTMLNode::replaceHTMLEntities(const std::string &text)
{
	std::string newText = text;
	// first time here -> fill the map!
	if (mHTMLEntityMap.size() == 0)
	{
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&nbsp;", " "));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&quot;", "\""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&#60;", "<"));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&#62;", ">"));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&lt;", "<"));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&gt;", ">"));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&amp;", "&"));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&copy;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&reg;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&cent;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&deg;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&sup2;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&raquo;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&laquo;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&frac14;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&frac12;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&frac34;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&plusmn;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&uuml;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&auml;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&ouml;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&Oslash;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&iexcl;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&ntilde;", ""));
		mHTMLEntityMap.insert(HTMLEntityMap::value_type("&szlig;", ""));

		// this could go endlessly, there are also numerical codes,
		// that one could replace...
	}
	std::string key;
	std::string value;
	HTMLEntityMap::iterator iter = mHTMLEntityMap.begin();
	while (iter != mHTMLEntityMap.end())
	{
		key = iter->first;
		value = iter->second;
		while (CSHelper::replace(newText, key, value))
			;
		iter++; 
	}
	return newText;
}
/******************************************************************************************
*******************************************************************************************/

/** If an error occured while opening, \a getError() returns
    the error code. (This should be checked!)
    \param htmlFilename Must be a valid html file.
    \sa getError()
*/
CSHTMLHelper::CSHTMLHelper(const std::string &htmlFilename)
{
	static char *functionName="CSHTMLHelper";
	LOG_ENTER 
	mError = CSHTML_OK;
	mXMLdoc = 0;
	mXMLdoc = new TiXmlDocument(htmlFilename);
	mXMLdoc->setIgnoreNoEndTag(true);		// for html the xml-Parser must also understand tags
											// which have no endtag (e.g. "<HR>")
	mXMLdoc->SetCondenseWhiteSpace(false);	// important for PRE Tag!
	mXMLdoc->ClearError();
	if (!mXMLdoc->LoadFile() )
	{
		mError = CSHTML_DOCUMENT_LOADING_ERROR;
		std::string errorMessageString;
		errorMessageString = std::string("Error - while loading document \"" + htmlFilename +"\"!\n");
		errorMessageString += mXMLdoc->ErrorDesc();
		mErrorMessageString = mErrorMessageString + "\n"+ errorMessageString;

		LOG_DEBUG_MESSAGE("HTML \"" + errorMessageString) 
		LOG_EXIT
		return;			
	}
	LOG_DEBUG_MESSAGE("HTML loaded!") 
	LOG_EXIT
}

//! Destructor cleans up all internal stuff!
//! \warning As long as the node got via \a getRootNode() is in any way active <B>DON'T</B> delete the \a CSHTMLHelper!!!
//! 
CSHTMLHelper::~CSHTMLHelper()
{
	static char *functionName="~CSHTMLHelper";
	LOG_ENTER 
	if (mXMLdoc)
	{
		delete (mXMLdoc);
		mXMLdoc = 0;
	}
	LOG_EXIT
}

/** Get the root-node of the currently loaded html-file. This will usually be
    the \<HTML\> - tag node!
    \return the \a rootNode 
    \sa CSHTMLNode
	\warning As long as the node got via \a getRootNode() is in any way active <B>DON'T</B> delete the \a CSHTMLHelper!!!
*/
CSHTMLNode CSHTMLHelper::getRootNode()
{
	return CSHTMLNode(mXMLdoc->FirstChildElement());
}

int CSHTMLHelper::getError(void)
{
	return mError;
}

std::string CSHTMLHelper::getErrorMessage(void)
{
	return mErrorMessageString;
}

