#ifndef CSHTMLHelper_H
#define CSHTMLHelper_H

#ifdef WIN32
#pragma warning(disable : 4786 )
#endif

#include "CSLog.h"
#include <map>
#include <string>
//#include <iostream.h>

#include "tinyxml.h"


const int CSHTML_OK = 0;						//!< Return code of \a getError() if everyting was ok.
const int CSHTML_DOCUMENT_LOADING_ERROR = 1;	//!< Return code of \a getError() html-page could not be loaded

typedef std::map<std::string, std::string> HTMLEntityMap;

class CSHTMLHelper;
//! Class that hold all information about a html-node.
/**
	\a CSHTMLNode holds all information of a single html-node and its children.
	(The root-node of a html-page e.g. holds the complete document structur.)
	The root node of a html-page can be got using the class \a CSHTMLHelper

    For an example how \a CSHTMLNode is used see \a CSHTMLHelper.
	\sa CSHTMLHelper

	\warning The corresponding \a CSHTMLHelper must not be deleted, as long as \b any node got from it is in any way active! <B>DON'T</B> delete the \a CSHTMLHelper!!!
*/
class CSHTMLNode
{
	friend CSHTMLHelper;						//!< to call private constructor
	private:
		static HTMLEntityMap mHTMLEntityMap;	//!< holds all "known" entities
		TiXmlNode *mNode;						//!< if empty 0, otherwise the corresponding \a tinyxml node
		TiXmlNode* mCurrentChild;				//!< current child node
		TiXmlElement* mCurrentElement;			//!< current element
		TiXmlAttribute* mCurrentAttribute;		//!< current attribute
		
		//! replace all known entities in the given \a text
		static std::string replaceHTMLEntities(const std::string &text);
	
		//! Constructor for use with \a tinyxml - node.
		/** Only internaly used!
		*/
		CSHTMLNode(TiXmlNode *node) 
		{
			mNode = node;
			mCurrentChild = 0;
			mCurrentAttribute = 0;
			mCurrentElement = mNode->ToElement();
		}

	public:
		//! The unique class identifier (the name of the class).
		static const char *CLASS;

		//! Get the unique class identifier
		virtual std::string getType() {return (std::string) CLASS;}

		//! Builds an empty \a CSHTMLNode
		CSHTMLNode() 
		{
			mNode = 0;
		}
		
		//! Copy constructor
		/** \param node the other \a CSHTMLNode that is copied.
		*/
		CSHTMLNode(const CSHTMLNode &node) 
		{
			// pointers are only copied!
			mNode = node.mNode;
			mCurrentChild = node.mCurrentChild;
			mCurrentElement = node.mCurrentElement;
			mCurrentAttribute = node.mCurrentAttribute;
		}

		//! Destructor
		virtual ~CSHTMLNode() {}	// pointers are never freed either!

		//! Get the next sibling node 
		CSHTMLNode nextSibling();

		//! Get the previous sibling node 
		CSHTMLNode previousSibling();	

		//! Get the parent node 
		CSHTMLNode parent();		

		//! Get the first child node 
		CSHTMLNode firstChild();	

		//! Get the last child node 
		CSHTMLNode lastChild();		

		//! Get the next child node 
		CSHTMLNode nextChild();		

		//! Get the previous child node 
		CSHTMLNode previousChild();

		//! Whether the node actually exits! 
		/** All methods return a \a CSHTMLNode, if an error occured, or
			a node does not exist, an empty node is returned. An empty
			node can be checked with this method.
			\return \a true on an empty node
			\return \a false on correct node
		*/
		bool isEmpty() {return (mNode==0);}

		//! Is the this node a "text" node?
		bool isText();

		//! Get the value of the node.
		std::string getValue();

		//! Is this node a single TAG node?
		bool isSingleTag();

		//! Get TAG (name)
		std::string getTag();

		//! Get first attribute (name)
		std::string firstAttribut();

		//! Get next attribute (name)
		std::string nextAttribut();

		//! Get value of attribte (with name as in "attribut")
		std::string attributValue(const std::string &attribut);
};

//! Class to load a html-page, and give access to it's root node.
/**
	A html-page can only be loaded from a local filesystem, not via an URL
	from the internet. This class is responsible for loading the page and giving
	access to it by calling the \a getRootNode() method. This method returns an \a CSHTMLNode,
	with which one can navigate thru the html-page and it's tags.
    \sa CSHTMLNode

	To parse a HTML-file use e.g. the following code:
	<PRE>
  	
	CSHTMLHelper html("index.html");
	CSHTMLNode rootNode = html.getRootNode();

	if (!rootNode.isEmpty())
	{
		printf("Root Node Value: %s\n", rootNode.getValue().c_str());
		printf("Root Node TAG: %s\n", rootNode.getTag().c_str());
		printf("Tree follows...\n");
		printChildren(rootNode);
	}
	</PRE>
    with the helper function as follows:
	<PRE>
	void indent(int ind)
	{
		for (int i=0; i<ind;i++) printf(" "); 
	}

	void printChildren(CSHTMLNode &rootNode)
	{
		static int ind = 0;
		ind++;
		CSHTMLNode node = rootNode.firstChild();
		while (!node.isEmpty())
		{
			std::string value = node.getValue();
			std::string tag = node.getTag();
			if (tag.size())
			{
				indent(ind); 
				printf("<%s", tag.c_str());
				
				std::string attribute = node.firstAttribut();
				while (attribute.size())
				{
					std::string avalue = node.attributValue(attribute);
					printf(" %s=", attribute.c_str());
					printf("\"%s\"", avalue.c_str());
					attribute = node.nextAttribut();
				}
				printf(">\n");
			}
			if (value.size())
			{
				indent(ind); 
				printf("%s\n", value.c_str());
			}
			printChildren(node);
			if (!node.isSingleTag())
			{
				if (tag.size())
				{
					indent(ind); 
					printf("</%s>\n", tag.c_str());
				}
			}
			node = rootNode.nextChild();
		}
		ind--;
	}
	</PRE>

	\warning As long as the node got via \a getRootNode() is in any way active <B>DON'T</B> delete the \a CSHTMLHelper!!!
*/

class CSHTMLHelper
{
	private:
		std::string mErrorMessageString;
		int	mError;
		TiXmlDocument	*mXMLdoc;
		
	public:
		//! The unique class identifier (the name of the class).
		static const char *CLASS;

		//! Get the unique class identifier
		virtual std::string getType() {return (std::string) CLASS;}

		//! Constructor of CSHTMLHelper
		CSHTMLHelper(const std::string &htmlFilename);

		//! Destructor CSHTMLHelper
		virtual ~CSHTMLHelper(); 

		//! Get the first TAG found in the loaded htmlfile (the root-node).
		CSHTMLNode getRootNode();

		//! Get the current error code.
		int getError(void);
	
		//! Textual representation of an error.
		std::string getErrorMessage(void);
};

#endif CSHTMLHelper_H