#ifdef WIN32
#pragma warning(disable : 4786 )
#endif

#ifndef CSLOADABLE_H
#define CSLOADABLE_H

#include <string>
#include <map>
//#include "CSUtility.h"
#include "CSLog.h"

/**
	A template class for loadable objects. These might be any
	kind of data loaded from somewhere.

    This class encapsulates the functionality of loading
	a needed resource only once. Upon a second call of the loader
	it gives back an allready loaded resource.

    A MUST for this to work is that this class is also responsible for
	freeing the (only) once loaded resources.

	If a resource is a
		virtual bool isSingleCreate(void) const;
	the resource will be put in a map, where it can be found later on for reuse.
	(map key is the filename of the resource)
	
	The object, that is created using this class (via a filename) must
	provide a 
		virtual T *create(const std::string &filename) = 0;
	-method.

	Per default an object is destroyed using c++ "delete".
	But the destroy-method can be overloaded by children!

  */

template <class T>
class Loadable
{
	private:
		std::map<std::string, T *> mMap;
	protected:
		virtual std::map<std::string, T *> *getMap() {return &mMap;}
		virtual void destroy(T *object);
		virtual bool isSingleCreate(void) const;
		virtual T *create(const std::string &filename) = 0;
		Loadable()
		{
			mMap.clear();
		}

	public:

		virtual T *load(const std::string &filename);

		virtual void unload(T *image);
		virtual void quit();	//!< unloads all knonw (singleton) 
								//!< instances of this loader class

		virtual ~Loadable()
		{
			quit();
		}
};

//! only produce singletons (and add to mapping)
//! if "isSingleCreate()" return true (default)
template <class T>
T *Loadable<T>::load(const std::string &filename)
{
	if (isSingleCreate())
	{
		LOG_MESSAGE((LOG_LEVEL_DEBUG),"Getting file from map: " + filename)
		std::map<std::string, T *>::iterator iter = mMap.find(filename.c_str());
		if (iter != mMap.end())
		{
			return iter->second;
		}
	}
	
	LOG_MESSAGE((LOG_LEVEL_DEBUG),"Loading file from disk: " + filename)
    T *anObject = create(filename.c_str());
	if (isSingleCreate())
	{
		mMap.insert(std::map<std::string, T *>::value_type(filename.c_str(), anObject));
	}
	return anObject;
}


template <class T>
void Loadable<T>::unload(T *anObject)
{
	//! delete single object
	std::map<std::string, T *>::iterator iter;
	for (iter = mMap.begin(); iter!=mMap.end(); iter++)
	{
		if (iter->second == anObject)
		{
			destroy(iter->second);
			mMap.erase(iter);
			return;
		}
	}
	SDLMain::shutdown((std::string)"Couldn't properly delete map object." + SDL_GetError(), 2);
}

template <class T>
void Loadable<T>::quit()
{
	// delete all object
	std::map<std::string, T *>::iterator iter;
	for (iter = mMap.begin(); iter!=mMap.end(); iter = mMap.begin())
	{
		unload(iter->second);
	}
	mMap.clear();
}

template <class T>
void Loadable<T>::destroy(T *object)
{
	delete object;
}

template <class T>
bool Loadable<T>::isSingleCreate(void) const
{
	return true;
}

#endif //CSLOADABLE_H
