#ifndef CS_MESSAGE_H
#define CS_MESSAGE_H

#include <vector>

class CSMessageDispatchable;

class CSTile;
class CSSprite;

class CSMessage;
class ActionMessage;
class GameMessage;
class SDLMessage;
class InputMessage;
class SpriteMessage;
class GuiMessage;

//! Message Types
const int ACTION_MESSAGE = 1;
const int GAME_MESSAGE = 2;
const int SDL_MESSAGE = 3;
const int INPUT_MESSAGE = 4;
const int SPRITE_MESSAGE = 5;
const int GUI_MESSAGE = 6;

//! Message Sub-Types (SDL)
const int SCREEN_CHANGED_MESSAGE = 1;

//! Message Sub-Types (Game)
const int END_MESSAGE = 1;
const int PAUSE_MESSAGE = 2;
const int TOGGLE_FULLSCREEN_MESSAGE = 3;
const int SPRITE_TILE_COLLISION_MESSAGE = 4;
const int SPRITE_SPRITE_COLLISION_MESSAGE = 5;
const int SPRITE_OUT_OF_BOUNDS_MESSAGE = 6;
const int TOGGLE_GUI_MESSAGE = 7;
const int LAST_GAME_MESSAGE = 10;

//! Message Sub-Types (Action)
//! following must correspond to type definition
//! in action file!
const int DEFAULT_MESSAGE = 0;
const int UP_MESSAGE = 1;
const int DOWN_MESSAGE = 2;
const int RIGHT_MESSAGE = 3;
const int LEFT_MESSAGE = 4;
const int FIRE_MESSAGE = 5;
const int DEATH_MESSAGE = 6;

//! Message Sub-Types (INPUT)
const int KEY_UP_MESSAGE = 1;
const int KEY_DOWN_MESSAGE = 2;
const int KEY_RIGHT_MESSAGE = 3;
const int KEY_LEFT_MESSAGE = 4;
const int KEY_FIRE_MESSAGE = 5;

//! Message SubSubType SPRITE_OUT_OF_BOUNDS_MESSAGE
const int X_LOWER = 0;
const int X_HIGHER = 1;
const int Y_LOWER = 2;
const int Y_HIGHER = 3;

//! Message Sub-Types (Sprite)
const int SPRITE_STATE_CHANGE_MESSAGE = 1;
const int SPRITE_ACTION_CHANGE_MESSAGE = 2;

//! Message Sub-Types (GUI_MESSAGE)
const int MOUSE_MOTION_MESSAGE = 1;
const int MOUSE_MOTION_LOST_MESSAGE = 2;
const int MOUSE_BUTTON_PRESSED_MESSAGE = 3;
const int MOUSE_BUTTON_RELEASED_MESSAGE = 4;

const int KEY_PRESSED_MESSAGE = 5;
const int KEY_RELEASED_MESSAGE = 6;

const int FOCUS_LOST_MESSAGE = 7;
const int FOCUS_GAINED_MESSAGE = 8;
const int WINDOW_OPEN_MESSAGE = 9;
const int WINDOW_CLOSE_MESSAGE = 10;

const int BUTTON_PRESSED_MESSAGE = 11;
const int BUTTON_RELEASED_MESSAGE = 12;
const int BUTTON_DRAGGED_MESSAGE = 13;
const int CHECKBOX_CHANGED_MESSAGE = 14;

const int SCROLLBAR_POSITION_CHANGED_MESSAGE = 15;
const int LIST_ITEM_SELECTED_MESSAGE = 16;
const int COMBOBOX_ITEM_SELECTED_MESSAGE = 17;

const int MENUITEM_ACTIVATED_MESSAGE = 18;
const int TEXT_CHANGED_MESSAGE = 19;

const int CLOSE_ACTION_ID = 0x1000;
const int SCROLL_BUTTON_LESS_ACTION_ID = 0x1100;
const int SCROLL_BUTTON_MORE_ACTION_ID = 0x1101;
const int SCROLL_BUTTON_DRAG_ACTION_ID = 0x1102;

const int COMBOBOX_BUTTON_MORE_ACTION_ID = 0x1200;

/**
	The CSMessage class is the base class for messages.
	It might have been possible to use this class as an
	overall massage system. but due to non inheritable 
	friend relations and data consistency 
	(most set methods should only be accessable from the
	message initiator) the actual use of the class is like:

    (Example from CSSprite)
	<PRE>
	class CSSprite : ...
	{
		...
		public:
			class SpriteMessage: public CSMessage
			{
				friend CSSprite;
			};
			SpriteMessage MESSAGE_SPRITE_NEXT_DEFAULT;
			...
	};
	</PRE>
    In the constructor of the message using class fixed messages
	might be pre-defined.
	That also meens, that messages are owned by the message - Source.
	And should never be freed from any place else!

	######################

	Also message types are defined here!
	(for now)
*/
class CSMessage
{
	friend CSMessageDispatchable;
	protected:

		CSMessageDispatchable *mOrigin;
		int mType;
		int mSubtype;
		int mSubsubtype;
		int mPriority;
		bool mIsDispatchAtOnce;
		bool mSuccess;

		virtual void setType(int type){mType = type;}

	public:
		bool mIsHandled;
		CSMessage() 
		{
			mIsHandled = false;
			mSuccess = false;
			mOrigin = 0;
			mSubtype = 0;
			mSubsubtype = 0;
			mType = 0;
			mPriority = 0;
			mIsDispatchAtOnce = true;
		}
		
		virtual ~CSMessage() {}

		virtual void setSuccess(bool b) {mSuccess = b;}
		virtual void setOrigin(CSMessageDispatchable *origin) {mOrigin = origin;}
		virtual void setSubtype(int subtype){mSubtype = subtype;}
		virtual void setSubsubtype(int type){mSubsubtype = type;}
		virtual void setPriority(int priority){mPriority = priority;}
		virtual void setDispatchAtOnce(bool dispatchAtOnce){mIsDispatchAtOnce = dispatchAtOnce;}

		virtual bool isSuccess() {return mSuccess;}
		virtual CSMessageDispatchable *getOrigin(){return mOrigin;}
		virtual int getSubtype(){return mSubtype;}
		virtual int getSubsubtype(){return mSubsubtype;}
		virtual int getType(){return mType;}
		virtual int getPriority(){return mPriority;}
		virtual bool isDispatchAtOnce(){return mIsDispatchAtOnce;}
		virtual void setHandled(bool isHandled){mIsHandled = isHandled;}
		bool isHandled(){return mIsHandled;}
};

class ActionMessage: public CSMessage
{
	public:
		ActionMessage() {mType = ACTION_MESSAGE;}
};

class GameMessage: public CSMessage
{
	public:
		CSSprite *sprite;
		CSSprite *other;
		std::vector<CSTile*> tiles;
		CSTile* tile;
		int pixels;
		int layer;

		GameMessage() 
		{
			mType = GAME_MESSAGE;
			sprite = 0;
			other = 0;
			pixels = 0;
			tile = 0;
			layer = -1;
		}
};

class SDLMessage: public CSMessage
{
	public:
		SDLMessage() {mType = SDL_MESSAGE;} 
};

class SpriteMessage: public CSMessage
{
	public:
		int state;
		SpriteMessage() {mType = SPRITE_MESSAGE;} 
};

class InputMessage: public CSMessage
{
	public:
		InputMessage() {mType = INPUT_MESSAGE;} 
};

class CSGrafikElement;
class GuiMessage: public CSMessage
{
	public:
		int selected;
		int start;
		int end;
		int screenX;
		int screenY;
		int receiverX;
		int receiverY;
		int deltaX;
		int deltaY;
		int button;
		int actionId;
		int keySymbolic;
		int keyUnicodeChar;
		bool newCheckBoxState;
		CSGrafikElement *sender;
		CSGrafikElement *receiver;
		CSGrafikElement *downReceiver;
		GuiMessage() 
		{
			mType = GUI_MESSAGE;
			selected = 0;
			start = 0;
			end = 0;
			screenX = 0;
			screenY = 0;
			receiverX = 0;
			receiverY = 0;
			deltaX = 0;
			deltaY = 0;
			button = 0;
			actionId = 0;
			keySymbolic = 0;
			keyUnicodeChar = 0;
			newCheckBoxState = false;
			sender = 0;
			receiver = 0;
			downReceiver = 0;
		} 
};

#endif // CS_MESSAGE_H


