#ifndef CSPicture_h
#define CSPicture_h

// USES SDL_Surface
// USES SDL_Rect 

#include <string>
#include <vector>
#include "SDLMain.h"
#include "CSLoadable.h"
#include "CSLog.h"

#ifdef WIN32
#pragma warning(disable : 4786 )
#endif

class CSBitmap;
class CSPicture;
class CSPictureLoader;
class CSPictureDirectLoader;

typedef std::vector<CSPicture *> CSPictures;

/** Class that holds DisplayValues, that are used by each CSPicture.
  * Thru this class (structure)
  * we have some sort of abstraction where we put something
  * on the screen - or on some world coordinates
  */
class CSDisplayParams
{
	public:
		int mXDisplayStart;	///< X startoffset of surface destination (0,0 of the blitting)
		int mYDisplayStart; ///< Y startoffset of surface destination (0,0 of the blitting)
		int mXWorldStart;	///< X where the display of the "world" starts
		int mYWorldStart;	///< Y where the display of the "world" starts
		int mXPos;			///< X actual position of this picture in the world
		int mYPos;			///< Y actual position of this picture in the world
};

const int RLE_YES = 1;
const int RLE_NO = 0;
const int RLE_PERHAPS = 2; // other

/** Data class that is used for loading a loadable class.
  * First the data is loaded via xml into this "DATA"-class.
  * From this data - the actual picture is created.
  * The data can be reused to create another instance of a
  * picture class - though probably not needed...
  */
class CSPictureData
{

	public:
		SDL_Rect position;	///< has position and size
		char *id;
		char *filename;		///< of bitmap file
		int rleUse;
		int alpha;

		CSPictureData()
		{
			rleUse = RLE_PERHAPS;
			id = 0;
			filename = 0;
			position.x=0;
			position.y=0;
			position.w=0;
			position.h=0;
			alpha = -1;
		}

		~CSPictureData()
		{
			if (id != 0) 
			{
				free(id);
				id = 0;
			}
			if (filename != 0) 
			{
				free(filename);
				filename = 0;
			}
		}
};
/** 
	CSPicture
	  CSPicture can only be created thru a loader (CSPictureLoader).
	  CSPicture are defined by (now ONLY thru) an xml-File, like:
		DTD:
\verbatim
			<!ELEMENT PICTURE (ID,(X,Y,WIDTH,HEIGHT)?,BITMAP)>
			<!ELEMENT ID (#PCDATA)>
			<!ELEMENT X (#PCDATA)>
			<!ELEMENT Y (#PCDATA)>
			<!ELEMENT WIDTH (#PCDATA)>
			<!ELEMENT HEIGHT (#PCDATA)>
			<!ELEMENT BITMAP (#PCDATA)>
\endverbatim		
		  Example:
\verbatim
			<PICTURE>
				<ID>PacManUp1</ID>
				<X>0</X>
				<Y>96</Y>
				<WIDTH>37</WIDTH>
				<HEIGHT>31</HEIGHT>
				<BITMAP>bitmap\\pac_pix.xml</BITMAP>
			</PICTURE>
\endverbatim	  
<PRE>
	  ID	           - supposed a unique identifier
	  X, Y			   - position of "picture" on the bitmap
	  WIDTH, HEIGHT    - Width and Hight (size) of the "picture" on the bitmap
	  BITMAP		   - File to be loaded for the bitmap.

	  CSPicture are singletons in the sense, that two "loaded" CSPicture of the 
	  same "ini" file ARE the same instances. 
	  (Bitmaps are loaded only once - by there respective loader entity)

	  these methods can be used:

		void display(CSDisplayParams *displayParams)

		unsigned int getMaxX()		// get width of the pciture
		unsigned int getMaxY()		// get height of the picture
		void setSolid(bool b)		// enable/disable transparency
		bool isSolid()				// has picture a transparent color
		SDL_Surface *getSurface() 	// for the font lib only

	  (SDLMain::addUpdateRect(mUpdateRegion) - is called after displaying, so at some 
	   stage - later (main-loop)
	   		void SDLMain::updateScreen()
	   must be called.)
</PRE>

	Class that holds and can display a picture.
	The picture class has no instance relevant data (state independant)
	for example: as position \par
	Thus one picture can be (re) used severall times!
	The data used for printing must come from the outside (CSDisplayParams).
*/
class CSPicture
{
	friend CSPictureLoader;
	friend CSPictureDirectLoader;
	private:
		static CSPictures mAllPictures;
		static void addPicture(CSPicture *picture);
		static void removePicture(CSPicture *picture);

		bool mIsDirectLoad;		
		std::string mLoadName;
		int mRLEUse;				///< flag whether RLE should be used by this image
									///< set via xml ini
		int mAlpha;					///< the alpha value of this picture
									///< can be set via xml-ini

		SDL_Rect mUpdateRegion;		///< has position and size where the picture will
									///< be displayed
									///< is set to mPixelPictureRegion.w
									///< and mPixelPictureRegion.h
		SDL_Rect mPixelPictureRegion; ///< the picture within the image ehich is loaded from disk
									///< will be blittet to a surface, where it alone resides
									///< this are the coordinates for that
									///< note: due to optimization, this region might also
									///< be smaller than the actual picture size...
									///< if the picture is "outlined" by its transparent
									///< color, than the x,y,w,h will be updated
									///< to where not only transaprent colors are visible
		SDL_Rect mPictureRegion;	///< position and size of this picture on a 
									///< "greater" image
									///< - does not change, once set!

		unsigned int mStartPixelsX;	///< like mPixelPictureRegion.x
									///< gives the offset from the original
									///< picture start to where the optimized
									///< picture starts
		unsigned int mStartPixelsY;	///< like mPixelPictureRegion.y
									///< gives the offset from the original
									///< picture start to where the optimized
									///< picture starts

		SDL_Surface *mPicture;		///< never! freed - is handled by BitmapLoader!
		bool mIsSolid;				///< picture has now transparency
		char *mId;					///< id of picture
		char *mFileName;			///< filename, from which picture was loaded (xml)
		int mColorKey;				///< current color key
		int mR;
		int mG;
		int mB;
		double mFactor;
		int mScaledWidth;
		int mScaledHeight;
		bool mIsCreatedAsScaled;
		SDL_Surface *mScaledPicture;///< 

		unsigned char **mProjection;///< pixel projection of image (8bit) for easier
									///< collision detection
		unsigned int mPixelProjectionCount; ///< how many non- transparent pixels
									///< are there within this image
		int mRLEAccelleration;		///< "OR" whether RLE acceleration is
									///< used when image is blitted
		
		///  constructor using "*.xml" file
		CSPicture(const std::string &filename, bool isDirectLoad = false);
		virtual void initialize(const CSPictureData &data, bool isDirectLoad = false);
		static void loadPictureData(const std::string &filename, CSPictureData &data);
		
		void resetSurface();		///< converts current surface to display surface
									///< used (e.g.) upon fullscreen-toggle
		SDL_Surface *convertToCurrentScreen(CSBitmap *image, SDL_Rect rect);
		void checkRLEAccelleration(); ///< this is not really well implemented yet :-)
		void buildPixelProjection(CSBitmap *image, const SDL_Rect &position);
		void buildOptimizeData();

	public:
		/// copy constructor
		CSPicture(const CSPicture &picture);

		/// destructor
		virtual ~CSPicture();	

		static const char *CLASS;
		virtual std::string getType() {return (std::string) CLASS;}

		void getRGB(int &r, int &b, int &g);
		virtual void display(CSDisplayParams *mDisplayParams);	///< display this picture on the srceen
		virtual void display(int x, int y);						///< display this picture on the srceen
		virtual void display(SDL_Surface *destination, int x, int y);						///< display this picture on the srceen
		unsigned int getMaxX(){return mPictureRegion.w;}		///< get width of the pciture
		unsigned int getMaxY(){return mPictureRegion.h;}		///< get height of the picture
		unsigned int getWidth(){return mPictureRegion.w;}		///< get width of the pciture
		unsigned int getHeight(){return mPictureRegion.h;}		///< get height of the picture
		void setSolid(bool b);									///< enable/disable transparency
		bool isSolid() {return mIsSolid;}						///< has picture a transparent color
		void setAlpha(int alpha);
		int getColorKey() {return mColorKey;}					///< get current transparent color
		SDL_Surface *getSurface() {return mPicture;}			///< for the font lib only
		unsigned char **getProjection() {return mProjection;}	///< for collision detection
		unsigned int getPixelProjectionCount()					///< for collision detection (optimization)
		{return mPixelProjectionCount;}
		
		static void resetSurfaces();							///< resets all surfaces
																///< loaded via the CSPictureLoader (all!)
		void scale(int width, int divideX, int height, int divideY);
		SDL_Surface *getScaledPicture(double factor);
		SDL_Surface *getScaledPicture(int w, int h);
		bool isTransparent(int x, int y);
		static CSPicture* createScaledPicture(CSPicture* picture, double factor);	// builds a complete new Picture!
};

/** Factory Loader-class for CSPictures.
  * Factory Loader-class for CSPictures
  * that is used to create CSPicture \par
  * Uses xml - data-files, produces "singletons"
  */
class CSPictureLoader : public Loadable<CSPicture>
{
	protected:
		virtual CSPicture *create(const std::string &filename)
		{
			return new CSPicture(filename, false);
		}

	public:
		static CSPictureLoader INSTANCE;
		void scale(char *name, int width, int divideX, int height, int divideY);
};

/** Factory Loader-class for CSPictures.
  * Factory Loader-class for CSPictures
  * that is used to create CSPicture \par
  * Uses xml - data-files, produces "singletons"
  */
class CSPictureDirectLoader : public Loadable<CSPicture>
{
	protected:
		virtual CSPicture *create(const std::string &filename)
		{
			return new CSPicture(filename, true);
		}

	public:
		static CSPictureDirectLoader INSTANCE;
		void scale(const char *name, int width, int divideX, int height, int divideY);
};
#endif // CSPicture_h
