#ifndef CSSPRITE_H
#define CSSPRITE_H

// USES SDL_Surface

#include <string>
#include <vector>
#include <map>
#include "CSTypes.h"
#include "CSLoadable.h"
#include "CSMessageListener.h"
#include "CSMessage.h"
#include "CSAction.h"		//!< for ActionState
#include "CSLog.h"

class CSTileMap;
class CSTile;
class CSNavigator;

class CSSpriteLoader;
class CSSprite;
class StateData;
class ActionBundle;

typedef std::vector<CSSprite*> CSSprites;
typedef std::vector<StateData *> StateDatas;
typedef std::map<int, ActionBundle *> ActionBundleMap;

class StateData
{
public:
		StringVector actionNames;			//!< filenames (xml) of all action belonging to sprite
		char *defaultActionName;			//!< default action (id)
		int stateId;
		
		StateData()
		{
			defaultActionName = 0;
			stateId = 0;
		}

		~StateData()
		{
			for (StringVector::iterator iter = actionNames.begin(); iter != actionNames.end(); iter++)
			{
				delete *iter;
			}
			actionNames.clear();

			if (defaultActionName != 0)
			{
				free(defaultActionName);
				defaultActionName = 0;
			}
			stateId = 0;
		}
};

//! data classes are used for loading a loadable class
//! first the data is loaded via xml into this "DATA"-class
//! from this data - the actual sprite is created
//! the data can be reused to create another instance of an
//! sprite class - though probably not needed...
class CSSpriteData
{
	public:
		char *id;							//!< id of sprite (unique)
		StateDatas stateDatas;
		int layer_position;					//!< sprite lies above layer (and under each higher layer)
		int type;
		int defaultBundle;
		int subtype;
		int y_position;						//!< within that layer a sprite can have a y position
											//!< so that one sprite can be above another within one layer
											//!< TODO: right now the squence of the sprite
											//!< is used, this parameter is still ignored!
		CSSpriteData()
		{
			type = 0;
			subtype = 0;
			defaultBundle = 0;
			id = 0;
			layer_position = 0;
			y_position = 0;
		}

		~CSSpriteData()
		{
			if (id != 0)
			{
				free (id);
				id = 0;
			}
			for (StateDatas::iterator iter = stateDatas.begin(); iter != stateDatas.end(); iter++)
			{
				delete *iter;
			}
			stateDatas.clear();
		}
};

//! the states hold all variable parts needed for printing the 
//! sprite on screen
//! putting these in a state class, enables us to
//! reuse the memory expensive sprite class!
class SpriteState
{
	public:
		ActionState mActionState;		//!< each sprite must have actions -> so each
										//!< sprite state needs an action state!

		//!< current active action
		CSAction *mCurrentAction;
		ActionBundle *mCurrentBundle;
		int mCurrentBundleId;
		int mUserState;

		SpriteState() 
		{
			mUserState = 0;
			mCurrentAction = 0;
			mCurrentBundleId = 0;
			mCurrentBundle = 0;
		}
		virtual ~SpriteState() {}
};

class ActionBundle
{
	public:
		CSActions *mActions;			//!< all actions belonging to this sprite
		CSAction *mDefaultAction;		//!< default action
		int stateId;

		ActionBundle()
		{
			mActions = 0;			//!< all actions belonging to this sprite
			mDefaultAction = 0;		//!< default action
			stateId = 0;
		}
		
		~ActionBundle()
		{
			if (mActions != 0)
			{
				//!< Actions are singletons -> do not delete them 
				delete mActions;
				mActions = 0;
			}
			mDefaultAction = 0;
		}
};

/**
	This file defines CSSprite:

	  CSSprite can only be created thru a loader (CSSpriteLoader).
	  CSSprite are defined by (now ONLY thru) an xml-File, like:
	  DTD:
\verbatim
		<!ELEMENT SPRITE (ID,(ACTION)+,DEFAULT_ACTION,LAYER_POSITION,Y_POSITION)>
		<!ELEMENT ID (#PCDATA)>
		<!ELEMENT ACTION (#PCDATA)>
		<!ELEMENT DEFAULT_ACTION (#PCDATA)>
		<!ELEMENT LAYER_POSITION (#PCDATA)>
		<!ELEMENT Y_POSITION (#PCDATA)>
\endverbatim
	  Example:

\verbatim
		  <SPRITE>
			<ID>PACMAN_SPRITE</ID>
			<ACTION>action\\pac_down.xml</ACTION>
			<ACTION>action\\pac_up.xml</ACTION>
			<ACTION>action\\pac_left.xml</ACTION>
			<ACTION>action\\pac_right.xml</ACTION>
			<ACTION>action\\pac_down_stay.xml</ACTION>
			<ACTION>action\\pac_up_stay.xml</ACTION>
			<ACTION>action\\pac_right_stay.xml</ACTION>
			<ACTION>action\\pac_left_stay.xml</ACTION>
			<ACTION>action\\pac_death.xml</ACTION>
			<DEFAULT_ACTION>action\\pac_right.xml</DEFAULT_ACTION>
			<LAYER_POSITION>0</LAYER_POSITION>
			<Y_POSITION>0</Y_POSITION>
		</SPRITE>
\endverbatim
<PRE>
	  ID	       - supposed a unique identifier

	  ACTION       - many actions are possible!
					 describes the action to be loaded,
	                 parameter gives the ini-file for a CSAction

  	  DEFAULT_ACTION - an action that is used as a default, if another actions
					 ends.
					 This default is overriden by the DEFAULT_NEXT_ACTION
					 setting of the current active action.
					 The default action should not appear listed in the 
					 'normal' action setting. (no double entries)

	  LAYER_POSITION - sprite lies above layer (and under each higher layer)
	  Y_POSITION	- within that layer a sprite can have a y position
					  so that one sprite can be above another within one layer
					  TODO: right now the squence of the sprite
			          is used, this parameter is still ignored!

	  Methods:
		virtual void next();
		virtual void display();
		SpriteState buildState();						// only for compatibility issues
		void resetState(SpriteState &state);
		virtual void reactOnMessage(CSMessage *message);// MessageListener implementation
		void setSecPerFrame(float secPerFrame);
		void setDisplayOffset(int xOffset, int yOffset);// where on the screen is 0,0 of our displaying	
		void setWorldPosition(int x, int y);			// where in the world will we start drawing the screen
		unsigned int checkCollisionPixels(CSSprite *other);	// Collision detection, returns # of pixels
		unsigned int checkCollisionPixels(CSTileMap *map);  // Collision detection, returns # of pixels
		unsigned int checkCollisionPixels(CSTile *tile);    // Collision detection, returns # of pixels
		bool checkCollision(CSSprite *other);			// Collision detection, returns boolean
		bool checkCollision(CSTileMap *map);			// Collision detection, returns boolean
		bool checkCollision(CSTile *tile);				// Collision detection, returns boolean
		std::vector<CSTile*> getCollidingTiles(CSTileMap *map); // return all tiles, that have pixels colliding with sprite
		void setSpeedY(int speed);						// overrides values in animation
		void setSpeedX(int speed);						// overrides values in animation
		void resetSpeed();								// returns to speed values of animation (default)
		SpriteState getState() 
		void setState(const SpriteState &state) 
		int getLayer() 
		int getLayerY() 
		void clearNavigator() 
		void setNavigator(CSNavigator *navigator)
		unsigned int getDisplayXLower()		
		unsigned int getDisplayYLower()
		unsigned int getDisplayXUpper() // pos + width of sprite
		unsigned int getDisplayYUpper() // pos + height of sprite
		unsigned int getXPos()	
		unsigned int getYPos()

	  Supported messages: 
	  (see Message.h)
	  Messages of type; ACTION_MESSAGE

	  CSSprite also implements:
			CSMessageListener
</PRE>
	the sprite class has (in opposite to action an aniamtion) state 
	dependent data -> the sprite state
	should at some point it be neccessary to have state information outside of
	sprites - than we must remove it here, than the sprite class will
	(as action and aniamtion) behave as a local singleton... (and change next(), display())
	*/
class CSSprite : public CSMessageListener, public CSMessageDispatchable
{
	friend CSSpriteLoader;
	private:
		SpriteMessage MESSAGE_SPRITE_ACTION_CHANGE;
		SpriteState mState;				//!< the only state dependent data within this class

		bool mActive;
		ActionBundleMap actionMap;
		int mDefaultBundle;
		CSNavigator *mNavigator;		//!< this class can 'navigate' our sprite (if set)
										//!< called within (at first chance) next()
		int mType;
		int mSubtype;
		int layer_position;				//!< sprite lies above layer (and under each higher layer)
		int y_position;					//!< within that layer a sprite can have a y position
										//!< so that one sprite can be above another within one layer
										//!< TODO: right now the squence of the sprite
										//!< is used, this parameter is still ignored!

		unsigned int mWidth;			//!< max Width of this sprite
		unsigned int mHeight;			//!< max Height of this sprite
		unsigned int mWorldPos;			//!< current on world # 		
		std::string mFilename;
		char *mId;
		//!< constructor using "*.xml" file
		CSSprite(const std::string &filename);
		static void loadSpriteData(const std::string &filename, CSSpriteData &data);
		bool changeStateTo(int state);

		void initialize(const CSSpriteData &data);
		void initialize();
	
	public:
		CSSprite(const CSSprite &Sprite);
		virtual ~CSSprite();

		static const char *CLASS;
		virtual std::string getType() {return (std::string) CLASS;}

		virtual void next();
		virtual void display();

		SpriteState buildState();						//!< only for compatibility issues
		void resetState(SpriteState &state);
		
		virtual void reactOnMessage(CSMessage *message);//!< MessageListener implementation

		void setSecPerFrame(float secPerFrame);
		
		void setDisplayOffset(int xOffset, int yOffset);//!< where on the screen is 0,0 of our displaying	
		void setWorldPosition(int x, int y);			//!< where in the world will we start drawing the screen
		unsigned int checkCollisionPixels(CSSprite *other);	//!< Collision detection, returns # of pixels
		unsigned int checkCollisionPixels(CSTileMap *map);  //!< Collision detection, returns # of pixels
		unsigned int checkCollisionPixels(CSTile *tile);    //!< Collision detection, returns # of pixels
		bool checkCollision(CSSprite *other);			//!< Collision detection, returns boolean
		bool checkCollision(CSTileMap *map);			//!< Collision detection, returns boolean
		bool checkCollision(CSTile *tile);				//!< Collision detection, returns boolean
		std::vector<CSTile*> getCollidingTiles(CSTileMap *map); //!< return all tiles, that have pixels colliding with sprite
		void setSpeedY(int speed);						//!< overrides values in animation
		void setSpeedX(int speed);						//!< overrides values in animation
		void adjustSpeed(int speed);					//!< overrides values in animation
		void resetSpeed();								//!< returns to speed values of animation (default)
		void setNavigator(CSNavigator *navigator);
		CSNavigator *getNavigator() {return mNavigator;}
		void setAlpha(int alpha);

		//!< inlines
		unsigned int getWorldPos() {return mWorldPos;}
		SpriteState getState() {return mState;}
		void setState(const SpriteState &state) {mState = state;}
		void setLayer(int z) {layer_position = z;}
		int getLayer() {return layer_position;}
		int getLayerY() {return y_position;}
		void clearNavigator() {mNavigator = 0;}
		void resetNavigator();
		int getMaxWidth() {return mWidth;}
		int getMaxHeight() {return mHeight;}
		int getSpriteType() {return mType;}
		int getSubtype() {return mSubtype;}
		int getCurrentStateId() {return mState.mCurrentBundleId;}
		std::string getFilename() {return mFilename;}
		std::string getId() {return mId;}
		void setPosition(int x, int y, unsigned int z);

		int getUserState(void) {return mState.mUserState;}
		void setUserState(int userState) {mState.mUserState = userState;}

		SDL_Surface *getScaledSprite(double factor)
		{
			return mState.mActionState.mAnimation->getScaledAnimation(mState.mActionState.mAnimationState, factor);
		}

		const char *getCurrentActionId()
		{
			return mState.mCurrentAction->getId();
		}
		int getCurrentActionType()
		{
			return mState.mCurrentAction->getActionType();
		}

		
		//!< return x start of surface coordinates of where the
		//!< the sprite will be drawn during next display
		unsigned int getDisplayXLower()		
		{
			int x = 
				mState.mActionState.mAnimationState.mDisplayParams.mXDisplayStart + 
				mState.mActionState.mAnimationState.mDisplayParams.mXPos -
				mState.mActionState.mAnimationState.mDisplayParams.mXWorldStart;
			return x;
		}

		//!< return y start of surface coordinates of where the
		//!< the sprite will be drawn during next display
		unsigned int getDisplayYLower()
		{
			int y = 
			mState.mActionState.mAnimationState.mDisplayParams.mYDisplayStart + 
			mState.mActionState.mAnimationState.mDisplayParams.mYPos -
			mState.mActionState.mAnimationState.mDisplayParams.mYWorldStart;
			return y;
		}

		//!< return x end of surface coordinates of where the
		//!< the sprite will be drawn during next display
		unsigned int getDisplayXUpper() //!< pos + width of sprite
		{
			int x = getDisplayXLower();
			x += mWidth;
			return x;
		}

		//!< return y end of surface coordinates of where the
		//!< the sprite will be drawn during next display
		unsigned int getDisplayYUpper() //!< pos + height of sprite
		{
			int y = getDisplayYLower();
			y += mHeight;
			return y;
		}
		
		//!< return the world x-coordinate of sprite that will be used during
		//!< the next display
		int getXPos()	
		{
			return mState.mActionState.mAnimationState.mDisplayParams.mXPos;
		}

		//!< return the world y-coordinate of sprite that will be used during
		//!< the next display
		int getYPos()
		{
			return mState.mActionState.mAnimationState.mDisplayParams.mYPos;
		}
		void setActive(bool active) 
		{
			mActive = active;
		}
};

//! factory - class
//! that is used to create CSSprite
//! uses xml - data-files, produces NOT "singletons"!!!
//! than means the result sprite must be freed by the caller (or whoever)!!!
class CSSpriteLoader : public Loadable<CSSprite>
{
	protected:
		//! not a singleton created!
		virtual bool isSingleCreate(void) const
		{
			return false;
		}

		virtual CSSprite *create(const std::string &filename)
		{
			return new CSSprite(filename);
		}

	public:
		static CSSpriteLoader INSTANCE;
};
 
#endif // CSSPRITE_H

