#ifndef CSTILE_H
#define CSTILE_H

// USES SDL_Surface

#include <string>
#include <vector>
#include "CSTypes.h"
#include "SDLMain.h"
#include "CSAnimation.h"
#include "CSLoadable.h"
#include "CSLog.h"

class CSTileLoader;
class CSTile;
typedef std::vector<CSTile *> CSTiles;

class CSTileData
{
	public:
		char *id;					//!< defining 
		unsigned char idChar;		//!< defining 
		char *animationName;		//!< animations
		int type;
		int subtype;

		CSTileData()
		{
			id = 0;
			animationName = 0;
			type = 0;
			subtype = 0;
		}

		~CSTileData()
		{
			if (id != 0)
			{
				free (id);
				id = 0;
			}
			if (animationName != 0)
			{
				free (animationName);
				animationName = 0;
			}	
		}
};

/**
	This file defines CSTile:
<PRE>
	  CSTile can only be created thru a loader (CSTileLoader).
	  CSTile are defined by (now ONLY thru) an ini-File, like:

			ID							"PAC_LEFT_ACTION"
			IS_REPEATABLE 				ACTION_INVERSE_ONCE
			PHASES						0
			ANIMATION					"animation\\pac_left.ani"
	  
	  ID	       - supposed a unique identifier
	  REAPEATABLE  - one of:
					 ACTION_NOT_REPEATABLE 		0
					 ACTION_REPEATABLE 			1
					 ACTION_INVERSE_ONCE 		2
					 ACTION_INVERSE_REPEATABLE 	3
					 ACTION_RANDOM 				4

	  PHASES       - many phases, each phase is represented by a number
					 the number corresponse to the ANIMATION that is
					 loaded with the ini - file (numbered from top to
					 bottom.
					 NOTE:
						REAPEATABLE			ACTION_INVERSE_ONCE
						PHASES				0
						PHASES				1
						PHASES				2
						PHASES				3
						PHASES				4
				
				     gives the same result as:
						REAPEATABLE			ACTION_NOT_REPEATABLE
						PHASES				0
						PHASES				1
						PHASES				2
						PHASES				3
						PHASES				4
						PHASES				3
						PHASES				2
						PHASES				1
						PHASES				0

	  ANIMATION    - many animations are possible!
					 describes the animation to be loaded,
	                 parameter gives the ini-file for a CSAnimation


	  Only two methods are really needed:
			virtual bool next() // if finished and not repeat -> false, true otherwise
</PRE>
	*/

class CSTile
{
	friend CSTileLoader;
	private:
		char *mId;							//!< defining 
		int mType;
		int mSubtype;
		unsigned char mIdChar;				//!< defining 
		CSAnimation *mAnimation;
		AnimationState mAnimationState;
		bool mChanged;
		bool mIsSolid;
		//! constructor using "*.act" file
		CSTile(const std::string &filename);
		static void loadTileData(const std::string &filename, CSTileData &data);
		unsigned int mMapPosX;
		unsigned int mMapPosY;
		void initialize(const CSTileData &data);
		void initialize();

	public:
		CSTile(const CSTile &tile);
		unsigned char getIDChar() {return mIdChar;}
		virtual ~CSTile();

		static const char *CLASS;
		virtual std::string getType() {return (std::string) CLASS;}

		unsigned int getX() {return mAnimationState.mXPos;}
		unsigned int getY() {return mAnimationState.mYPos;}
		void setX(unsigned int x) {mAnimationState.mXPos = x;}
		void setY(unsigned int y) {mAnimationState.mYPos = y;}

		unsigned int getWitdh() {return mAnimation->getMaxX();}
		unsigned int getHeight() {return mAnimation->getMaxY();}
		SDL_Surface *getScaledTile(double factor)
		{
			return mAnimation->getScaledAnimation(mAnimationState, factor);
		}

		void setMapPosition(unsigned int x, unsigned int y){mMapPosX = x;mMapPosY = y;}
		unsigned int getMapPositionX(){return mMapPosX;} //!< in pixel of World
		unsigned int getMapPositionY(){return mMapPosY;} //!< in pixel of World
		int getTileType(){return mType;}
		int getSubtype(){return mSubtype;}
		void setChanged(bool b){mChanged = b;}
		bool isChanged(){return mChanged;}
		virtual bool next(const CSDisplayParams &displayParams)
		{
			mAnimationState.mDisplayParams = displayParams;
			bool finished = false;
			finished = !(mAnimation->next(mAnimationState));

			if (finished)
			{
				mAnimation->resetState(mAnimationState);
			} 

			return !finished;
		}

		virtual void display()
		{
			mAnimation->display(mAnimationState);
		}

		void setSolid(bool b);
		bool isSolid() {return mIsSolid;}
		void setSecPerFrame(float secPerFrame);
		AnimationState *getAnimationState() {return &mAnimationState;}
		void scale(int width, int tilesX, int height, int tilesY);
};		

class CSTileLoader : public Loadable<CSTile>
{
	protected:
		virtual bool isSingleCreate(void) const
		{
			return false;
		}

		virtual CSTile *create(const std::string &filename)
		{
			return new CSTile(filename);
		}

	public:
		static CSTileLoader INSTANCE;
};

#endif // CSTILE_H
