/**
*/

#ifdef WIN32
#pragma warning(disable : 4786 )
#endif

#include "CSTileSet.h"
#include "CSXMLHelper.h"

CSTileSetLoader CSTileSetLoader::INSTANCE;
const char *CSTileSet::CLASS = "CSTileSet";

CSTileSet::CSTileSet(const std::string &filename)
{
	static char *functionName="CSTileSet";
	LOG_ENTER 
	CSTileSetData data;
	loadTileSetData(filename, data);
	initialize(data);
	LOG_EXIT
}

// not done
CSTileSet::CSTileSet(const CSTileSet &set)
{
	static char *functionName="CSTileSet";
	LOG_ENTER 
	initialize();
	mId = strdup(set.mId);
	mIsSolid = set.mIsSolid;
	mIsOneSize = set.mIsOneSize;
	mXSize = set.mXSize;
	mYSize = set.mYSize;
	LOG_EXIT
}

void CSTileSet::initialize(void)
{
	static char *functionName="initialize";
	LOG_ENTER 
	mId = 0;
	mTileCount = 0;
	mIsOneSize = false;
	mXSize = 0;
	mYSize = 0;
	mIsSolid = false;
	mSecPerFrame = 0;
	for (int i=0; i< MAX_TILE; i++)
	{
		mTiles[i] = 0;
	}
	LOG_EXIT
}

void CSTileSet::initialize(const CSTileSetData &data)
{
	static char *functionName="initialize";
	LOG_ENTER 
	initialize();
	mId = strdup(data.id);
	mTileCount = 0;
	for (StringVector::iterator iter = data.tileNames->begin(); iter != data.tileNames->end(); iter++)
	{
		char *tilename = (char *) *iter;
		CSTile *tile = CSTileLoader::INSTANCE.load(tilename);
		unsigned char value = tile->getIDChar();
		mTiles[value] = tile;
		mTileCount++;
	}
	mIsOneSize = data.onesize;
	mXSize = data.width;
	mYSize = data.height;

	LOG_EXIT
}

CSTileSet::~CSTileSet()
{
	static char *functionName="~CSTileSet";
	LOG_ENTER 
	if (mId != 0)
	{
		free(mId);
		mId = 0;
	}

	for (int i=0; i< MAX_TILE; i++)
	{
		if (mTiles[i] != 0)
		{
			delete (mTiles[i]);
			mTiles[i] = 0;
			mTileCount--;
		}
	}

	LOG_EXIT
}

// get a char that is not used by a tile
int CSTileSet::getFreeTile()
{
	for (int i=0; i< MAX_TILE; i++)
	{
		if (mTiles[i] == 0)
		{
			return i;
		}
	}
	return MAX_TILE;
}

// add a tile to the tileset, the tile will be using "c" as it's "key"
void CSTileSet::setTile(CSTile* tile, unsigned char c)
{
	if (mTiles[c] == 0)
	{
		if (tile == 0)
		{
			return;
		}
		mTiles[c] = tile;
		mTileCount++;
	}
	else
	{
		mTiles[c] = tile;
		if (tile == 0)
		{
			mTileCount--;
		}
	}
	if (tile != 0)
	{
		mTiles[c]->setSecPerFrame(mSecPerFrame);
		mTiles[c]->setSolid(mIsSolid);
		// experimental only for single tile, therefore shouldn't be changed...
		// mTiles[c]->scale(mWidth, mTilesX, mHeight, mTilesY);
	}
}

void CSTileSet::setSolid(bool b)
{
	static char *functionName="setSolid";
	LOG_ENTER 
	mIsSolid = b;
	for (int i=0; i< MAX_TILE; i++)
	{
		if (mTiles[i] != 0)
		{
			mTiles[i]->setSolid(b);
		}
	}
	LOG_EXIT
}

void CSTileSet::setSecPerFrame(float secPerFrame)
{
	static char *functionName="setSecPerFrame";
	LOG_ENTER 
	mSecPerFrame = secPerFrame;
	for (int i = 0; i< MAX_TILE; i++)
	{
		if (mTiles[i] != 0)
		{
			mTiles[i]->setSecPerFrame(secPerFrame);
		}
	}
	LOG_EXIT
}

void CSTileSet::loadTileSetData(const std::string &filename, CSTileSetData &data)
{
	static char *functionName="loadTileSetData";
	LOG_ENTER 
	char tmp[10];
	CSXMLHelper xmlSupport(filename, "TILESET");
	try
	{
		if (xmlSupport.getError())
		{
			LOG_EXIT
			throw "error";
		}

		data.id = strdup(xmlSupport.getString("ID").c_str());

		int tileCount = xmlSupport.getInt("count(TILES/NAME)");
		for (int i=0; i<tileCount; i++)
		{
			// i+1, predicates in xpath start with 1
			std::string a = xmlSupport.getString((std::string)"TILES/NAME["+itoa(i+1,tmp,10)+"]");
			data.tileNames->push_back(strdup(a.c_str()));
		}
		
		// One size does not neccessarily mean, that all tiles have the same size
		// but within tilemap and replacing of tiles at runtime
		// same sizes are assumed to center position of a "new" tile at a 
		// previously occupied position
		// sizes are used to calculate the position in such an event
		int oneSizeCount = xmlSupport.getInt("count(ONE_SIZE)");
		if (oneSizeCount>0)
		{
			data.onesize = xmlSupport.getInt("ONE_SIZE") == 1;
			data.width = xmlSupport.getInt("WIDTH");
			data.height = xmlSupport.getInt("HEIGHT");
		}
	}
	catch(...)
	{
		LOG_EXIT
		SDLMain::shutdown((std::string)"XML error \"" + filename + "\": " + xmlSupport.getErrorMessage().c_str(), 1);
	}
	LOG_EXIT
}

void CSTileSet::scaleToFit(int width, int tilesX, int height, int tilesY)
{
	static char *functionName="scaleToFit";
	LOG_ENTER 
	// scaling is experimental!!!
	// only for single tile background usage!
	if (mTileCount != 1)
	{
		LOG_EXIT
		return;
	}
	for (int i=0; i< MAX_TILE; i++)
	{
		if (mTiles[i] != 0)
		{
			mTiles[i]->scale(width, tilesX, height, tilesY);
		}
	}
	LOG_EXIT
}

