#ifndef CSWORLD_H
#define CSWORLD_H

// USES SDL_Surface
// USES SDL_Rect 

#include <string>
#include <vector>
#include "CSTypes.h"
#include "SDLMain.h"
#include "CSTileMap.h"
#include "CSSprite.h"
#include "CSLoadable.h"
#include "CSLog.h"

class CSWorldLoader;
class CSWorld;
class CSMusic;

typedef std::vector<CSWorld *> CSWorlds;

class MapData
{
	public:
		char *name;					
		int offsetX;
		int offsetY;
		bool isBackground;
		bool isParallax;
	MapData()
	{
		name = 0;
		offsetX = 0;
		offsetY = 0;
		isBackground = false;;
		isParallax = false;
	}
	
	~MapData()
	{
		if (name != 0)
		{
			free(name);
			name = 0;
		}
	}
};

typedef std::vector<MapData *> MapDatas;
class CSWorldData
{
	public:
		char *id;					// defining 
		char *musicName;
		MapDatas *maps;
		int layers;

		CSWorldData()
		{
			musicName = 0;
			id = 0;
			layers = 0;
			maps = new MapDatas();
		}

		~CSWorldData()
		{
			if (id != 0)
			{
				free (id);
				id = 0;
			}
			if (musicName != 0)
			{
				free (musicName);
				musicName = 0;
			}
			if (maps != 0)
			{
				for (MapDatas::iterator iter = maps->begin(); iter != maps->end(); iter++)
				{
					MapData *help = (MapData *) *iter;
					free(help->name);
					help->name = 0;
					delete help;
				}
				delete maps;
			}
		}
};

class CSWorld
{
	friend CSWorldLoader;
	private:
		int mXPos;
		int mYPos;
		SDL_Rect mScaledWorldDestination;
		SDL_Rect mMapArea;
		int mScaleMapCount;
		double mFactor;
		char *mId;							// defining 
		CSTileMap **mWorld;
		unsigned int mLayers;
		int mXOffset;
		int mYOffset;
		unsigned int mHeight;	// Height in pixel
		unsigned int mWidth;	// Width in pixel
		unsigned int mWorldNo;
		bool mActive;
		CSSprites mSprites;
		CSMusic *mMusic;

		// constructor using "*.wrl" file
		CSWorld(const std::string &filename);
		static void loadWorldData(const std::string &filename, CSWorldData &data);

		void initialize(const CSWorldData &data);
		void initialize();
	
	public:
		CSWorld(const CSWorld &world);
		virtual ~CSWorld();
		static const char *CLASS;
		virtual std::string getType() {return (std::string) CLASS;}

		void display();
		void displayScaled(CSSprite *centerSprite);
		void update();
		void setSecPerFrame(float secPerFrame);
		void setDisplayArea(SDL_Rect area);
		void setWorldPosition(int x, int y);
		void addSprite(CSSprite *sprite);
		void removeSprite(CSSprite *sprite);
		void setWorldNo(unsigned int no);

		void scaleBackgroundTiles(int w, int h);
		void buildScaledWorld(double factor, const SDL_Rect &srcRect, SDL_Surface *destination, const SDL_Rect &destRect);
		void setScaleFactor(double factor) {mFactor = factor;}

		void setScaleDisplayArea(SDL_Rect area);
		void resetScaledWorld();

		void setScaleDisplayLayer(bool b, int layer)
		{
			if (mWorld[layer]->isScaleDisplay() == b)
			{
				return;
			}
			if (b)
			{
				mScaleMapCount++;
			}
			else 
			{
				mScaleMapCount--;
			}
			mWorld[layer]->setScaleDisplay(b);
		}
		
		unsigned int getDisplayXLowerBound() {return mMapArea.x;}
		unsigned int getDisplayXUpperBound() {return mMapArea.x + mMapArea.w;}
		unsigned int getDisplayYLowerBound() {return mMapArea.y;}
		unsigned int getDisplayYUpperBound() {return mMapArea.y + mMapArea.h;}
		unsigned int getHeight(){return mHeight;}
		unsigned int getWidth(){return mWidth;}
		void setTile(unsigned int x, unsigned int y, unsigned int z, char idChar); // old tile deleted!
		void setTile(unsigned int x, unsigned int y, unsigned int z, CSTile *tile); // old tile not deleted, but overwritten
		CSTile *getTileAtPixel(unsigned int x, unsigned int y, unsigned int z)
		{
			return mWorld[z]->getTileAtPixel(x, y);
		}
		CSTile *getTile(unsigned int x, unsigned int y, unsigned int z)
		{
			return mWorld[z]->getTile(x, y);
		}
		
		unsigned int getLayerCount() {return mLayers;}
		CSTileMap *getLayer(unsigned int layer) 
		{
			if (layer > mLayers) return 0;
			return mWorld[layer];
		}
		int getWorldPositionX() {return mXPos;}
		int getWorldPositionY() {return mYPos;}
		bool isBorderTile(CSTile * tile, int z) {return mWorld[z]->isBorderTile(tile);}
		void setActive(bool active) 
		{
			mActive = active;
			if (mWorld != 0)
			{
				for (int l=0; l< mLayers; l++)
				{
					mWorld[l]->setActive(mActive);
				}
			}
		}
};		

class CSWorldLoader : public Loadable<CSWorld>
{
	protected:
		virtual bool isSingleCreate(void) const
		{
			return false;
		}

		virtual CSWorld *create(const std::string &filename)
		{
			return new CSWorld(filename);
		}

	public:
		static CSWorldLoader INSTANCE;
};

#endif // CSWORLD_H
