#ifdef USE_XALAN

#include "CSXMLHelper.h"

#ifdef WIN32
#pragma warning(disable : 4786 )
#endif

#include <vector>
const char *CSXMLHelper::CLASS = "CSXMLHelper";


// Base header file.  Must be first.
#include <Include/PlatformDefinitions.hpp>

#if defined(XALAN_OLD_STREAM_HEADERS)
#include <iostream.h>
#else
#include <iostream>
#endif
#include <util/PlatformUtils.hpp>
#include <framework/LocalFileInputSource.hpp>

#include <XalanDOM/XalanDocument.hpp>
#include <XalanDOM/XalanElement.hpp>

#include <XPath/XObject.hpp>
#include <XPath/XPathEvaluator.hpp>

#include <XalanSourceTree/XalanSourceTreeDOMSupport.hpp>
#include <XalanSourceTree/XalanSourceTreeInit.hpp>
#include <XalanSourceTree/XalanSourceTreeParserLiaison.hpp>

#include <sax/ErrorHandler.hpp>



int CSXMLHelper::mLock = 0;
CSXMLHelper::CSXMLHelper(const std::string &xmlFilename, const std::string &contextNode)
{
	static char *functionName="CSXMLHelper";
	LOG_ENTER 
	mError = CSXML_OK;
	mInit = 0;
	theEvaluator = 0;
	theLiaison = 0;
	theDOMSupport = 0;
	theSourceTreeInit = 0;
	
	if (mLock != 0)
	{
		std::string errorMessageString;
		errorMessageString = std::string("Error - only one single instance of CSXMLHelper allowed!");
		mErrorMessageString = mErrorMessageString + "\n"+ errorMessageString;
		mError = CSXML_ONLY_SINGLE_XML_HELPER_ALLOWED;
	}
	mLock = 1;

	try
	{
		theSourceTreeInit = new XalanSourceTreeInit();
		theDOMSupport = new XalanSourceTreeDOMSupport();
		theLiaison = new XalanSourceTreeParserLiaison(*theDOMSupport);

		theLiaison->setUseValidation(true);
		theLiaison->setErrorHandler(this);
		
		// Hook the two together...
		theDOMSupport->setParserLiaison(theLiaison);

		XMLPlatformUtils::Initialize();		// attention
		XPathEvaluator::initialize();		// attention
		mInit = 1;

		// Create an input source that represents a local file...
		const LocalFileInputSource	theInputSource(c_wstr(XalanDOMString(
				xmlFilename.c_str()
				)));

		theEvaluator = new XPathEvaluator();

		// Parse the document...
		theDocument = theLiaison->parseXMLStream(theInputSource);
		if (theDocument == 0)
		{
			mError = CSXML_DOCUMENT_LOADING_ERROR;
			std::string errorMessageString;
			errorMessageString = std::string("Error - while loading document \"" + xmlFilename +"\"!");
			mErrorMessageString = mErrorMessageString + "\n"+ errorMessageString;
		}

		// OK, let's find the context node...

		theContextNode =
				theEvaluator->selectSingleNode(
					*theDOMSupport,
					theDocument,
					XalanDOMString(contextNode.c_str()).c_str(),
					theDocument->getDocumentElement());

		if (theContextNode == 0)
		{
			mError = CSXML_CONTEXT_LOADING_ERROR;
			std::string errorMessageString;
			errorMessageString = std::string("Error - while loading context \"" + contextNode +"\"!");
			mErrorMessageString = mErrorMessageString + "\n"+ errorMessageString;
		}
	}
	catch(const SAXException&)
	{
		mError = CSXML_INITIALIZE_ERROR;
		std::string errorMessageString;
		errorMessageString = std::string("SAXException") + mErrorMessageString;
		mErrorMessageString = mErrorMessageString + "\n"+ errorMessageString;
	}
	catch(const XMLException&)
	{
		mError = CSXML_INITIALIZE_ERROR;
		std::string errorMessageString;
		errorMessageString = std::string("XMLException") + mErrorMessageString;
		mErrorMessageString = mErrorMessageString + "\n"+ errorMessageString;
	}
	catch(const XalanDOMException&)
	{
		mError = CSXML_INITIALIZE_ERROR;
		std::string errorMessageString;
		errorMessageString = std::string("XalanDOMException") + mErrorMessageString;
		mErrorMessageString = mErrorMessageString + "\n"+ errorMessageString;
	}
	LOG_EXIT
}

CSXMLHelper::~CSXMLHelper()
{
	static char *functionName="~CSXMLHelper";
	LOG_ENTER 

	delete (theEvaluator);
	delete (theLiaison);
	delete (theDOMSupport);
	delete (theSourceTreeInit);

	if (mInit != 1)
	{
		LOG_EXIT
		return;
	}
	XPathEvaluator::terminate();		// attention
	XMLPlatformUtils::Terminate();		// attention
	mLock = 0;
	LOG_EXIT
}

int CSXMLHelper::getInt(const std::string &xpath)
{
	static char *functionName="getInt";
	int ret = 0;
	if (mError != 0)
	{
		LOG_WARN_MESSAGE("Error")
		throw "error";
	}
	try
	{
		XObjectPtr xml = getXPathResult(xpath);
		ret = DOMStringToInt(xml->str());
	}
	catch(const SAXException&)
	{
		mError = CSXML_XPATH_ERROR;
		std::string errorMessageString;
		errorMessageString = std::string("SAXException") + mErrorMessageString;
		mErrorMessageString = mErrorMessageString + "\n"+ errorMessageString;
		LOG_WARN_MESSAGE(mErrorMessageString)
		throw "error";
	}
	catch(const XMLException&)
	{
		mError = CSXML_XPATH_ERROR;
		std::string errorMessageString;
		errorMessageString = std::string("XMLException") + mErrorMessageString;
		mErrorMessageString = mErrorMessageString + "\n"+ errorMessageString;
		LOG_WARN_MESSAGE(mErrorMessageString)
		throw "error";
	}
	catch(const XalanDOMException&)
	{
		mError = CSXML_XPATH_ERROR;
		std::string errorMessageString;
		errorMessageString = std::string("XalanDOMException") + mErrorMessageString;
		mErrorMessageString = mErrorMessageString + "\n"+ errorMessageString;
		LOG_WARN_MESSAGE(mErrorMessageString)
		throw "error";
	}
	return ret;
}

std::string CSXMLHelper::getString(const std::string &xpath)
{
	static char *functionName="getString";
	std::string cstr;
	if (mError != 0)
	{
		LOG_WARN_MESSAGE("XML! Not properly initialized!")
		throw std::string("XML! Not properly initialized!");
	}
	try
	{
		XObjectPtr xml = getXPathResult(xpath);
		cstr = toAscii(xml);
	}
	catch(const SAXException&)
	{
		mError = CSXML_XPATH_ERROR;
		std::string errorMessageString;
		errorMessageString = std::string("SAXException") + mErrorMessageString;
		mErrorMessageString = mErrorMessageString + "\n"+ errorMessageString;
		LOG_WARN_MESSAGE(mErrorMessageString)
		throw "error";
	}
	catch(const XMLException&)
	{
		mError = CSXML_XPATH_ERROR;
		std::string errorMessageString;
		errorMessageString = std::string("XMLException") + mErrorMessageString;
		mErrorMessageString = mErrorMessageString + "\n"+ errorMessageString;
		LOG_WARN_MESSAGE(mErrorMessageString)
		throw "error";
	}
	catch(const XalanDOMException&)
	{
		mError = CSXML_XPATH_ERROR;
		std::string errorMessageString;
		errorMessageString = std::string("XalanDOMException") + mErrorMessageString;
		mErrorMessageString = mErrorMessageString + "\n"+ errorMessageString;
		LOG_WARN_MESSAGE(mErrorMessageString)
		throw "error";
	}
	return cstr;
}


XObjectPtr CSXMLHelper::getXPathResult(const std::string &xpath)
{
	static char *functionName="getXPathResult";
	static char h[1025];
	LOG_ENTER
	h[0] = 0;
	strcat(h, "normalize-space(");
	strcat(h, xpath.c_str());
	strcat(h, ")");

	const XObjectPtr	theResult(
		theEvaluator->evaluate(
				*theDOMSupport,
				theContextNode,
				XalanDOMString(h).c_str(),
				theDocument->getDocumentElement())
				);
	if (theResult.null() != false)
	{
		mError = CSXML_XPATH_ERROR;
		std::string errorMessageString;
		errorMessageString = std::string("General XPATH-Result = 0") + mErrorMessageString;
		mErrorMessageString = mErrorMessageString + "\n"+ errorMessageString;
		LOG_EXIT
		throw "error";
	}
	LOG_DEBUG_MESSAGE(xpath + "->" + toAscii(theResult) ) 
	LOG_EXIT
	return theResult;
}

std::string CSXMLHelper::toAscii(const XObjectPtr &xObjectPtr)
{
	static char *functionName="toAscii";
	std::string cstr;
	const XalanDOMString xmlString = xObjectPtr->str();
	return toAscii(xmlString);
}

std::string CSXMLHelper::toAscii(const XalanDOMString &xmlString)
{
	static char *functionName="toAscii";
	const CharVectorType charVec = TranscodeToLocalCodePage (xmlString);
	return toAscii(charVec);
}

// typedef std::vector<char> CharVectorType
std::string CSXMLHelper::toAscii(const CharVectorType &charVec)
{
	static char *functionName="toAscii";
	std::string cstr;
	CharVectorType::const_iterator iter = charVec.begin();
	char helper[2];
	helper[1] = 0;
	while (iter != charVec.end())
	{
		helper[0] = *iter;
		cstr = cstr+helper;
		iter++;
	}
	return std::string(cstr);
}

std::string CSXMLHelper::toAscii(const DOMString &domString)
{
	static char *functionName="toAscii";
    char* charArray = domString.transcode();
	std::string ret(charArray);
	free(charArray);
	return ret;
}

std::string CSXMLHelper::toAscii(const XMLCh* xmlChars)
{
	return toAscii(DOMString(xmlChars));
}

int CSXMLHelper::getError(void)
{
	return mError;
}

std::string CSXMLHelper::getErrorMessage(void)
{
	return mErrorMessageString;
}

void CSXMLHelper::warning(const SAXParseException &toCatch)
{
	static char *functionName="warning";
	LOG_ENTER 
	char line[10];
	char column[10];
	itoa(toCatch.getLineNumber(),line,10);
	itoa(toCatch.getColumnNumber(),column,10);
	std::string errorMessageString;
	errorMessageString = errorMessageString
						  + std::string("Warning at File \"")
						  + toAscii(toCatch.getSystemId()) + "\", "
						  + std::string(line) + ", "
						  + std::string(column) + " : "
						  + toAscii(toCatch.getMessage()) + "\n";
	mErrorMessageString = errorMessageString;
	mError = CSXML_DOCUMENT_LOADING_ERROR;
	LOG_EXIT
}

void CSXMLHelper::error(const SAXParseException& toCatch)
{
	static char *functionName="error";
	LOG_ENTER 
	char line[10];
	char column[10];
	itoa(toCatch.getLineNumber(),line,10);
	itoa(toCatch.getColumnNumber(),column,10);
	std::string errorMessageString;
	errorMessageString = errorMessageString
						  + std::string("Error at File \"")
						  + toAscii(toCatch.getSystemId()) + "\", "
						  + std::string(line) + ", "
						  + std::string(column) + " : "
						  + toAscii(toCatch.getMessage()) + "\n";
	mErrorMessageString = errorMessageString;
	mError = CSXML_DOCUMENT_LOADING_ERROR;
	LOG_EXIT
}

void CSXMLHelper::fatalError(const SAXParseException& toCatch)
{
	static char *functionName="fatal";
	LOG_ENTER 
	char line[10];
	char column[10];
	itoa(toCatch.getLineNumber(),line,10);
	itoa(toCatch.getColumnNumber(),column,10);
	std::string errorMessageString;
	errorMessageString = errorMessageString
						  + std::string("Fatal Error at File \"")
						  + toAscii(toCatch.getSystemId()) + "\", "
						  + std::string(line) + ", "
						  + std::string(column) + " : "
						  + toAscii(toCatch.getMessage()) + "\n";
	mErrorMessageString = errorMessageString;
	mError = CSXML_DOCUMENT_LOADING_ERROR;
	LOG_EXIT
}

void CSXMLHelper::resetErrors()
{
	static char *functionName="resetErrors";
	LOG_ENTER 
	mError = CSXML_OK;
	mErrorMessageString = std::string();
	LOG_EXIT
}

#endif // USE_XALAN