/**
	This class is responsible for setting up a screen,
	providing access to the screen and handling screen updates.

	This class is a singleton and one can only access it thru:
		SDLMain *SDLMain::getInstance()

	Upon the first call of this method the SDL Screen is initialized.
	(Later from a config - file - no hardcoded parameters)

	Using the methods:
		void SDLMain::shutdown()
		void SDLMain::shutdown(std::string message, unsigned int errorNo)
	
	The client can shutdown SDL usage, either with or without an error message.
	(exit() is called - so the program quits!)

	With the methods:

		void SDLMain::addUpdateRect(const SDL_Rect &saveRegion)
		void SDLMain::addUpdateRect(signed short x, signed short y, unsigned short w, unsigned short h)
	
	The Client can add update regions of the screen for the next update "round".
	At the moment only 
		  const int NUM_UPDATE_RECTS = 500;
	update-rectangles are supported. But changing the value for higher needs is ok.

	Using:
		void SDLMain::updateScreen()
	Updates the screen, the update methods respects last set update regions, and
	page flipping for double buffering!

	The methods:
		int SDLMain::getPageNo()
		int SDLMain::getCurrentPage()
	Are provided for double-buffering, they 
	a. return the number of pages currently used (right now either 1 or 2)
	b. the current page number (0 or 1)
  */

#ifdef WIN32
#pragma warning(disable : 4786 )
#endif
#include "SDLMain.h"
#include "SDL_rotozoom.h"

#include <stdlib.h>
#include <stdio.h> 
#include <stdarg.h>

SDLMain *SDLMain::mInstance = 0;
SDL_Surface *SDLMain::mScreen = 0;
SDL_Rect *SDLMain::mRects = 0;

int SDLMain::mCurrentCount = 0;
int SDLMain::mNoPages = 1;		// default
int SDLMain::mPage = 0;

int SDLMain::mWidth = 640;
int SDLMain::mHeight = 480;
int SDLMain::mDepth = 32;
bool SDLMain::mUpdateWholeScreen = false;
	
bool SDLMain::mIsHardware = true;
bool SDLMain::mIsDoublebuffer = true;
bool SDLMain::mIsAnyFormat = true;
bool SDLMain::mIsFullscreen = false;
unsigned int SDLMain::mMode = 0;

unsigned int SDLMain::mAudioFormat = AUDIO_S16;
unsigned int SDLMain::mAudioChannels = 2;
unsigned int SDLMain::mAudioBuffers = 4096;
unsigned int SDLMain::mAudioRate = 44100;
bool SDLMain::mSDLInit = false;
const char *SDLMain::CLASS = "SDLMain";

SDLMain::SDLMain() // private
{
	static char *functionName="SDLMain";
	LOG_ENTER 
	MESSAGE_SCREEN_CHANGED.setSubtype(SCREEN_CHANGED_MESSAGE);

	LOG_INFO_MESSAGE("Initializing SDL.") 

    /* Initialize the SDL library */
    if( SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO) < 0 ) 
	{
		shutdown((std::string) "Couldn't initialize SDL: " + SDL_GetError(), 1);
    }
	LOG_INFO_MESSAGE("SDL initialized.") 

	if (Mix_OpenAudio(mAudioRate, mAudioFormat, mAudioChannels, mAudioBuffers))
	{
		shutdown((std::string) "Couldn't initialize SDL_Mixer: " + SDL_GetError(), 1);
	}
	LOG_INFO_MESSAGE("SDL_Mixer initialized.") 
	SDL_EnableUNICODE(1);


	resetVideo();
	
	// build number of update recangles!
	mRects = new SDL_Rect[NUM_UPDATE_RECTS];
	mCurrentCount = 0;
	mSDLInit = true;
	LOG_EXIT
}

SDLMain::~SDLMain()
{
	static char *functionName="~SDLMain";
	LOG_ENTER 
	SDLMain *instance = mInstance;
	mInstance = 0;
	if (instance != 0)
	{
		delete instance;
		instance = 0;
	}
	if (mRects != 0)
	{
		delete []mRects;
		mRects = 0;
	}
	LOG_EXIT
}

void SDLMain::clearScreen( int red, int green, int blue )
{
	long int color = SDL_MapRGB( mScreen->format, red, green, blue );
	SDL_FillRect( mScreen, 0, color ); 
	SDL_UpdateRect( mScreen, 0, 0, 0, 0 );
}

SDLMain *SDLMain::getInstance()
{
	static char *functionName="getInstance";
	LOG_ENTER 
	if (mInstance == 0)
	{
		mInstance = new SDLMain();
	}
	LOG_EXIT
	return mInstance;
}

void SDLMain::shutdown()
{
	shutdown("Quiting SDL.\n", 0);
}

void SDLMain::shutdown(std::string message, unsigned int errorNo)
{
	static char *functionName="shutdown";
	LOG_ENTER 
	SDLMain *instance = mInstance;
	mInstance = 0;
	if (instance != 0)
	{
		delete instance;
	}

    /* Shutdown all subsystems */
	Mix_CloseAudio();
    SDL_Quit();
	mSDLInit = false;
	LOG_ERROR_PRINT(message, (long)errorNo) 

	LOG_EXIT
	exit(errorNo);
}

void SDLMain::adjustClipping(SDL_Rect &newRect, SDL_Rect &oldRect)
{
	if (newRect.x<oldRect.x)
	{
		newRect.w -= oldRect.x-newRect.x;
		newRect.x = oldRect.x;
	}
	if (newRect.w > 60000)
	{
		newRect.w = 0;
	}
	if (newRect.y<oldRect.y)
	{
		newRect.h -= oldRect.y-newRect.y;
		newRect.y = oldRect.y;
	}
	if (newRect.h > 60000)
	{
		newRect.h = 0;
	}
	if (newRect.x + newRect.w > oldRect.x + oldRect.w)
	{
		newRect.w -= (newRect.x + newRect.w) - (oldRect.x + oldRect.w); 
	}
	if (newRect.y + newRect.h > oldRect.y + oldRect.h)
	{
		newRect.h -= (newRect.y + newRect.h) - (oldRect.y + oldRect.h); 
	}

	if (newRect.x < 0) 
	{
		newRect.w += newRect.x;
		newRect.x = 0;
	}
	if (newRect.y < 0) 
	{
		newRect.h += newRect.y;
		newRect.y = 0;
	}
	if (newRect.w > 60000) 
		newRect.w = 0;
	if (newRect.h > 60000) 
		newRect.h = 0;
}

void SDLMain::setScreenWidth(int w)
{
	static char *functionName="setScreenWidth";
	LOG_ENTER 
	if (mWidth == w)
	{
		LOG_EXIT
		return;
	}
	mWidth = w;
	resetVideo();
	LOG_EXIT
}

void SDLMain::setScreenHeight(int h)
{
	static char *functionName="setScreenHeight";
	LOG_ENTER 
	if (mHeight == h)
	{
		LOG_EXIT
		return;
	}
	mHeight = h;
	resetVideo();
	LOG_EXIT
}

void SDLMain::setScreenDepth(int d)
{
	static char *functionName="setScreenDepth";
	LOG_ENTER 
	if (mDepth == d)
	{
		LOG_EXIT
		return;
	}
	mDepth = d;
	resetVideo();
	LOG_EXIT
}

void SDLMain::setHardwareSurface(bool h)
{
	static char *functionName="setHardwareSurface";
	LOG_ENTER 
	if (mIsHardware == h)
	{
		LOG_EXIT
		return;
	}
	mIsHardware = h;
	resetVideo();
	LOG_EXIT
}

void SDLMain::setDoubleBufferSurface(bool d)
{
	static char *functionName="setDoubleBufferSurface";
	LOG_ENTER 
	if (mIsDoublebuffer == d)
	{
		LOG_EXIT
		return;
	}
	mIsDoublebuffer = d;
	resetVideo();
	LOG_EXIT
}

void SDLMain::setFullscreenSurface(bool f)
{
	static char *functionName="setFullscreenSurface";
	LOG_ENTER 
	if (mIsFullscreen == f)
	{
		LOG_EXIT
		return;
	}
	mIsFullscreen = f;
	resetVideo();
	LOG_EXIT
}

void SDLMain::setUseScreenAnyhowSurface(bool a)
{
	static char *functionName="setUseScreenAnyhowSurface";
	LOG_ENTER 
	if (mIsAnyFormat == a)
	{
		LOG_EXIT
		return;
	}
	mIsAnyFormat = a;
	resetVideo();
	LOG_EXIT
}

unsigned int SDLMain::getMode()
{
	static char *functionName="getMode";
	LOG_ENTER 
	unsigned int mode = 0;
	if (isHardwareSurface())
	{
		mode = mode | SDL_HWSURFACE;
	}
	else 
	{
		mode = mode | SDL_SWSURFACE;
	}

	if (isDoubleBufferSurface())
	{
		mode = mode | SDL_DOUBLEBUF;
	}

	if (isFullscreenSurface())
	{
		mode = mode | SDL_FULLSCREEN;
	}

	if (isUseScreenAnyhowSurface())
	{
		mode = mode | SDL_ANYFORMAT;
	}
	mMode = mode;	
	LOG_EXIT
	return mMode;
}

void SDLMain::resetVideo(void)
{
	static char *functionName="resetVideo";
	LOG_ENTER 
	// should be read from an ini file at some stage
	mScreen = SDL_SetVideoMode(	getScreenWidth(), 
								getScreenHeight(), 
								getScreenDepth(), 
								getMode());

    if ( mScreen == NULL ) 
	{
		std::string message;
		message = message + "Couldn't set video mode: " + SDL_GetError();
		shutdown(message, 1);
    }
	LOG_INFO_PRINT("Set Mode at bits-per-pixel mode:", (long)mScreen->format->BitsPerPixel) 
	if (mScreen->flags & SDL_DOUBLEBUF) 
	{
		mNoPages = 2;
		mPage = 0;
	}
//	mIsDoublebuffer = (mScreen->flags & SDL_DOUBLEBUF) == SDL_DOUBLEBUF;
//	mIsHardware = (mScreen->flags & SDL_HWSURFACE) == SDL_HWSURFACE;
//	mIsFullscreen = (mScreen->flags & SDL_FULLSCREEN) == SDL_FULLSCREEN;

	if (mIsHardware) 
	{
		LOG_INFO_MESSAGE("SDL_HWSURFACE") 
	}
	if (mIsDoublebuffer) 
	{
		LOG_INFO_MESSAGE("DOUBLEBUFFER") 
	}
	if (mIsFullscreen) 
	{
		LOG_INFO_MESSAGE("FULLSCREEN") 
	}

	if (mInstance != 0)
	{
		mInstance->sendMessage(mInstance->MESSAGE_SCREEN_CHANGED);
	}
	SDL_ShowCursor(SDL_DISABLE);
	LOG_EXIT
}

void SDLMain::toggleFullscreen()
{
	static char *functionName="toggleFullscreen";
	LOG_ENTER 
	setFullscreenSurface(!isFullscreenSurface());
	LOG_EXIT
}

void SDLMain::addUpdateRect(const SDL_Rect &saveRegion)
{
	static char *functionName="addUpdateRect";
	if (mUpdateWholeScreen)
	{
		return;
	}

	if (mRects == 0)
	{
		return;
	}
	if (mCurrentCount >= NUM_UPDATE_RECTS)
	{
		SDLMain::shutdown((std::string)"Not enough Update Rects available: " + SDL_GetError(), 1);
	}
	mRects[mCurrentCount].x = saveRegion.x;
	mRects[mCurrentCount].y = saveRegion.y;
	mRects[mCurrentCount].w = saveRegion.w;
	mRects[mCurrentCount].h = saveRegion.h;
	mCurrentCount++;
}

void SDLMain::addUpdateRect(signed short x, signed short y, unsigned short w, unsigned short h)
{
	static char *functionName="addUpdateRect";
	if (mUpdateWholeScreen)
	{
		return;
	}
	if (mRects == 0)
	{
		return;
	}
	if (mCurrentCount >= NUM_UPDATE_RECTS)
	{
		SDLMain::shutdown((std::string)"Not enough Update Rects available: " + SDL_GetError(), 1);
	}
	mRects[mCurrentCount].x = x;
	mRects[mCurrentCount].y = y;
	mRects[mCurrentCount].w = w;
	mRects[mCurrentCount].h = h;
	mCurrentCount++;
}

void SDLMain::updateScreen()
{
	static char *functionName="updateScreen";
	if (mScreen->flags & SDL_DOUBLEBUF)
	{
		SDL_Flip(mScreen);
		mPage = (mPage + 1) % mNoPages;
	}
	else
	{
		if (mUpdateWholeScreen)
		{
			SDL_UpdateRect(mScreen, 0,0,0,0);
		}
		else
		{
			if (mRects == 0)
			{
				return;
			}
		    SDL_UpdateRects(mScreen, mCurrentCount, mRects);
		}
	}

	mCurrentCount = 0;
}

int SDLMain::getPageNo()
{
	return mNoPages;
}

int SDLMain::getCurrentPage()
{
	return mPage;
}

Uint32 SDLMain::getPixel(SDL_Surface *Surface, Sint32 X, Sint32 Y)
{
	static char *functionName="getPixel";
	Uint32 Bpp;

	if ((X<0) || (X>=Surface->w))
	{
		return (Uint32)-1;
	}
   
	Bpp = Surface->format->BytesPerPixel;

	// Get the pixel
	switch(Bpp) 
	{
		case 1:
			return *((Uint8 *)Surface->pixels + Y * Surface->pitch + X);
			break;
		case 2:
			return *((Uint16 *)Surface->pixels + Y * Surface->pitch/2 + X);
			break;
		case 3: 
		{ // Format/endian independent 
			Uint8 r, g, b;
			Uint8  *bits;
			bits = ((Uint8 *)Surface->pixels)+Y*Surface->pitch+X*Bpp;
         
			r = *((bits)+Surface->format->Rshift/8);
			g = *((bits)+Surface->format->Gshift/8);
			b = *((bits)+Surface->format->Bshift/8);
         
			return SDL_MapRGB(Surface->format, r, g, b);
		}
		break;
		case 4:
			return *((Uint32 *)Surface->pixels + Y * Surface->pitch/4 + X);
			break;
	}
	return -1;
}

void SDLMain::setPixel(SDL_Surface *surface, Sint32 x, Sint32 y, Uint32 color)
{
	static char *functionName="setPixel";
	static Uint8 ri, gi, bi;
	static Uint8 *bits;
	static Uint8 bpp;
	static SDL_Rect rect;

	bpp = surface->format->BytesPerPixel;
	bits = ((Uint8 *) surface->pixels) + y * surface->pitch + x * bpp;
	/* Set the pixel */
	switch (bpp) {
		case 1:
			*((Uint8 *) (bits)) = (Uint8) color;
			break;

		case 2:
			*((Uint16 *) (bits)) = (Uint16) color;
			break;

		case 3: {			/* Format/endian independent */

				ri = (color >> surface->format->Rshift) & 0xFF;
				gi = (color >> surface->format->Gshift) & 0xFF;
				bi = (color >> surface->format->Bshift) & 0xFF;
				*((bits) + surface->format->Rshift / 8) = ri;
				*((bits) + surface->format->Gshift / 8) = gi;
				*((bits) + surface->format->Bshift / 8) = bi;
			}
			break;

		case 4:
			*((Uint32 *) (bits)) = (Uint32) color;
			break;
	}
}

SDL_Surface *SDLMain::scale(SDL_Surface *surface, int width, int height)
{
	static char *functionName="scale";
	LOG_ENTER 
	double zoomx;
	double zoomy;
	zoomx = (double)(((double)(width)) / ((double) surface->w));
	zoomy = (double)(((double)(height)) / ((double) surface->h));
	SDL_Surface *ret = zoomSurface(surface, zoomx, zoomy, 0);
	LOG_EXIT
	return ret;
}

SDL_Surface *SDLMain::scale(SDL_Surface *surface, int width, int divideX, int height, int divideY)
{
	static char *functionName="scale";
	LOG_ENTER 
	double zoomx;
	double zoomy;
	zoomx = (double)(((double)(width)) / ((double) divideX));
	zoomy = (double)(((double)(height)) / ((double) divideY));
	SDL_Surface *ret = zoomSurface(surface, zoomx, zoomy, 0);
	LOG_EXIT
	return ret;
}

SDL_Surface *SDLMain::scale(SDL_Surface *surface, double factor)
{
	return zoomSurface(surface, factor, factor, 0);
}

int SDLMain::scaleX(int width, int divideX, int x)
{
	static char *functionName="scaleX";
	LOG_ENTER 
	double zoomx;
	zoomx = (double)(((double)(width)) / ((double) divideX));
    int destx, desty;
	
	zoomSurfaceSize(x, 1, zoomx, 1, &destx, &desty);
	if (x==0) destx = 0;
	LOG_EXIT
	return destx;
}

int SDLMain::scaleY(int height, int divideY, int y)
{
	static char *functionName="scaleY";
	LOG_ENTER 
	double zoomy;
	zoomy = (double)(((double)(height)) / ((double) divideY));
    int destx, desty;
	zoomSurfaceSize(1, y, 1, zoomy, &destx, &desty);
	if (y==0) desty = 0;
	LOG_EXIT
	return desty;
}


/* Clipping based heavily on code from                       */
/* http://www.ncsa.uiuc.edu/Vis/Graphics/src/clipCohSuth.c   */
#define CLIP_LEFT_EDGE   0x1
#define CLIP_RIGHT_EDGE  0x2
#define CLIP_BOTTOM_EDGE 0x4
#define CLIP_TOP_EDGE    0x8
#define CLIP_INSIDE(a)   (!a)
#define CLIP_REJECT(a,b) (a&b)
#define CLIP_ACCEPT(a,b) (!(a|b))

static int clipEncode(Sint16 x, Sint16 y, Sint16 left, Sint16 top, Sint16 right, Sint16 bottom)
{
	char *CLASS = "None";
	static char *functionName="clipEncode";
	LOG_ENTER 
    int code = 0;
    if (x < left) {
	code |= CLIP_LEFT_EDGE;
    } else if (x > right) {
	code |= CLIP_RIGHT_EDGE;
    }
    if (y < top) {
	code |= CLIP_TOP_EDGE;
    } else if (y > bottom) {
	code |= CLIP_BOTTOM_EDGE;
    }
	LOG_EXIT
    return code;
}

static int clipLine(SDL_Surface * dst, Sint16 * x1, Sint16 * y1, Sint16 * x2, Sint16 * y2)
{
	char *CLASS = "None";
	static char *functionName="clipLine";
	LOG_ENTER 
    Sint16 left, right, top, bottom;
    int code1, code2;
    int draw = 0;
    Sint16 swaptmp;
    float m;

    /*
     * Get clipping boundary 
     */
    left = dst->clip_rect.x;
    right = dst->clip_rect.x + dst->clip_rect.w - 1;
    top = dst->clip_rect.y;
    bottom = dst->clip_rect.y + dst->clip_rect.h - 1;

    while (1) {
	code1 = clipEncode(*x1, *y1, left, top, right, bottom);
	code2 = clipEncode(*x2, *y2, left, top, right, bottom);
	if (CLIP_ACCEPT(code1, code2)) {
	    draw = 1;
	    break;
	} else if (CLIP_REJECT(code1, code2))
	    break;
	else {
	    if (CLIP_INSIDE(code1)) {
		swaptmp = *x2;
		*x2 = *x1;
		*x1 = swaptmp;
		swaptmp = *y2;
		*y2 = *y1;
		*y1 = swaptmp;
		swaptmp = code2;
		code2 = code1;
		code1 = swaptmp;
	    }
	    if (*x2 != *x1) {
		m = (*y2 - *y1) / (float) (*x2 - *x1);
	    } else {
		m = 1.0f;
	    }
	    if (code1 & CLIP_LEFT_EDGE) {
		*y1 += (Sint16) ((left - *x1) * m);
		*x1 = left;
	    } else if (code1 & CLIP_RIGHT_EDGE) {
		*y1 += (Sint16) ((right - *x1) * m);
		*x1 = right;
	    } else if (code1 & CLIP_BOTTOM_EDGE) {
		if (*x2 != *x1) {
		    *x1 += (Sint16) ((bottom - *y1) / m);
		}
		*y1 = bottom;
	    } else if (code1 & CLIP_TOP_EDGE) {
		if (*x2 != *x1) {
		    *x1 += (Sint16) ((top - *y1) / m);
		}
		*y1 = top;
	    }
	}
    }

	LOG_EXIT
    return draw;
}


#define clip_xmin(surface) surface->clip_rect.x
#define clip_xmax(surface) surface->clip_rect.x+surface->clip_rect.w-1
#define clip_ymin(surface) surface->clip_rect.y
#define clip_ymax(surface) surface->clip_rect.y+surface->clip_rect.h-1

int _putPixelAlpha(SDL_Surface * surface, Sint16 x, Sint16 y, Uint32 color, Uint8 alpha)
{
	char *CLASS = "None";
	static char *functionName="_putPixelAlpha";
	LOG_ENTER 

	Uint32 Rmask = surface->format->Rmask, Gmask =
	surface->format->Gmask, Bmask = surface->format->Bmask, Amask = surface->format->Amask;
    Uint32 R, G, B, A = 0;

    if (x >= clip_xmin(surface) && x <= clip_xmax(surface)
	&& y >= clip_ymin(surface) && y <= clip_ymax(surface)) {

	switch (surface->format->BytesPerPixel) {
	case 1:{		/* Assuming 8-bpp */
		if (alpha == 255) {
		    *((Uint8 *) surface->pixels + y * surface->pitch + x) = color;
		} else {
		    Uint8 *pixel = (Uint8 *) surface->pixels + y * surface->pitch + x;

		    Uint8 dR = surface->format->palette->colors[*pixel].r;
		    Uint8 dG = surface->format->palette->colors[*pixel].g;
		    Uint8 dB = surface->format->palette->colors[*pixel].b;
		    Uint8 sR = surface->format->palette->colors[color].r;
		    Uint8 sG = surface->format->palette->colors[color].g;
		    Uint8 sB = surface->format->palette->colors[color].b;

		    dR = dR + ((sR - dR) * alpha >> 8);
		    dG = dG + ((sG - dG) * alpha >> 8);
		    dB = dB + ((sB - dB) * alpha >> 8);

		    *pixel = SDL_MapRGB(surface->format, dR, dG, dB);
		}
	    }
	    break;

	case 2:{		/* Probably 15-bpp or 16-bpp */
		if (alpha == 255) {
		    *((Uint16 *) surface->pixels + y * surface->pitch / 2 + x) = color;
		} else {
		    Uint16 *pixel = (Uint16 *) surface->pixels + y * surface->pitch / 2 + x;
		    Uint32 dc = *pixel;

		    R = ((dc & Rmask) + (((color & Rmask) - (dc & Rmask)) * alpha >> 8)) & Rmask;
		    G = ((dc & Gmask) + (((color & Gmask) - (dc & Gmask)) * alpha >> 8)) & Gmask;
		    B = ((dc & Bmask) + (((color & Bmask) - (dc & Bmask)) * alpha >> 8)) & Bmask;
		    if (Amask)
			A = ((dc & Amask) + (((color & Amask) - (dc & Amask)) * alpha >> 8)) & Amask;

		    *pixel = R | G | B | A;
		}
	    }
	    break;

	case 3:{		/* Slow 24-bpp mode, usually not used */
		Uint8 *pix = (Uint8 *) surface->pixels + y * surface->pitch + x * 3;
		Uint8 rshift8 = surface->format->Rshift / 8;
		Uint8 gshift8 = surface->format->Gshift / 8;
		Uint8 bshift8 = surface->format->Bshift / 8;
		Uint8 ashift8 = surface->format->Ashift / 8;


		if (alpha == 255) {
		    *(pix + rshift8) = color >> surface->format->Rshift;
		    *(pix + gshift8) = color >> surface->format->Gshift;
		    *(pix + bshift8) = color >> surface->format->Bshift;
		    *(pix + ashift8) = color >> surface->format->Ashift;
		} else {
		    Uint8 dR, dG, dB, dA = 0;
		    Uint8 sR, sG, sB, sA = 0;

		    pix = (Uint8 *) surface->pixels + y * surface->pitch + x * 3;

		    dR = *((pix) + rshift8);
		    dG = *((pix) + gshift8);
		    dB = *((pix) + bshift8);
		    dA = *((pix) + ashift8);

		    sR = (color >> surface->format->Rshift) & 0xff;
		    sG = (color >> surface->format->Gshift) & 0xff;
		    sB = (color >> surface->format->Bshift) & 0xff;
		    sA = (color >> surface->format->Ashift) & 0xff;

		    dR = dR + ((sR - dR) * alpha >> 8);
		    dG = dG + ((sG - dG) * alpha >> 8);
		    dB = dB + ((sB - dB) * alpha >> 8);
		    dA = dA + ((sA - dA) * alpha >> 8);

		    *((pix) + rshift8) = dR;
		    *((pix) + gshift8) = dG;
		    *((pix) + bshift8) = dB;
		    *((pix) + ashift8) = dA;
		}
	    }
	    break;

	case 4:{		/* Probably 32-bpp */
		if (alpha == 255) {
		    *((Uint32 *) surface->pixels + y * surface->pitch / 4 + x) = color;
		} else {
		    Uint32 *pixel = (Uint32 *) surface->pixels + y * surface->pitch / 4 + x;
		    Uint32 dc = *pixel;

		    R = ((dc & Rmask) + (((color & Rmask) - (dc & Rmask)) * alpha >> 8)) & Rmask;
		    G = ((dc & Gmask) + (((color & Gmask) - (dc & Gmask)) * alpha >> 8)) & Gmask;
		    B = ((dc & Bmask) + (((color & Bmask) - (dc & Bmask)) * alpha >> 8)) & Bmask;
		    if (Amask)
			A = ((dc & Amask) + (((color & Amask) - (dc & Amask)) * alpha >> 8)) & Amask;

		    *pixel = R | G | B | A;
		}
	    }
	    break;
	}
    }

	LOG_EXIT
    return (0);
}

/* ----- Pixel - fast, no blending, no locking, clipping */
int pixelColorNolock(SDL_Surface * dst, Sint16 x, Sint16 y, Uint32 color)
{
	char *CLASS = "None";
	static char *functionName="pixelColorNolock";
	LOG_ENTER 
    Uint8 alpha;
    Uint32 mcolor;
    int result = 0;

    /*
     * Setup color 
     */
    alpha = color & 0x000000ff;
    mcolor =
	SDL_MapRGBA(dst->format, (color & 0xff000000) >> 24,
		    (color & 0x00ff0000) >> 16, (color & 0x0000ff00) >> 8, alpha);

    /*
     * Draw 
     */
    result = _putPixelAlpha(dst, x, y, mcolor, alpha);

	LOG_EXIT
    return (result);
}


int SDLMain::line(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint32 color)
{
	static char *functionName="line";
	LOG_ENTER 
    int pixx, pixy;
    int x, y;
    int dx, dy;
    int ax, ay;
    int sx, sy;
    int swaptmp;
    Uint8 *pixel;
    Uint8 *colorptr;

    /*
     * Clip line and test if we have to draw 
     */
    if (!(clipLine(dst, &x1, &y1, &x2, &y2))) 
	{
		LOG_EXIT
		return (0);
    }

    /*
     * Variable setup 
     */
    dx = x2 - x1;
    dy = y2 - y1;
    sx = (dx >= 0) ? 1 : -1;
    sy = (dy >= 0) ? 1 : -1;

    /* Lock surface */
    if (SDL_MUSTLOCK(dst)) 
	{
		if (SDL_LockSurface(dst) < 0) 
		{
			LOG_EXIT
			return (-1);
		}
    }

    /*
     * Check for alpha blending 
     */
    if ((color & 255) == 255) 
	{

		/*
		 * No alpha blending - use fast pixel routines 
		 */

		/*
		 * Setup color 
		 */
		colorptr = (Uint8 *) & color;
		if (SDL_BYTEORDER == SDL_BIG_ENDIAN) 
		{
			color = SDL_MapRGBA(dst->format, colorptr[0], colorptr[1], colorptr[2], colorptr[3]);
		} 
		else 
		{
			color = SDL_MapRGBA(dst->format, colorptr[3], colorptr[2], colorptr[1], colorptr[0]);
		}

		/*
		 * More variable setup 
		 */
		dx = sx * dx + 1;
		dy = sy * dy + 1;
		pixx = dst->format->BytesPerPixel;
		pixy = dst->pitch;
		pixel = ((Uint8 *) dst->pixels) + pixx * (int) x1 + pixy * (int) y1;
		pixx *= sx;
		pixy *= sy;
		if (dx < dy) 
		{
			swaptmp = dx;
			dx = dy;
			dy = swaptmp;
			swaptmp = pixx;
			pixx = pixy;
			pixy = swaptmp;
		}

		/*
		 * Draw 
		 */
		x = 0;
		y = 0;
		switch (dst->format->BytesPerPixel) 
		{
			case 1:
				for (; x < dx; x++, pixel += pixx) 
				{
					*pixel = color;
					y += dy;
					if (y >= dx) 
					{
						y -= dx;
						pixel += pixy;
					}
				}
				break;
			case 2:
				for (; x < dx; x++, pixel += pixx) 
				{
					*(Uint16 *) pixel = color;
					y += dy;
					if (y >= dx) 
					{
						y -= dx;
						pixel += pixy;
					}
				}
				break;
			case 3:
				for (; x < dx; x++, pixel += pixx) 
				{
					if (SDL_BYTEORDER == SDL_BIG_ENDIAN) 
					{
						pixel[0] = (color >> 16) & 0xff;
						pixel[1] = (color >> 8) & 0xff;
						pixel[2] = color & 0xff;
					} 
					else 
					{
						pixel[0] = color & 0xff;
						pixel[1] = (color >> 8) & 0xff;
						pixel[2] = (color >> 16) & 0xff;
					}
					y += dy;
					if (y >= dx) 
					{
						y -= dx;
						pixel += pixy;
					}
				}
				break;
			default:		/* case 4 */
				for (; x < dx; x++, pixel += pixx) 
				{
					*(Uint32 *) pixel = color;
					y += dy;
					if (y >= dx) 
					{
						y -= dx;
						pixel += pixy;
					}
				}
				break;
			}

		} 
		else 
		{

		/*
		 * Alpha blending required - use single-pixel blits 
		 */
		ax = (((dx)<0) ? -(dx) : (dx)) << 1;
		ay = (((dy)<0) ? -(dy) : (dy)) << 1;
		x = x1;
		y = y1;
		if (ax > ay) 
		{
			int d = ay - (ax >> 1);
			while (x != x2) 
			{
				pixelColorNolock (dst, x, y, color);
				if (d > 0 || (d == 0 && sx == 1)) 
				{
					y += sy;
					d -= ax;
				}
				x += sx;
				d += ay;
			}
		} 
		else 
		{
			int d = ax - (ay >> 1);

			while (y != y2) 
			{
				pixelColorNolock (dst, x, y, color);
				if (d > 0 || ((d == 0) && (sy == 1))) 
				{
					x += sx;
					d -= ay;
				}
				y += sy;
				d += ax;
			}
		}
		pixelColorNolock (dst, x, y, color);

    }

    /* Unlock surface */
    if (SDL_MUSTLOCK(dst)) 
	{
		SDL_UnlockSurface(dst);
    }
	LOG_EXIT
    return (0);
}

// this function wraps freeing of SDL surfaces
// since data is held in static classes
// some surfaces are tried to be freed after quiting SDL
// especially on hardware surfaces this produces errros
// the hard way is -> don't free 'em if SDL is not initialized anymore!
void  SDLMain::freeSurface(SDL_Surface *surface)
{
	if (surface == 0)
	{
		return;
	}
	if (!mSDLInit)
	{
		if (surface->hwdata)
		{
			return;
		}
	}
	SDL_FreeSurface(surface);
}

void SDLMain::DrawRoundRect(SDL_Surface * screen, SDL_Rect *rect, Uint32 color, int corner, int thickness)
{
	DrawRoundRect(screen, rect->x, rect->y, rect->w, rect->h, color, corner, thickness);
}

//! Draws a rectangle with rounded corners (based on a routine
//! from SDL_draw library)
void SDLMain::DrawRoundRect(SDL_Surface * screen, int x0, int y0, int w, int h, Uint32 color, int corner, int thickness, bool north, bool south, bool east, bool west)
{
    int dx, dy;
    int Xcenter, Ycenter, X2center, Y2center;
    int d, x = 0;
    int diagonalInc, rightInc = 6;

    if (w == 0 || h == 0)
	{
        return;
	}

    if (corner != 0)
    {
        d = w < h ? w : h;
        corner--;
        if(corner != 0 && corner + 2 >= d)
        {
            if(corner + 2 == d)
                corner--;
            else
                corner = 0;
        }
    }

    d = 3 - (corner << 1);
    diagonalInc = 10 - (corner << 2);

	// Rectangles
	dx = w - (corner << 1);
	dy = h - (corner << 1);
	Xcenter = x0 + corner;
	Ycenter = y0 + corner;

	// Centers
	X2center = Xcenter + dx - 1;
	Y2center = Ycenter + dy - 1;
	SDL_Rect area;

	if (north)
	{
		area.x=x0+corner;
		area.y=y0;
		area.w=dx;
		area.h=thickness;
		if (dx >0)
			SDL_FillRect( screen, &area, color); 
	}
	if (south)
	{
		area.x=x0+corner;
		area.y=y0+h-thickness;
		area.w=dx;
		area.h=thickness;
		if (dx >0)
			SDL_FillRect( screen, &area, color); 
	}
	if (west)
	{
		area.x=x0;
		area.y=y0+corner;
		area.w=thickness;
		area.h=dy;
		if (dy >0)
			SDL_FillRect( screen, &area, color); 
	}
	if (east)
	{
		area.x=x0+w-thickness;
		area.y=y0+corner;
		area.w=thickness;
		area.h=dy;
		if (dy >0)
			SDL_FillRect( screen, &area, color); 
	}

	
	color = color*256+255;
	for (int t = 0; t < thickness; t++)
	{
		while(x <= corner*2)
		{
			// north
			if (north)
			{
				line(screen, Xcenter - corner,				Ycenter-x, 
							 Xcenter - corner + thickness,	Ycenter-x, color);
				line(screen, Xcenter - x,					Ycenter-corner, 
							 Xcenter - x,					Ycenter-corner + thickness, color);
			}
			// west
			if (west)
			{
				line(screen, Xcenter - corner,				Y2center + x, 
							 Xcenter - corner + thickness,	Y2center + x, color);
				line(screen, Xcenter - x,					Y2center+corner - thickness, 
							 Xcenter - x,					Y2center+corner, color);
			}

			// east
			if (east)
			{
				line(screen, X2center + corner - thickness,	Ycenter - x, 
							 X2center + corner,				Ycenter - x, color);
				line(screen, X2center + x,					Ycenter - corner, 
							 X2center + x,					Ycenter - corner + thickness, color);
			}

			// south
			if (south)
			{
				line(screen, X2center + x,					Y2center + corner - thickness, 
							 X2center + x,					Y2center + corner, color);
				line(screen, X2center + corner - thickness,	Y2center + x, 
							 X2center + corner,				Y2center + x, color);
			}
			if(d >= 0)
			{
				d += diagonalInc;
				diagonalInc += 8;
				corner--;
			}
			else
			{
				d += rightInc;
				diagonalInc += 4;
			}
			rightInc += 4;
			x++;
		}
	}
}

// Draws a rectangle with rounded corners (based on a routine
// from SDL_draw library)
void SDLMain::DrawRoundRectFill(SDL_Surface * screen, int x0, int y0, int w, int h, Uint32 color, int corner)
{
    int dx, dy;
    int Xcenter, Ycenter, X2center, Y2center;
    int d, x = 0;
    int diagonalInc, rightInc = 6;

    if(w == 0 || h == 0)
        return;

    if(corner != 0)
    {
        d = w < h ? w : h;
        corner--;
        if(corner != 0 && corner + 2 >= d)
        {
            if(corner + 2 == d)
                corner--;
            else
                corner = 0;
        }
    }

    d = 3 - (corner << 1);
    diagonalInc = 10 - (corner << 2);

    // Rectangles
    dx = w - (corner << 1);
    Xcenter = x0 + corner;
    dy = h - (corner << 1);
    Ycenter = y0 + corner;

    // Centers
    X2center = Xcenter + dx - 1;
    Y2center = Ycenter + dy - 1;

	SDL_Rect area;
	area.x = x0; 
	area.y = Ycenter;
	area.w = w; 
	area.h = dy;
	
	SDL_FillRect(screen, &area, color);
	color = color *256 + 255;
	
	// because of = there are some lines drawn twice
	while (x <= corner) 
	{

		line(screen, Xcenter - x,	Ycenter - corner, 
					 X2center + x,	Ycenter - corner, color);
		line(screen, Xcenter - x,	Y2center + corner, 
					 X2center + x,	Y2center + corner, color);
		line(screen, Xcenter - corner,		Ycenter - x, 
					 X2center + corner,		Ycenter - x, color);
		line(screen, Xcenter - corner,		Y2center + x, 
					 X2center + corner,		Y2center + x, color);
		

		if (d >= 0) 
		{
			d += diagonalInc;
			diagonalInc += 8;
			--corner;
		} 
		else 
		{
			d += rightInc;
			diagonalInc += 4;
		}
		rightInc += 4;
		++x;
	}/*while*/
}

#define R(color) (((color)&0xff0000)>>16)
#define G(color) (((color)&0x00ff00)>>8)
#define B(color) (((color)&0x0000ff))
// following to methods stolen from aedGUI!
void SDLMain::horizgradient(SDL_Surface * surf, SDL_Rect & gradRect, int c1, int c2, Uint8 alpha)
{
    int x, width, height;
    Uint8 r1=R(c1), g1 = G(c1), b1 = B(c1);
    Uint8 r2=R(c2), g2 = G(c2), b2 = B(c2);
    Uint8 r, g, b;
    SDL_Rect dest;
    Uint32 pixelcolor;

    width = gradRect.w;
    height = gradRect.h;


    // Some optimization:
    if ((r1 == r2) && (g1 == g2) && (b1 == b2))
    {
        SDL_FillRect(surf, NULL, SDL_MapRGBA(surf->format, r1, g1, b1, alpha));
    }
    else
    {
        for(x = 0; x < width; x++)
        {
            r = (r2 * x / width) + (r1 * (width - x) / width);
            g = (g2 * x / width) + (g1 * (width - x) / width);
            b = (b2 * x / width) + (b1 * (width - x) / width);

            dest.x = gradRect.x + x;
            dest.y = gradRect.y;
            dest.w = 1;
            dest.h = height;

            pixelcolor = SDL_MapRGBA(surf->format, r, g, b, alpha);
            SDL_FillRect(surf, &dest, pixelcolor);
        }
    }
}

void SDLMain::vertgradient(SDL_Surface * surf, SDL_Rect & gradRect, int c1, int c2, Uint8 alpha)
{
    int y, width, height;
    Uint8 r1=R(c1), g1 = G(c1), b1 = B(c1);
    Uint8 r2=R(c2), g2 = G(c2), b2 = B(c2);
    Uint8 r, g, b;
    SDL_Rect dest;
    Uint32 pixelcolor;

    width = gradRect.w;
    height = gradRect.h;


    // Some optimization:
    if ((r1 == r2) && (g1 == g2) && (b1 == b2))
    {
        SDL_FillRect(surf, NULL, SDL_MapRGBA(surf->format, r1, g1, b1, alpha));
    }
    else
    {
        for(y = 0; y < height; y++)
        {
            r = (r2 * y / height) + (r1 * (height - y) / height);
            g = (g2 * y / height) + (g1 * (height - y) / height);
            b = (b2 * y / height) + (b1 * (height - y) / height);

            dest.x = gradRect.x;
            dest.y = gradRect.y + y;
            dest.w = width;
            dest.h = 1;

            pixelcolor = SDL_MapRGBA(surf->format, r, g, b, alpha);
            SDL_FillRect(surf, &dest, pixelcolor);
        }
    }
}
