#ifdef WIN32
#pragma warning(disable : 4786 )
#endif

#include "CSGrafikElement.h"
#include "CSDesktop.h"
#include "CSWindow.h"
#include "CSBorder.h"
#include "CSFont.h"
#include "CSArea.h"
#include "CSLayoutManager.h"

const char *CSGrafikElement::CLASS = "CSGrafikElement";
long CSGrafikElement::mInLayout=0;

CSGrafikElements CSGrafikElement::mLostElements;

CSGrafikElement::CSGrafikElement(int height, int width)
{
	static char *functionName="CSGrafikElement";
	LOG_ENTER 
	CSLAF *laf = CSLAF::getCurrentLAF(); // make sure laf is loaded/initialized at least once
	CSLAF::addElement(this);
	init(height, width);
	addMessageListener(this, GUI_MESSAGE);
	mTooltipText = "";
	LOG_EXIT
}


/** Destructor, destroys also all child elements (recursivly) and layoutmanager
*/
CSGrafikElement::~CSGrafikElement(void)
{
	static char *functionName="~CSGrafikElement";
	LOG_ENTER 
	removeMessageListener(this);
	CSGrafikElement *parent = getParent();

	CSDesktop *desktop = getDesktop();
	if (desktop != 0)
	{
		desktop->correctElementHandling(this);
	}
	
	if (parent) 
	{
		parent->removeMainElement(this);
	}

	// Main element must be reomved first, since the "normal" remove element
	// is mapped to the mainElement (if available)
	// a "normal" delete removes itself from the parent
	// if that remove is done on a parent with a mainElement, the element tries
	// to remove itself from the mainElement instead of the parent...
	if (mMainElement != this)
	{
		CSGrafikElements::iterator iter = mElements.begin();
		while (iter != mElements.end())
		{
			CSGrafikElement *element = *iter;
			if (element == mMainElement)
			{
				mElements.erase(iter);
				delete (element);
				break;
			}
			iter++;
		}
		mMainElement = this;
	}

	CSGrafikElements::iterator iter = mElements.begin();
	while (iter != mElements.end())
	{
		CSGrafikElement *element = *iter;
		mElements.erase(iter);
		delete (element);
		iter = mElements.begin();
	}
	if (mBorder)
	{
		delete mBorder;
		mBorder = 0;
	}
	if (mArea)
	{
		delete mArea;
		mArea = 0;
	}
	if (mLayoutManager != 0)
	{
		delete mLayoutManager;
		mLayoutManager = 0;
	}
	CSLAF::removeElement(this);
	addLostElement(this);

	LOG_EXIT
}

void CSGrafikElement::init(int height, int width)
{
	static char *functionName="init";
	LOG_ENTER 
	mMainElement = this;
	CSLAF *laf = CSLAF::getCurrentLAF(); // make sure laf is loaded/initialized at least once
	MESSAGE_FOCUS_GAINED.setSubtype(FOCUS_GAINED_MESSAGE);
	MESSAGE_FOCUS_LOST.setSubtype(FOCUS_LOST_MESSAGE);
	MESSAGE_KEY_PRESSED.setSubtype(KEY_PRESSED_MESSAGE);
	MESSAGE_KEY_RELEASED.setSubtype(KEY_RELEASED_MESSAGE);
	MESSAGE_MOUSE_PRESSED.setSubtype(MOUSE_BUTTON_PRESSED_MESSAGE);
	MESSAGE_MOUSE_RELEASED.setSubtype(MOUSE_BUTTON_RELEASED_MESSAGE);
	MESSAGE_MOUSE_MOTION.setSubtype(MOUSE_MOTION_MESSAGE);
	MESSAGE_MOUSE_MOTION_LOST.setSubtype(MOUSE_MOTION_LOST_MESSAGE);
	mFrontElement = false;
	mElementBackadded = false;
	mBorderState = BORDER_STATE_DEFAULT;
	mActionId = 0;
	mActionObject = 0;
	mParent = 0;
	mFont = 0;
	mBorder = 0;
	mArea = 0;
	mLayoutManager = 0;
	mLayoutChanged = true; // layout at least once!
	mLayoutAllowed = true;
	mParentUpdateAllowed = true;
	setLayoutManager(new CSLayoutManagerXY());
	mX = 0;
	mY = 0;
	mHeight = height;
	mWidth = width;
	mMinWidth = 0;
	mMinHeight = 0;
	mChangeType = 0;
	if (mHeight < getBorder()->getTotalHeight() + mInset.getTotalHeight()) 
	{
		mHeight = getBorder()->getTotalHeight() + mInset.getTotalHeight();
	}
	if (mWidth < getBorder()->getTotalWidth() + mInset.getTotalWidth()) 
	{
		mWidth = getBorder()->getTotalWidth() + mInset.getTotalWidth();
	}

	mViewportX = getBorder()->getSizeWest() + mInset.getSizeWest();
	mViewportY = getBorder()->getSizeNorth() + mInset.getSizeNorth();
	mViewportHeight = getHeight() - getBorder()->getTotalHeight() - mInset.getTotalHeight();
	mViewportWidth = getWidth() - getBorder()->getTotalWidth() - mInset.getTotalWidth();

	if (mViewportHeight < 0) mViewportHeight = 0;
	if (mViewportWidth < 0) mViewportWidth = 0;

	mViewportOffsetX = 0;
	mViewportOffsetY = 0;

	mStyle = FONT_STYLE_NONE;
	mEnabled = true;
	mActivated = false;
	mActive = false;
	mVisible = true;
	mModal = false;
	mFocused = false;
	mFocusable = true;
	mHeightSet = false;
	mWidthSet = false;

	mBackGroundColorSet = false;
	mTextColorSet = false;
	mHorizontalElementSpacingSet = false;
	mVerticalElementSpacingSet = false;

	changeLAFDefaults();

	mHorizontalElementSpacing = laf->getHorizontalElementSpacing(getGUIType());
	mVerticalElementSpacing = laf->getVerticalElementSpacing(getGUIType());
	cursorDefault();
	layoutSetup(); //combobox an menus correct if this is here
	LOG_EXIT
}

void CSGrafikElement::changeLAFDefaults()
{
	CSLAF *laf = CSLAF::getCurrentLAF(); // make sure laf is loaded/initialized at least once
	if (mEnabled)
	{
		if (!mBackGroundColorSet)
			mBackgroundColor = laf->getBackgroundColorEnabled(getGUIType());
		if (!mTextColorSet)
			mTextColor = laf->getTextColorEnabled(getGUIType());
	}
	else
	{
		if (!mBackGroundColorSet)
			mBackgroundColor = laf->getBackgroundColorDisabled(getGUIType());
		if (!mTextColorSet)
			mTextColor = laf->getTextColorDisabled(getGUIType());
	}

}

//void CSGrafikElement::setPosition(int x, int y, bool centerX, bool centerY)
void CSGrafikElement::setPosition(int x, int y)
{
	static char *functionName="setPosition";
	if ((mX != x) || (mY != y))
	{
		if (mParent)
		{
			// no position information to be set to non xy
			if (mParent->getLayoutManager()->getType().compare(CSLayoutManagerXY::CLASS) != 0)
			{
				return;
			}
		}

		mX = x;
		mY = y;
		getLayoutDataInternal()->setX(mX);
		getLayoutDataInternal()->setY(mY);
		layoutChanged(true, false);
	}
}

void CSGrafikElement::setHeight(int h) 
{
	static char *functionName="setHeight";
	LOG_ENTER 
	if (h == -1)
	{
		mHeightSet = false;
	}
	else
	{
		if (h < mMinHeight) 
		{
			h = mMinHeight;
		}
		if (mHeight != h)
		{
			mHeightSet = true;
			mHeight = h;
			layoutChanged();
		}
	}
	LOG_EXIT
}

void CSGrafikElement::setWidth(int w) 
{
	static char *functionName="setWidth";
	LOG_ENTER 
	if (w == -1)
	{
		mWidthSet = false;
	}
	else
	{
		if (w < mMinWidth) 
			w = mMinWidth;
		if (mWidth != w)
		{
			mWidthSet = true;
			mWidth = w;
			layoutChanged();
		}
	}
	LOG_EXIT
}

void CSGrafikElement::paintBorder(SDL_Surface *destination, SDL_Rect *parentArea)
{
	SDL_Rect area = getElementArea();
	area.x += parentArea->x;
	area.y += parentArea->y;
	CSBorder *border = getBorder();
	if (border)
	{
		border->paint(destination, &area);
	}
}

void CSGrafikElement::paintAreaStandard(SDL_Surface *destination, SDL_Rect *parentArea)
{
	SDL_Rect elementArea = getElementArea();
	elementArea.x += parentArea->x;
	elementArea.y += parentArea->y;
	CSArea *area = getArea();
	if (area)
	{
		area->setColor(COLOR(destination, getBackgroundColor()));
		area->paint(destination, &elementArea);
	}
}

//! \todo the clipping and filling here in Graphic element, not where the children live!
//!
void CSGrafikElement::paintChildren(SDL_Surface *destination, SDL_Rect *parentViewport)
{
	static char *functionName="paintChildren";
	CSGrafikElement *modalDetected = 0;
	SDL_Rect saveRect;

	SDL_GetClipRect(destination, &saveRect);
	SDL_Rect thisViewportArea;
	SDL_Rect thisViewportAreaCopy;

	thisViewportArea.x = parentViewport->x + mViewportX + mX;
	thisViewportArea.y = parentViewport->y + mViewportY + mY;
	thisViewportArea.h = mViewportHeight;
	thisViewportArea.w = mViewportWidth;
	thisViewportAreaCopy = thisViewportArea;

	int offsetX = 0;
	int offsetY = 0;
	getViewportOffset(offsetX, offsetY);

	thisViewportArea.x -= offsetX;
	thisViewportArea.y -= offsetY;


	SDLMain::adjustClipping(thisViewportAreaCopy, saveRect);
	SDL_SetClipRect(destination, &thisViewportAreaCopy);

	CSGrafikElements::iterator iter = mElements.begin();
	while (iter != mElements.end())
	{
		CSGrafikElement *element = (*iter);
		if (element->getVisible())
		{
			if (!element->getModal())
			{
				element->paintArea(destination, &thisViewportArea);
				element->paint(destination, &thisViewportArea);
				element->paintBorder(destination, &thisViewportArea);
			}
			else 
			{
				modalDetected = *iter;
			}
		}
		iter++;
	}

	SDL_SetClipRect(destination, &saveRect);

	if (modalDetected)
	{
		if (getDesktop())
		{
			getDesktop()->runModalDesktop(modalDetected, destination);
		}
	}

}

void CSGrafikElement::toFront(CSGrafikElement *element)
{
	CSGrafikElements::iterator iter = mElements.begin();
	bool elementFound = false;
	while (iter != mElements.end())
	{
		if ((*iter) == element)
		{
			mElements.erase(iter);
			elementFound = true;
			break;
		}
		iter++;
	}	
	if (elementFound == true)
	{
		mElements.push_back(element);
		mElementBackadded = true;
	}
}

void CSGrafikElement::setParent(CSGrafikElement *parent) 
{
	static char *functionName="setParent";
	LOG_ENTER 
	mParent = parent;
	LOG_EXIT
}

/** 
	no clipping
	// raw (no border, inset respected)
*/
void CSGrafikElement::line(SDL_Surface *destination, SDL_Rect *parentViewport, int x1, int y1, int x2, int y2, Uint32 color)
{
	static char *functionName="line";
	// colors are alpha colors!!!
	SDLMain::line(destination, parentViewport->x + x1, parentViewport->y + y1, parentViewport->x + x2, parentViewport->y + y2, (color<<8)+255);
}
 
/** 
	no clipping
*/
void CSGrafikElement::putString(SDL_Surface *destination, SDL_Rect *parentViewport, int x1, int y1, const std::string &string)
{
	static char *functionName="putString";
	getFont()->putString(destination, parentViewport->x + x1, parentViewport->y + y1, mTextColor ,string, mStyle);
}

/** 
	no clipping
	// raw (no border, inset respected)
*/
void CSGrafikElement::putString(SDL_Surface *destination, SDL_Rect *parentViewport, int x1, int y1, int color, const std::string &string)
{
	static char *functionName="putString";
	getFont()->putString(destination, parentViewport->x + x1, parentViewport->y + y1, color, string, mStyle);
}

// x and y in screen coordinates
bool CSGrafikElement::isIn(int x, int y, SDL_Rect *parentElementArea)
{
	static char *functionName="isIn";
	if (!getVisible())
	{
		return false;
	}
	int startX, startY;
	int endX, endY;
	
	if (parentElementArea)
	{
		startX = parentElementArea->x+mX;
		startY = parentElementArea->y+mY;
		endY = mHeight + parentElementArea->y+mY;
		if (mHeight>parentElementArea->h-mY)
		{
			// die 4 sind sowas wie 2* abstand zwischen zwei Panels (border layout), ist schlecht so - aber keinen bock es schoen zu machen zur zeit
			endY = parentElementArea->y+parentElementArea->h-4;
		}

		endX = mWidth + parentElementArea->x+mX;
		if (mWidth>parentElementArea->w-mX)
		{
			// die 4 sind sowas wie 2* abstand zwischen zwei Panels (border layout), ist schlecht so - aber keinen bock es schoen zu machen zur zeit
			endX = parentElementArea->w + parentElementArea->x-4;
		}
	}
	else
	{
		startX = mX;
		startY = mY;
		endY = mHeight + mY;
		endX = mWidth + mX;
	}


	if ((x>=startX) && (x<endX) && (y>=startY) && (y<endY))
	{
		bool b = !isTransparent(x, y, parentElementArea);
		return b;
	}

	return false;
}

//! x and y in screen coordinates
//! returnes coordinates in (int &receiverX, int &receiverY) of receiver component within it's parent!
CSGrafikElement *CSGrafikElement::getElement(int x, int y, SDL_Rect *parentElementArea, int &receiverX, int &receiverY)
{
	static char *functionName="getElement";
	static int ypos=0;
/*
ypos+=16;
getFont()->setSolid(true);
std::string type = this->getType();
char str[100];
if (type.compare(CSWindow::CLASS) == 0)
{
	getFont()->putString(SDLMain::getScreen(), 5, 280, "Window searched!");
}
sprintf(str, "Screen: %i, %i (x, y)     ",x,y);
getFont()->putString(SDLMain::getScreen(), 5, 300, str);
*/

	CSGrafikElements::reverse_iterator riter = mElements.rbegin();
	SDL_Rect thisElementArea = mElementArea;
	thisElementArea.x += parentElementArea->x;
	thisElementArea.y += parentElementArea->y;
	SDL_Rect thisElementAreaViewported = thisElementArea;


	int offsetX = 0;
	int offsetY = 0;
	getViewportOffset(offsetX, offsetY);

	thisElementAreaViewported.x += mViewportX - offsetX;
	thisElementAreaViewported.y += mViewportY - offsetY;


/*
	// buggy?
	if (thisElementAreaViewported.h>parentElementArea->h+mElementArea.x+mViewportX)
	{
		thisElementAreaViewported.h = parentElementArea->h-mElementArea.x-mViewportX;
	}
	if (thisElementAreaViewported.w>parentElementArea->w+mElementArea.y+mViewportY)
	{
		thisElementAreaViewported.w = parentElementArea->w-mElementArea.y-mViewportY;
	}
*/

/*
htmlpage must be sized down in order to be updated correctly when resizing
sonst bleiben die worte immer auf der grten size und der text passt sich nicht an
*/
	riter = mElements.rbegin();
	while (riter != mElements.rend())
	{
		CSGrafikElement *element = (*riter);
		bool isInElement;
		isInElement = element->isIn(x, y, &thisElementAreaViewported);
		if (isInElement)
		{

			
			CSGrafikElement *element = (*riter)->getElement(x, y, &thisElementAreaViewported, receiverX, receiverY);
/*

sprintf(str, "%s: %i, %i (w:%i:, h:%i, sx:%i, sy:%i)     ",getType().c_str(),x-thisElementArea.x,y-thisElementArea.y,getWidth(), getHeight(), getX(), getY());
getFont()->putString(SDLMain::getScreen(), 5, 300+ypos, str);
ypos-=16;
*/
			return element;
		}
		riter++;
	}

	receiverX = x - thisElementArea.x;
	receiverY = y - thisElementArea.y;

/*
sprintf(str, "%s: %i, %i (w:%i:, h:%i, sx:%i, sy:%i)     ",getType().c_str(),x-thisElementArea.x,y-thisElementArea.y,getWidth(), getHeight(), getX(), getY());
getFont()->putString(SDLMain::getScreen(), 5, 300+ypos, str);
ypos-=16;
*/
	return this;
}

void CSGrafikElement::setFocusable(bool b) 
{
	static char *functionName="setFocusable";
	mFocusable = b;
	
	CSGrafikElements::iterator iter = mElements.begin();
	while (iter != mElements.end())
	{
		CSGrafikElement *test = *iter;
		test->setFocusable(b);
		iter++;
	}
}

void CSGrafikElement::cursorSizing()
{
	mCursorType = MOUSE_CURSOR_TYPE_SIZING;
}

void CSGrafikElement::cursorDefault()
{
	mCursorType = MOUSE_CURSOR_TYPE_DEFAULT;
}

CSMouseCursor *CSGrafikElement::getCursor() 
{
	CSLAF *laf = CSLAF::getCurrentLAF(); // make sure laf is loaded/initialized at least once
	if (mCursorType == MOUSE_CURSOR_TYPE_SIZING)
	{
		return laf->getMouseCursor(LAF_MOUSE_CURSOR_SIZING_DOWN_RIGHT);
	}

	return laf->getMouseCursor(LAF_MOUSE_CURSOR_DEFAULT);
}

int CSGrafikElement::getElementMaxX()
{
	int maxX = 0;
	CSGrafikElements::iterator iter = mElements.begin();
	while (iter != mElements.end())
	{
		CSGrafikElement *test = *iter;
		int x = test->getX() + test->getWidth();
		if (x > maxX)
		{
			maxX = x;
		}
		iter++;
	}
	return maxX;
}

int CSGrafikElement::getElementMaxY()
{
	int maxY = 0;
	CSGrafikElements::iterator iter = mElements.begin();
	while (iter != mElements.end())
	{
		CSGrafikElement *test = *iter;
		int y = test->getY() + test->getHeight();
		if (y > maxY)
		{
			maxY = y;
		}
		iter++;
	}
	return maxY;
}

CSWindow *CSGrafikElement::getParentWindow(void)
{
	static char *functionName="getParentWindow";
	CSGrafikElement *parent = this;
	while (parent != 0)
	{
		std::string type = parent->getType();
		if (type.compare(CSWindow::CLASS) == 0)
		{
			return (CSWindow *) parent;
		}
		parent = parent->getParent();
	}

	return 0;
}

CSDesktop *CSGrafikElement::getDesktop(void)
{
	static char *functionName="getDesktop";
	CSGrafikElement *parent = this;
	while (parent != 0)
	{
		std::string type = parent->getType();
		if (type.compare(CSDesktop::CLASS) == 0)
		{
			return (CSDesktop *) parent;
		}
		parent = parent->getParent();
	}
	return 0;
}

void CSGrafikElement::translateToDesktop(int &x, int &y)
{
	static char *functionName="translateToDesktop";
	CSGrafikElement *parent = this;
	int offsetX = 0;
	int offsetY = 0;
	getViewportOffset(offsetX, offsetY);
	x = x - offsetX;
	y = y - offsetY;
	while (parent != 0)
	{
		std::string type = parent->getType();
		if (type.compare(CSDesktop::CLASS) == 0)
		{
			break;
		}
		x = x + parent->getX();
		y = y + parent->getY();
		x = x + parent->getParent()->getViewportX();
		y = y + parent->getParent()->getViewportY();

		parent->getViewportOffset(offsetX, offsetY);
		x = x - offsetX;
		y = y - offsetY;

		parent = parent->getParent();
	}
}

//! is \a element a child of myself?
bool CSGrafikElement::isChild(CSGrafikElement *element)
{
	static char *functionName="isChild";
	while (element != 0)
	{
		if (element == this)
		{
			return true;
		}
		element = element->getParent();
	}
	return false;
}

//! a bottom up laf change, inner most elements allways first,
//! so outer lying elements get right sizing information from their children!
void CSGrafikElement::lafChanged(int id)
{
	if (mCurrentLAFId != id)
	{
		CSGrafikElements::iterator iter = mElements.begin();
		while (iter != mElements.end())
		{
			(*iter)->lafChanged(id);
			iter++;
		}
		rebuildElement();
		layoutChanged();
		mCurrentLAFId = id;
	}
}

/**  returns the given gui font (see CSLAF)
*/
CSFont *CSGrafikElement::getFont() 
{
	if (mFont) 
	{
		return mFont; 
	}
	return CSLAF::getCurrentLAF()->getFont(getGUIType());
}	

// there should be:
// called after this call
// cannot do it from inside, since
//! in destructor this class deletes the currently set area!
//! if set (and kept) from outside this should be taken into account
//! if outside wants to delete the border itself
//! setArea(0) should be called to prevent deleting here!
void CSGrafikElement::setArea(CSArea *area, bool deleteOldArea) 
{
	if (deleteOldArea)
	{
		if (mArea)
		{
			delete mArea;
			mArea = 0;
		}
	}
	mArea = area;
}

CSArea *CSGrafikElement::getArea() 
{
	if (!mArea)
	{
		return CSLAF::getCurrentLAF()->getArea(getGUIType());
	}
	return mArea;
}

// there should be:
//! in destructor this class deletes the currently set border!
//! if set (and kept) from outside this should be taken into account
//! if outside wants to delete the border itself
//! setBorder(0) should be called to prevent deleting here!
void CSGrafikElement::setBorder(CSBorder *border, bool deleteOldBorder) 
{
	if (deleteOldBorder)
	{
		if (mBorder)
		{
			delete mBorder;
			mBorder = 0;
		}
	}
	mBorder = border;
	layoutChanged();
}

CSBorder *CSGrafikElement::getBorder() 
{
	if (!mBorder)
	{
		return CSLAF::getCurrentLAF()->getBorder(getGUIType(), mBorderState);
	}
	return mBorder;
}

int CSGrafikElement::getBackgroundColor() 
{
	if (mBackGroundColorSet == true)
	{
		return mBackgroundColor;
	}
	if (getEnabled())
	{
		int color = -1;
		CSGrafikElement * element = this;
		while ((color == -1) && (element != 0))
		{
			color = CSLAF::getCurrentLAF()->getBackgroundColorEnabled(element->getGUIType());
			if (color != -1)
			{
				return color;
			}
			element = element->getParent();
		}
		return 255*256*256+255; // Magenta!
	}
	int color = -1;
	CSGrafikElement * element = this;
	while ((color == -1) && (element != 0))
	{
		color = CSLAF::getCurrentLAF()->getBackgroundColorDisabled(element->getGUIType());
		if (color != -1)
		{
			return color;
		}
		element = element->getParent();
	}
	return 255*256*256+255; // Magenta!
}

void CSGrafikElement::setBackgroundColor(int backgroundColor) 
{
	mBackGroundColorSet = true;
	mBackgroundColor = backgroundColor;
}

int CSGrafikElement::getTextColor() 
{
	if (mTextColorSet == true)
	{
		return mTextColor;
	}
	if (getEnabled())
	{
		int color = -1;
		CSGrafikElement * element = this;
		while ((color == -1) && (element != 0))
		{
			color = CSLAF::getCurrentLAF()->getTextColorEnabled(element->getGUIType());
			if (color != -1)
			{
				return color;
			}
			element = element->getParent();
		}
		return 255*256*256+255; // Magenta!
	}
	int color = -1;
	CSGrafikElement * element = this;
	while ((color == -1) && (element != 0))
	{
		color = CSLAF::getCurrentLAF()->getTextColorDisabled(element->getGUIType());
		if (color != -1)
		{
			return color;
		}
		element = element->getParent();
	}
	return 255*256*256+255; // Magenta!
}

void CSGrafikElement::setTextColor(int textColor) 
{
	mTextColorSet = true;
	mTextColor = textColor;
}

int CSGrafikElement::getVerticalElementSpacing() 
{
	if (mVerticalElementSpacingSet == true)
	{
		return mVerticalElementSpacing;
	}
	return CSLAF::getCurrentLAF()->getVerticalElementSpacing(getGUIType());
}

void CSGrafikElement::setVerticalElementSpacing(int verticalElementSpacing) 
{
	mVerticalElementSpacingSet = true;
	mVerticalElementSpacing = verticalElementSpacing;
	layoutChanged();
}

int CSGrafikElement::getHorizontalElementSpacing()
{
	if (mHorizontalElementSpacingSet == true)
	{
		return mHorizontalElementSpacing;
	}
	return CSLAF::getCurrentLAF()->getHorizontalElementSpacing(getGUIType());
}

void CSGrafikElement::setHorizontalElementSpacing(int horizontalElementSpacing) 
{
	mHorizontalElementSpacingSet = true;
	mHorizontalElementSpacing = horizontalElementSpacing;
	layoutChanged();
}

//! removed element is NOT freed!
CSGrafikElement *CSGrafikElement::removeElement(CSGrafikElement *element)
{
	static char *functionName="removeElement";
	if (mMainElement == this)
	{
		return removeMainElement(element);
	}
	CSGrafikElement *returnee = mMainElement->removeElement(element);
	layoutChanged();
	return returnee;
}

//! removed element is NOT freed!
CSGrafikElement *CSGrafikElement::removeMainElement(CSGrafikElement *element)
{
	static char *functionName="removeMainElement";
	LOG_ENTER 
	CSGrafikElement *removedElement = 0;
	CSGrafikElements::iterator iter = mElements.begin();
	while (iter != mElements.end())
	{
		CSGrafikElement *test = *iter;
		if (test == element)
		{
			removedElement = test;
			mElements.erase(iter);
			removeMessageListener(test, GUI_MESSAGE); 

			if (getDesktop())
			{
				getDesktop()->correctElementHandling(test);
			}

			test->setParent(0);
			test->mFrontElement = false;
			break;
		}
		iter++;
	}
	layoutChanged();
	LOG_EXIT
	return removedElement;
}

//! element will be freed by this
//! 
void CSGrafikElement::addElement(CSGrafikElement *element, int x, int y)
{
	static char *functionName="addElement";
	LOG_ENTER 

	// there can only be one parent! (not tested - the old one is simply removed!)
	if (mMainElement == this)
	{
		addMainElement(element, x, y);
	}
	else
	{
		mMainElement->addElement(element, x, y);
	}
	LOG_EXIT
}

//! element will be freed by this
//! 
void CSGrafikElement::addElement(CSGrafikElement *element, int position)
{
	static char *functionName="addElement";
	LOG_ENTER 

	// there can only be one parent! (not tested - the old one is simply removed!)
	if (mMainElement == this)
	{
		addMainElement(element, position);
	}
	else
	{
		mMainElement->addElement(element, position);
	}
	LOG_EXIT
}

void CSGrafikElement::addElement(CSGrafikElement *element)
{
	static char *functionName="addElement";
	LOG_ENTER 

	// there can only be one parent! (not tested - the old one is simply removed!)
	if (mMainElement == this)
	{
		addMainElement(element);
	}
	else
	{
		mMainElement->addElement(element);
	}
	LOG_EXIT
}

//! element will be freed by this
//! 
void CSGrafikElement::addMainElement(CSGrafikElement *element)
{
	static char *functionName="addElement";
	LOG_ENTER 
	// there can only be one parent! (not tested - the old one is simply removed!)
	if (element->getParent())
	{
		element->getParent()->removeMainElement(element);
	}
	element->setParent(this);
	mElements.push_back(element);
	mElementBackadded = true;

	CSLayoutData *layoutData = element->getLayoutDataInternal();
	if (!layoutData->getSpacingSet())
	{
		layoutData->setSpacing(getVerticalElementSpacing(), getHorizontalElementSpacing());
	}
	element->setLayoutData(*layoutData);
	element->layoutChanged();
	LOG_EXIT
}

//! element will be freed by this
//! 
void CSGrafikElement::addMainElement(CSGrafikElement *element, int position)
{
	static char *functionName="addElement";
	LOG_ENTER 
	// there can only be one parent! (not tested - the old one is simply removed!)
	if (element->getParent())
	{
		element->getParent()->removeMainElement(element);
	}
	element->setParent(this);
	mElements.push_back(element);
	mElementBackadded = true;

	CSLayoutData *layoutData = element->getLayoutDataInternal();
	if (!layoutData->getSpacingSet())
	{
		layoutData->setSpacing(getVerticalElementSpacing(), getHorizontalElementSpacing());
	}
	layoutData->setPosition(position);
	element->setLayoutData(*layoutData);
	element->layoutChanged();
	LOG_EXIT
}

//! element will be freed by this
//! 
void CSGrafikElement::addMainElement(CSGrafikElement *element, int x, int y)
{
	static char *functionName="addElement";
	LOG_ENTER 
	// there can only be one parent! (not tested - the old one is simply removed!)

//	to in layout wird hier eine message box geaddet und damit der aktuelle
//	layout iterator drcheinander gewirbelt
// todo add as "add pending" und dann nach dem layout als lezttes

	if (element->getParent())
	{
		element->getParent()->removeMainElement(element);
	}
	element->setParent(this);
	mElements.push_back(element);
	mElementBackadded = true;
	CSLayoutData *layoutData = element->getLayoutDataInternal();
	if (!layoutData->getSpacingSet())
	{
		layoutData->setSpacing(getVerticalElementSpacing(), getHorizontalElementSpacing());
	}
	layoutData->setX(x);
	layoutData->setY(y);
	element->setLayoutData(*layoutData);
	element->layoutChanged();
	LOG_EXIT
}

CSGrafikElement *CSGrafikElement::getUppermostAncestor()
{
	static char *functionName="getUppermostAncestor";
	CSGrafikElement *parent = this;
	CSGrafikElement *lastParent = this;
	while (parent != 0)
	{
		lastParent = parent;
		parent = parent->getParent();
	}
	return lastParent;
}

// perhaps this is not right
// glass Pane?
// now it works without interfering with (Border) Layoutmanager, but why?
void CSGrafikElement::addFrontElement(CSGrafikElement *element, int x, int y) // koordinataes in Screen!
{
	CSDesktop *desktop = getDesktop();
	if (desktop == 0)
	{
		return;
	}
	element->mFrontElement = true;
	getLayoutDataInternal()->setX(x);
	getLayoutDataInternal()->setY(y);
	desktop->addMainElement(element, x, y);
}

void CSGrafikElement::addFrontElementCenter(CSGrafikElement *element) // koordinataes in Screen!
{
	CSDesktop *desktop = getDesktop();
	if (desktop == 0)
	{
		return;
	}
	element->mFrontElement = true;
	getLayoutDataInternal()->setX(desktop->getWidth()/2 - element->getWidth()/2);
	getLayoutDataInternal()->setY(desktop->getHeight()/2 - element->getHeight()/2);
	desktop->addMainElement(element, desktop->getWidth()/2 - element->getWidth()/2, desktop->getHeight()/2 - element->getHeight()/2);
}

void CSGrafikElement::removeFrontElement(CSGrafikElement *element)
{
	CSDesktop *desktop = getDesktop();
	if (desktop == 0)
	{
		return;
	}
	element->mFrontElement = false;
	desktop->removeMainElement(element);
}

void CSGrafikElement::setVisible(bool v) 
{
	if (v != mVisible)
	{
		mVisible = v;
		layoutChanged();
	}
}

void CSGrafikElement::deleteElements()
{
	CSDesktop *desktop = getDesktop();
	if (mMainElement != this)
	{
		CSGrafikElements::iterator iter = mElements.begin();
		while (iter != mElements.end())
		{
			CSGrafikElement *element = *iter;
			if (element == mMainElement)
			{
				mElements.erase(iter);
				if (desktop != 0)
				{
					desktop->correctElementHandling(element);
				}
				delete (element);
				break;
			}
			iter++;
		}
		mMainElement = this;
	}

	CSGrafikElements::iterator iter = mElements.begin();
	while (iter != mElements.end())
	{
		CSGrafikElement *element = *iter;
		mElements.erase(iter);
		delete (element);
		iter = mElements.begin();
	}
}

// only unhandled messages are "handled" here
// unhandled messaged get passed (if possible) to their parent
// xy position ist converted to parent position!
void CSGrafikElement::reactOnMessageGrafikElement(CSMessage *message) 
{
	static char *functionName="reactOnMessage";
	if (message->mIsHandled)
	{
		return;
	}
	switch (message->getType())
	{
		case GUI_MESSAGE:
		{
			GuiMessage *gm = (GuiMessage *) message;
			CSGrafikElement *parent = getParent();
			if (!parent) 
			{
				return;
			}
//			gm->receiverX += parent->getViewportX();
//			gm->receiverY += parent->getViewportY();
//			parent->reactOnMessage(gm);
			break;
		}
	}
}

//const int SIZING_TYPE_FIXED = 1;
//const int SIZING_TYPE_STRETCHED = 2;
//const int SIZING_TYPE_PACKED = 4;
int CSGrafikElement::getSizingType()
{
	CSLayoutData layout = getLayoutData();
	bool stretched = layout.getStretchedHorizontal() ||	layout.getStretchedVertical();
	bool packed = layout.getPackedHorizontal() ||	layout.getPackedVertical();
	if (stretched)
	{
		return SIZING_TYPE_STRETCHED;
	}
	if (packed)
	{
		return SIZING_TYPE_PACKED;
	}
	return SIZING_TYPE_FIXED;
}

void CSGrafikElement::layoutChanged(bool positionChanged, bool sizeChanged)
{
/*
jedes element kann nur einen von drei Zustnden (bzgl Gre) haben:
 fix (gesetzt)
 min (packed)
 max (stretched)

 layoutChanged()
1. sperren, wenn im doLayout()
2. alle changes zu parent propagieren
3. propagierte changes sehen anders aus, als eigene changes (type) [elementAnzahlChanged, posChange, sizeChange]
   - wenn parent fix, dann niemals propagieren!
   - wenn parent min / max dann immer propagieren!

doLayout()
1. alle changes fr betroffene setzen (typed)
2. changes sperren
3. algorithmus:

 parent = getParent()
 while (!parent->isFixed())
	 parent = parent->parent();
*/
	
	if (mInLayout!=0)
	{
		return;
	}
	
	mLayoutChanged = true;
	mChangeType = 0;

	if (positionChanged)
	{
		mChangeType += CHANGE_TYPE_POSITION;
	}
	if (sizeChanged)
	{
		mChangeType += CHANGE_TYPE_SIZE;
	}
	if (mChangeType == 0)
	{
		mChangeType = CHANGE_TYPE_UNKOWN;
	}

	CSDesktop *desktop = getDesktop();
	if (desktop)
	{
		desktop->requestLayout();
	}
}

void CSGrafikElement::doLayout()
{
	static int watch=0;
	watch++;
	// don't allow any changed while in layout!	
	mInLayout++;
	
	if (mLayoutChanged)
	{
		mLayoutChanged = false;
		layoutComponent();
		mChangeType = 0;
	}

	if (watch==17)
	{
		printf("");
	}
		printf("");
	
	CSGrafikElements::iterator iter = mElements.begin();
	while (iter != mElements.end())
	{
		CSGrafikElement *element = (*iter);
		mElementBackadded = false;
		if (element->getVisible())
		{
			element->doLayout();

			// check if layout caused elements structure to be changed
			// actually this should not happen if elements are only layouted
			// but as html is implemented now, the
			// layout is changes elements, and that may xause errors, erros messages and popup messages
			// these CAN be added in between - ist sort of REALLY BAD!
			if (mElementBackadded)
			{
				iter = mElements.begin();
				while ((iter != mElements.end()) && (element != (*iter)))
					iter++;
				if (iter == mElements.end())
					break;
			}
		}
		iter++;
	}

	mInLayout--;
}

void CSGrafikElement::layoutComponent()
{
	// todo: respect type of changes
	// look what kind of change occured (pos, add, size)

	// and what type of element am I (fix, min, max)
	int sizingType = getLayoutDataInternal()->getSizingType();
	if (sizingType == SIZING_MAX)
	{
		if ((mParent) && (isParentLayoutUpdate()))
		{
			mParent->layoutFromChild();
		}
		layoutSelf();
	}
	else // MIN & FIX
	{
		layoutSelf();
		if ((mParent) && (isParentLayoutUpdate()))
		{
			mParent->layoutFromChild();
		}
	}	

	if (( (mChangeType&CHANGE_TYPE_SIZE) == CHANGE_TYPE_SIZE) || ((mChangeType&CHANGE_TYPE_UNKOWN) == CHANGE_TYPE_UNKOWN))
	{
		layoutChildren();
	}
}
	
void CSGrafikElement::layoutChildren()
{
	CSGrafikElements::iterator iter = mElements.begin();
	while (iter != mElements.end())
	{
		CSGrafikElement *element = (*iter);
		if (element->getVisible())
		{
			element->layoutFromParent();
		}
		iter++;
	}
}

void CSGrafikElement::layoutFromParent()
{
	layoutSelf();
	layoutChildren();
}
	
void CSGrafikElement::layoutSelf()
{
	if (isLayoutAllowed())
	{
		layoutSetup();
		mLayoutManager->layout();
	}
}

void CSGrafikElement::layoutFromChild()
{
	// look what kind of change occured (pos, add, size)
	// and what type of element am I (fix, min, max)
	int sizingType = getLayoutDataInternal()->getSizingType();
	if (sizingType == SIZING_MAX)
	{
		if (mParent)
		{
			mParent->layoutFromChild();
		}
	}
	layoutSelf();
}
