// USES SDL_Surface
// USES SDL_Rect 

#ifndef CSGRAFIK_ELEMENT_h
#define CSGRAFIK_ELEMENT_h

#ifdef WIN32
#pragma warning(disable : 4786 )
#endif

const int SIZING_TYPE_FIXED = 1;
const int SIZING_TYPE_STRETCHED = 2;
const int SIZING_TYPE_PACKED = 4;

const int CHANGE_TYPE_UNKOWN = 1;	
const int CHANGE_TYPE_SIZE = 2;	
const int CHANGE_TYPE_POSITION = 4;	
const int CHANGE_TYPE_ADD = 7;	
const int CHANGE_TYPE_REMOVE = 7;	


const int BORDER_STATE_VISIBLE = 1;
const int BORDER_STATE_PRESSED = 2;
const int BORDER_STATE_HOVERING = 4;
const int BORDER_STATE_FOCUSED = 8;
const int BORDER_STATE_ACTIVATED = 16;
const int BORDER_STATE_ENABLED = 32;

const int BORDER_STATE_DEFAULT = BORDER_STATE_VISIBLE;

// General GUI constant definitions
const int BUTTON_VISIBLE =		BORDER_STATE_VISIBLE;	//!< Button state -> pressed
const int BUTTON_PRESSED =		BORDER_STATE_PRESSED;	//!< Button state -> pressed
const int BUTTON_HOVERING =		BORDER_STATE_HOVERING;	//!< Button state -> hovering (unpressed) (implemented in \a CSMenu)
const int BUTTON_FOCUSED =		BORDER_STATE_FOCUSED;	//!< Button state -> focused (unpressed) (not done yet)
const int BUTTON_ACTIVATED =	BORDER_STATE_ACTIVATED;	//!< Button state -> activated (unpressed) (not done yet)
const int BUTTON_ENABLED =		BORDER_STATE_ENABLED;	//!< Button state -> disabled (unpressed) (not done yet)

const int BUTTON_DEFAULT =		BUTTON_VISIBLE + BUTTON_ENABLED;	//!< Button state -> unpressed

const int POSITION_NOT_SET = -1;//!< positon information is not given
const int POSITION_NORTH = 0;	//!< positon NORTH, item is placed above the element with the smallest y - pos value
const int POSITION_SOUTH = 1;	//!< positon SOUTH, item is placed below the element with the greatest y - pos value
const int POSITION_EAST = 2;	//!< positon EAST, item is placed left the element with the greatest x - pos value
const int POSITION_WEST = 3;	//!< positon WEST, item is placed right the element with the smallest y - pos value
const int POSITION_CENTER = 4;	//!< positon CENTER, makes only sense if component is autosizing both x and y
const int POSITION_ABSOLUT = 5; //!< not used

//! same as directions, if changed to something unique, one must add a nonViewported special case in buildArea() in LayoutManager!
const int POSITION_ORDER_TOP	= POSITION_NORTH; //!< same as POSITION_NORTH
const int POSITION_ORDER_BOTTOM = POSITION_SOUTH; //!< same as POSITION_SOUTH
const int POSITION_ORDER_LEFT	= POSITION_WEST;  //!< same as POSITION_WEST
const int POSITION_ORDER_RIGHT	= POSITION_EAST;  //!< same as POSITION_EAST

const int TYPE_HORIZONTAL = 0;	//!< element is arranged horizontal
const int TYPE_VERTICAL = 1;	//!< element is arranged vertical

const int TYPE_CHANGE_UP = 1;	
const int TYPE_CHANGE_INIT = 2;
const int TYPE_CHANGE_DOWN = -1;
const int TYPE_CHANGE_ONLY_ME = -2;

const int LAYOUT_ELEMENT = 1;	// add remove visibility
const int LAYOUT_INTERNAL = 8;	// font, border, area... changed
const int LAYOUT_LAYOUT = 16;	// layourt changed
const int LAYOUT_MOVE = 2;
const int LAYOUT_SIZE = 4;
const int LAYOUT_NOT = 0;

#include <vector>
#include <string>

#include "SDLMain.h"
#include "CSObject.h"
#include "CSMessageListener.h"
#include "CSLAF.h"
#include "CSLog.h"
#include "CSLayoutManager.h "
#include "CSLayoutManagerBorder.h "
#include "CSLayoutManagerStackVertical.h "
#include "CSLayoutManagerStackHorizontal.h "
#include "CSLayoutManagerXY.h "
#include "CSInset.h"

class CSGrafikElement;
class CSFont;
class CSBorder;
class CSDesktop;
class CSWindow;
class CSScrollbar;
class CSArea;
class CSLayoutManager;

typedef std::vector<CSGrafikElement *> CSGrafikElements;

const int MOUSE_CURSOR_TYPE_DEFAULT = 0;
const int MOUSE_CURSOR_TYPE_SIZING = 1;

/**
	This is the base class of all GUI elements.
	All GUI elements are derived from this class.
	The "base" class at runtime is the CSDesktop class. To a desktop all Gui 
	elements are added (or to its children).

    The GUI elements are event driven (as almost all GUI's are). The events are
	dispatched from the desktop. The desktop itself gets "hardware" events (in this
	case SDL events) from outside (simple method calls). 

    Within this set of classes the events are generated at runtime from the 
	CSGame class, which "polls" SDL every "game round".

    Events are send to listeners. One can simply add a listener with the "addlistener" method.

    C++ does not per default have (java has that!) the ability to call a super() method.
	(a method that is overloaded by a child method)

    With event passing and wanting a default behaviour for all classes, a strategy (convention)
	was used within the gui classes, that allows us to call the parents overloaded 
	methods from the child.
	
	The default method to wait for events is:
	<PRE>
		reactOnMessage(...)
	</PRE>
	
	That method is virtual and therefor overloadable.
	Each Gui-element overloads this method.

    BUT:

    Each element also implements a method (non virtual, protected or public)
	<PRE>
  		reactOnMessage"ElementName"(...)
	</PRE>

    The "reactOnMessage()" allway! only calls the element based method. Therefor
	each child can also call (all) the parent methods!

	e.g.
	CSButton implements following methods:

	<PRE>
  		void reactOnMessageDragButton(CSMessage *message);
		virtual void reactOnMessage(CSMessage *message);
	</PRE>

    the bodies might look like:

	<PRE>
		void CSButton::reactOnMessage(CSMessage *message)
		{
			reactOnMessageButton(message);
		}

		void CSButton::reactOnMessageButton(CSMessage *message)
		{
			static char *functionName="reactOnMessage";
			if (message->mIsHandled)
			{
				reactOnMessageGrafikElement(message);
				return;
			}
			... handle events
			reactOnMessageGrafikElement(message);
		}
	</PRE>

	Thus the parent methods are allways called, and can still
	handle events not handled by the subclass.
	(This can even be used further - see e.g. MenuItems...)

    
    This class implements the basic methods all gui classes need.
    Child classes must overload the paint(destination) method to display themselves.
   
    All GrafikElements can be added to another GrafikElement. (But not everything makes sense - 
    like adding a desktop to a window...)
    The strukture should be like: 
	<PRE>
			Screen-> Desktop->Window->Elements (like Label, Button, Textfiled)
	</PRE>
  */
class CSGrafikElement : public CSMessageDispatchable, public CSMessageListener
{
	friend CSDesktop;			// because it does a modal Desktop "painting"
	friend CSLayoutManager;			
	friend CSWindow;			// for getElements
	private:

		static long mInLayout;
		int mChangeType;
		int getSizingType();

		CSLayoutManager *mLayoutManager;
		CSLayoutData mLayoutData;
		bool mFrontElement;
		bool mLayoutChanged;
		bool mLayoutAllowed;
		bool mParentUpdateAllowed;
		bool mElementBackadded;


		CSGrafikElement *mMainElement;
		int mX;					//!< Position in parent
		int mY;					//!< Position in parent
		int mViewportX;			//!< "physical" start of the viewport within this element (if there is viewport)
		int mViewportY;			//!< "physical" start of the viewport within this element (if there is viewport)
		int mViewportHeight;	//!< "physical" Dimension of viewport
		int mViewportWidth;		//!< "physical" Dimension of viewport
		int mViewportOffsetX;	//!< where the viewport starts displaying the childElements!
		int mViewportOffsetY;	//!< where the viewport starts displaying the childElements!
		SDL_Rect mElementArea;	//!< mX, mY, mHeight, mWidth (for easier updating)
		SDL_Rect mViewportArea;	//!< mViewportX, mViewportY, mViewportHeight, mViewportWidth (for easier updating)
		int getElementMaxX();
		int getElementMaxY();
		//! not deleting, no listener removed!, only removing!
		void removeMainAllElements() {mElements.clear(); }		
		CSGrafikElement *getElement(int x, int y, SDL_Rect *parentViewport, int &receiverX, int &receiverY); //!< get (sub-) element, that is at the given (screen) coordinates 
		CSGrafikElements *getElements() {return &mElements; }//!<  this returns the "real thing - do not change!
		
		bool mFocusable;		//!< can this element receive the focus of a desktop?

		bool mFocused;			//!< is this element in focus of a desktop?
		bool mModal;			//!< is this element modal (makes sense only for windows)
		bool mEnabled;
		bool mActivated;
		bool mActive;

		bool mBackGroundColorSet;	//!< if background color is set from outside, even after a layout(), the once set color is used
		int mBackgroundColor;		//!< Background color of component
		
		bool mTextColorSet;			//!< if text color is set from outside, even after a layout(), the once set color is used
		int mTextColor;				//!< Text color of component
		
		bool mHorizontalElementSpacingSet;
		int mHorizontalElementSpacing;
		bool mVerticalElementSpacingSet;
		int mVerticalElementSpacing;

		int mCurrentLAFId;		//!< currently ID of the LAF in use
		int mCursorType;		//!< what kind of mouse cursor should the cursor, when over this element?

		CSBorder *mBorder;		//!< the currently set CSBorder
		CSArea *mArea;
		CSFont *mFont;			//!< the currently set CSFont
		int mStyle;
		std::string mType;
		static CSGrafikElements mLostElements;
		static void addLostElement(CSGrafikElement *element)
		{
			mLostElements.push_back(element);
		}
		static void clearLostElements()
		{
			mLostElements.clear();
		}

		static bool isLostElement(CSGrafikElement *element)
		{
			CSGrafikElements::iterator iter = mLostElements.begin();
			while (iter != mLostElements.end())
			{
				CSGrafikElement *lostElement = *iter;
				if (element == lostElement)
				{
					return true;
				}
				iter++;
			}
			return false;
		}
		void layoutComponent();
		void layoutSelf();
		void layoutChildren();
		void layoutFromParent();
		void layoutFromChild();
		void changeLAFDefaults();

	protected:
		bool mVisible;			//!< is this element visible?
		CSInset mInset;			//!< the currently set CSInset
		CSInset *getInsetInternal() {return &mInset;}
		CSLayoutData *getLayoutDataInternal() {return &mLayoutData;} ///< pointer is faster, but more dangerous


		void setLayoutAllowed(bool b) {mLayoutAllowed=b;}
		bool isLayoutAllowed() {return mLayoutAllowed;}

		
		void deleteElements();
		virtual int getViewportWidth() {return mViewportWidth;}	//!< Width that elements, that are added to this component can use
		virtual int getViewportHeight() {return mViewportHeight;}	//!< Height that elements, that are added to this component can use
		virtual int getViewportX() {return mViewportX;}	//!< Coordinates in relation to this element, 0,0 is left upper corner
		virtual int getViewportY() {return mViewportY;}	//!< 
		
		int mBorderState;
		bool mHeightSet;		//!< was height of component set from outside
		bool mWidthSet;			//!< was width of component set from outside
		int mHeight;			//!< "physical" Dimension of component (as is seen by parent)
		int mWidth;				//!< "physical" Dimension of component (as is seen by parent)
		int mMinWidth;			//!< without border, without spacing, without inset
		int mMinHeight;			//!< without border, without spacing, without inset

		int mActionId;			//!< the action id (if there is one) of an element (e.g Button)
		void *mActionObject;	//!< 


		CSGrafikElements mElements;	//!< all child elements are stored here
		CSGrafikElement *mParent;	//!< the direct parent (if any) is stored here

		std::string mTooltipText;	// text of tooltip (default "")

		GuiMessage MESSAGE_FOCUS_GAINED;	//!< Message that can be sent from CS Graphic element (CSDesktop)
		GuiMessage MESSAGE_FOCUS_LOST;		//!< Message that can be sent from CS Graphic element (CSDesktop)
		GuiMessage MESSAGE_KEY_PRESSED;		//!< Message that can be sent from CS Graphic element (CSDesktop)
		GuiMessage MESSAGE_KEY_RELEASED;	//!< Message that can be sent from CS Graphic element (CSDesktop)
		GuiMessage MESSAGE_MOUSE_PRESSED;	//!< Message that can be sent from CS Graphic element (CSDesktop)
		GuiMessage MESSAGE_MOUSE_RELEASED;	//!< Message that can be sent from CS Graphic element (CSDesktop)
		GuiMessage MESSAGE_MOUSE_MOTION;	//!< Message that can be sent from CS Graphic element (CSDesktop)
		GuiMessage MESSAGE_MOUSE_MOTION_LOST;	//!< Message that can be sent from CS Graphic element (CSDesktop)

		SDL_Rect getElementArea() {return mElementArea;}


		//! Set the parent element of this
		virtual void setParent(CSGrafikElement *parent);
		void init(int height, int width);		//!< initializes this element, called once from constructor
		//! Translates the given element coordinates to Desktop coordinates
		void translateToDesktop(int &x, int &y);
		//! Builds the viewport sizes from \a mHeight and \a mWidth (and sets the rectangles)
		// void buildArea();
		virtual void CSGrafikElement::paintArea(SDL_Surface *destination, SDL_Rect *parentArea)
		{
			paintAreaStandard(destination, parentArea);
		}
		void paintBorder(SDL_Surface *destination, SDL_Rect *parentArea);
		void paintAreaStandard(SDL_Surface *destination, SDL_Rect *parentArea);
		void paintChildren(SDL_Surface *destination, SDL_Rect *parentViewport); 
		CSGrafikElement *removeMainElement(CSGrafikElement *element);
		CSLayoutManager *getMainLayoutManager()
		{
			return mLayoutManager;
		}

		void setMainLayoutManager(CSLayoutManager *manager)
		{
			if (mLayoutManager != 0)
			{
				delete mLayoutManager;
				mLayoutManager = 0;
			}
			mLayoutManager = manager;
			if (mLayoutManager != 0)
			{
				mLayoutManager->setHostingElement(this);
			}
		}

		virtual void rebuildElement() {} // if a component is made up of elements dependend of LAF, than this method is called
								 // during a lafChanged, so element can be changed (like Icons...)
		// protected, so desktop can call it befor painting
		void doLayout();	// return negative -> parentChangerelevant, positive -> children relevant
		void setMainElement(CSGrafikElement *element) {mMainElement = element;}
	
		void addMainElement(CSGrafikElement *element, int position);
		void addMainElement(CSGrafikElement *element, int x, int y);
		void addMainElement(CSGrafikElement *element);
	public:
		static const char *CLASS;				///< static element, name of this class (introsepection)
		virtual std::string getType() {return (std::string) CLASS;}

		CSGrafikElement(int height, int width);
		~CSGrafikElement(void);
		CSGrafikElement *getUppermostAncestor();
		CSWindow *getParentWindow();
		CSDesktop *getDesktop(void);				///< get the next parent, that is a CSDesktop (or null)
		CSGrafikElement *getParent(void) {return mParent;}
		virtual void paint(SDL_Surface *destination, SDL_Rect *parentViewport)
		{
			paintChildren(destination, parentViewport);
		}
		void setPosition(int x, int y); 		///< set X, Y coordinates in relation to parent (useaable Area)
		bool isFrontElement() {return mFrontElement;}
		void toFront(CSGrafikElement *element);
		void setGUIType(std::string type) {mType = type;}
		std::string getGUIType() {if (mType.size() == 0) return getType(); return mType;}
		int getX() {return mX;} 				///< X coordinates in relation to parent
		int getY() {return mY;} 				///< Y coordinates in relation to parent

		//!< expects coordinates in relation to THIS (current) element
		bool isIn(int x, int y, SDL_Rect *parentViewport);			//!< coordinates of screen 
		void reactOnMessageGrafikElement(CSMessage *message);
		virtual void reactOnMessage(CSMessage *message) {reactOnMessageGrafikElement(message);}	///< to be overloaded
		virtual bool isTransparent(int x, int y, SDL_Rect *parentViewport) {return false;}			///< to be overloaded (default false)
		virtual void putString(SDL_Surface *destination, SDL_Rect *parentViewport, int x1, int y1, const std::string &string); 	///< coordinates relativ to this element
		virtual void putString(SDL_Surface *destination, SDL_Rect *parentViewport, int x1, int y1, int color, const std::string &string); 	///< coordinates relativ to this element
		
		void line(SDL_Surface *destination, SDL_Rect *parentViewport, int x1, int y1, int x2, int y2, Uint32 color); 		//!< coordinates relativ to this element
		void setActionId(int id) {mActionId=id;};
		int getActionId() {return mActionId;};
		void setActionObject(void *object) {mActionObject=object;};
		void *getActionObject() {return mActionObject;};

		virtual void lafChanged(int id);	//!< each element is registered to CSLAF, if LAF changes, this method is called (default does nothing!)
		bool isChild(CSGrafikElement *element);
		virtual void cursorDefault();
		virtual void cursorSizing();
		virtual CSMouseCursor *getCursor();
		void setCurrentLAFId(int id) {mCurrentLAFId = id;}
		int getCurrentLAFId(void) {return mCurrentLAFId;}
		virtual void setTooltipText(const std::string &s) {mTooltipText = s;}
		virtual std::string getTooltipText() {return mTooltipText;}

		virtual void layoutSetup(){}
		virtual bool getActive(){return mActive;}
		virtual void setActive(bool b){mActive = b;}
		virtual void setEnabled(bool b){mEnabled = b;changeLAFDefaults();}
		virtual bool getEnabled(){return mEnabled;}
		virtual void setActivated(bool b){mActivated = b;}
		virtual bool getActivated(){return mActivated;}
		virtual bool getModal() {return mModal;}
		virtual void setModal(bool m) {mModal = m;}
		virtual bool getVisible() {return mVisible;}
		virtual void setVisible(bool v); 
		virtual void setFocusable(bool b);
		virtual bool getFocusable() {return mFocusable;}
		bool getFocused() {return mFocused;}
		void setFocused(bool b) {mFocused = b;} 	///< only to be used from CSDesktop!
		virtual SDL_Rect getViewportArea() {return mViewportArea;}

		int getHeight() {if (mHeight<mMinHeight) return mMinHeight; return mHeight;}		///< height of component
		int getWidth() {if (mWidth<mMinWidth) return mMinWidth; return mWidth;}			///< width of component
		void setHeight(int h);					///< set height of component
		void setWidth(int w);					///< set width of component
		virtual CSFont *getFont(); //!<  returns the given gui font (see CSLAF)
		
		//! Font is not deleted here, since it is loaded by a LOADER::INSTANCE!
		//! (which keeps track of the instances!)
		virtual void setFont(CSFont *font) 
		{
			if (mFont != font)
			{
				mFont = font;
				layoutChanged();
			}
		}

		virtual void setInset(const CSInset &inset) {mInset = inset;}
		virtual CSInset getInset() {return mInset;}
		virtual void setBorder(CSBorder *border, bool deleteOldBorder = false);
		virtual CSBorder *getBorder();
		virtual void setArea(CSArea *area, bool deleteOldArea = false);
		virtual CSArea *getArea();
		int getHorizontalElementSpacing();
		void setHorizontalElementSpacing(int horizontalElementSpacing);
		int getVerticalElementSpacing();
		void setVerticalElementSpacing(int verticalElementSpacing);
		
		int getBackgroundColor();
		void setBackgroundColor(int backgroundColor);
		int getTextColor();
		void setTextColor(int textColor);
		void setStyle(int style) {mStyle = style;}
		void setBorderState(int state){mBorderState = state;}

		CSLayoutManager *getLayoutManager()
		{
			if (mMainElement == this)
			{
				return getMainLayoutManager();
			}
			else
			{
				return mMainElement->getLayoutManager();
			}
		}
		void setLayoutManager(CSLayoutManager *manager)
		{
			if (mMainElement == this)
			{
				setMainLayoutManager(manager);
			}
			else
			{
				mMainElement->setLayoutManager(manager);
			}
		}

		CSLayoutData getLayoutDataMainElement() ///< pointer is faster, but more dangerous
		{
			if (mMainElement)
			{
				return mMainElement->getLayoutData();
			}
			return mLayoutData;
		} 
		
		void setLayoutDataMainElement(const CSLayoutData &layoutData) 
		{
			if ((mMainElement!=this) && (mMainElement))
			{
				mMainElement->setLayoutData(layoutData);
			}
		}

		CSLayoutData getLayoutData() ///< pointer is faster, but more dangerous
		{
			return mLayoutData;
		} 
		
		void setLayoutData(const CSLayoutData &layoutData) 
		{
			mLayoutData = layoutData;
			setPosition(mLayoutData.getX(), mLayoutData.getY()); 		///< set X, Y coordinates in relation to parent (useaable Area)
			mLayoutManager->buildArea(this);
			layoutChanged();
		}


		//! each and every CSLayoutManager is freed by the CSGraphicElement, which
		//! receives it. Freed either in th edestructor, or by setting a new CSLayoutManager!
		void addElement(CSGrafikElement *element, int position);
		void addElement(CSGrafikElement *element, int x, int y);
		void addElement(CSGrafikElement *element);
		void addFrontElement(CSGrafikElement *element, int x, int y); // koordinataes in Screen!
		void addFrontElementCenter(CSGrafikElement *element); // koordinataes in Screen!

		int getMinWidth() {if (getVisible()) return mMinWidth; return 0;}		//!< if not set defaults to 5 pixel
		void setMinWidth(int w) {mMinWidth = w;}
		int getMinHeight() {if (getVisible()) return mMinHeight; return 0;}		//!< if not set defaults to 5 pixel
		void setMinHeight(int h) {mMinHeight = h;}
		void removeFrontElement(CSGrafikElement *element);
		CSGrafikElement *removeElement(CSGrafikElement *element);

		void setViewportOffset(int x, int y)
		{
			mViewportOffsetX = x;
			mViewportOffsetY = y;
		}

		void getViewportOffset(int &x, int &y)
		{
			x = 0;
			y = 0;
			if (mParent)
			{
//				mParent->getViewportOffset(x, y);
			}
			x += mViewportOffsetX;
			y += mViewportOffsetY;
		}
		void getViewportOffsetParent(int &x, int &y)
		{
			x = 0;
			y = 0;
			if (mParent)
			{
				mParent->getViewportOffsetParent(x, y);
			}
			x += mViewportOffsetX;
			y += mViewportOffsetY;
		}
		
		void removeAllElements() 
		{
			if (mMainElement == this)
			{
				removeMainAllElements();
			}
			else
			{
				mMainElement->removeAllElements();
			}
		}		
		void layoutChanged(bool positionChanged = true, bool sizeChanged = true);
		void setParentLayoutUpdate(bool b) {mParentUpdateAllowed=b;}
		bool isParentLayoutUpdate() {return mParentUpdateAllowed;}
};
#endif // CSGRAFIK_ELEMENT_h











