#ifndef CSLayouManager_h
#define CSLayouManager_h

// USES SDL_Surface

#ifdef WIN32
#pragma warning(disable : 4786 )
#endif

#include <vector>
#include <string>

class CSGrafikElement;
typedef std::vector<CSGrafikElement *> CSGrafikElements;
class CSLayoutManager;

const int SIZING_UNKOWN = 0;
const int SIZING_MAX = 1;
const int SIZING_MIN = 2;
const int SIZING_FIX = 4;
//typedef std::vector<CSLayoutManager *> CSLayoutManagers;

/** LayoutData is responsible ONLY for the GraphicElement it is set to.
*/
class CSLayoutData
{
	private:
		int mX;						// Position in Parent
		int mY;						// Position in Parent
		int mPosition;				// Position in Parent
		bool mCenteredHorizontal;	// centered in Parent
		bool mCenteredVertical;		// centered in Parent
		bool mStretchedHorizontal;	// stretched self
		bool mStretchedVertical;	// stretched self
		bool mPackedHorizontal;		// packed self
		bool mPackedVertical;		// packed self		
		bool mOversize;				// don`t allow element to grow larger than parent
		int mVerticalSpacing;		// spacing of children inside
		int mHorizontalSpacing;		// spacing of children inside
		bool mSpacingSet;
	public:
		CSLayoutData()
		{
			init();
		}
		CSLayoutData(int x, int y)
		{
			init();
			mX = x;
			mY = y;
		}

		CSLayoutData(int position, bool stretchAndCentered = false)
		{
			init();
			mPosition = position;
			mCenteredHorizontal = stretchAndCentered;
			mCenteredVertical = stretchAndCentered;
			mStretchedVertical = stretchAndCentered;
			mStretchedHorizontal = stretchAndCentered;
		}
		void init();
		bool getStretched()
		{
			return mStretchedVertical||mStretchedHorizontal;
		}
		bool getPacked()
		{
			return mPackedVertical||mPackedHorizontal;
		}
		void setOversize(bool b) {mOversize = b; } // only implemented for false in Border Layout
		bool getOversize() {return mOversize; }

		void setPackedHorizontal(bool c) {mPackedHorizontal = c; if (c) mStretchedHorizontal = false;}
		const bool getPackedHorizontal() {return mPackedHorizontal;}
		void setPackedVertical(bool c) {mPackedVertical = c; if (c) mStretchedVertical = false;}
		const bool getPackedVertical() {return mPackedVertical;}
		void setCenteredHorizontal(bool c) {mCenteredHorizontal = c;}
		const bool getCenteredHorizontal() {return mCenteredHorizontal;}
		void setCenteredVertical(bool c) {mCenteredVertical = c;}
		const bool getCenteredVertical() {return mCenteredVertical;}
		const bool getStretchedVertical() {return mStretchedVertical;}
		void setStretchedVertical(bool s) {mStretchedVertical = s;if (s) mPackedVertical = false;}
		const bool getStretchedHorizontal() {return mStretchedHorizontal;}
		void setStretchedHorizontal(bool s) {mStretchedHorizontal = s;if (s) mPackedHorizontal = false;}
		void setSpacing(int v, int h) {mVerticalSpacing = v; mHorizontalSpacing = h; mSpacingSet = true;}
		const void getSpacing(int &v, int &h) {v = mVerticalSpacing; h = mHorizontalSpacing;}
		const int getX() {return mX;}
		const int getY() {return mY;}
		void setX(int x) {mX = x;}
		void setY(int y) {mY = y;}
		const int getPosition() {return mPosition;}
		void setPosition(int position) {mPosition = position;}
		bool getSpacingSet() {return mSpacingSet;}
		bool isFixSized()
		{
			return !(getPacked() || getStretched());
		}

		// notice! both methods isMinSized() and isMaxSized() can return return true, sind
		// it is possible to have packing in one axis and stretching on the other axis
		bool isMinSized()
		{
			return getPacked();
		}
		bool isMaxSized()
		{
			return getStretched();
		}
		int getSizingType()
		{
			if (isMaxSized())
				return SIZING_MAX;
			if (isMinSized())
				return SIZING_MIN;
			if (isFixSized())
				return SIZING_FIX;
			return SIZING_UNKOWN;
		}

};

class CSLayoutManager
{
	private:
		CSGrafikElement *mHostingElement;

	protected:
		bool mPropagateSizeChangesOnly;
		CSGrafikElement *getHostingElement() {return mHostingElement;}
		CSGrafikElement *getParent(CSGrafikElement *element);
		CSGrafikElements *getElements(CSGrafikElement *element);
		bool setLayout(CSGrafikElement *visitor, int x, int y, int width ,int height);
		CSLayoutData *getLayoutData(CSGrafikElement *visitor);
		int getViewportWidth(CSGrafikElement *element);
		int getViewportHeight(CSGrafikElement *element);
		int getWidth(CSGrafikElement *element);
		int getHeight(CSGrafikElement *element);
		int getMinWidth(CSGrafikElement *element);
		int getMinHeight(CSGrafikElement *element);
		int getParentWidth(CSGrafikElement *element);
		int getParentHeight(CSGrafikElement *element);
		int getBorderWidth(CSGrafikElement *element);
		int getBorderHeight(CSGrafikElement *element);
		int getInsetWidth(CSGrafikElement *element);
		int getInsetHeight(CSGrafikElement *element);
		bool isNextUnstretchedParentHorizontalFixed(CSGrafikElement *element);
		bool isNextUnstretchedParentVerticalFixed(CSGrafikElement *element);
		virtual void layoutElements(){}
		

	public:
		static const char *CLASS;
		virtual std::string getType() {return (std::string) CLASS;}
		CSLayoutManager() 
		{
			mHostingElement = 0;
			mPropagateSizeChangesOnly = false;
		}
		virtual ~CSLayoutManager() {}
		void setHostingElement(CSGrafikElement *hostingElement) {mHostingElement = hostingElement;}
		virtual int getMinimumWidth(CSGrafikElement *element) = 0;	
		virtual int getMinimumHeight(CSGrafikElement *element) = 0;	
		//! not all LayoutManagers support Packing!
		//! Packing changes the size of the element and possibly the size of
		//! stretching children
		//! it does not change the position of THIS element
		//! Packing is done AFTER layout is finished
		//! it cannot be called straight, but must be set with setPack(true);
		//		virtual void pack() {}	
		void pack(CSGrafikElement *element);
		void layout(); //!< layout does not change the size of the component! (might change size of children!)
		void setPropagateSizeChangesOnly(bool b) {mPropagateSizeChangesOnly = b;}
		bool getPropagateSizeChangesOnly() {return mPropagateSizeChangesOnly;}
		void buildArea(CSGrafikElement *element);
};

#endif // CSLayouManager_h
